///////////////////////////////////////////////////////////////////////////////
// DC_CombineImages_dialog.js
///////////////////////////////////////////////////////////////////////////////
//
// DC_CombineImages: Version 0.61
//
// A script to combine multiple grayscale images to form a single color image
// with user-defined colors
//
// Copyright (c) 2018-2022 Dean Carr
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __DC_CombineImages_dialog_js
#define __DC_CombineImages_dialog_js

//------------------------------------------------------------------------------

#include <pjsr/Sizer.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/StdButton.jsh>
#include <pjsr/StdIcon.jsh>
#include <pjsr/ColorSpace.jsh>
#include <pjsr/Interpolation.jsh>
#include <pjsr/UndoFlag.jsh>
#include <pjsr/ColorComboBox.jsh>
#include <pjsr/Slider.jsh>
#include <pjsr/NumericControl.jsh>
#include <pjsr/FontFamily.jsh>
#include <pjsr/FontStretch.jsh>

#include "DC_CombineImages_the_app.js"

///////////////////////////////////////////////////////////////////////////////

// This script maintains three primary objects: 'g_control_data', 'theApp',
// and 'dialog'.

// g_control_data, based on GlobalControlData, records basic 'wrapper'
// information such as 'script_title', 'script_version', and
// 'root_folder' (where the top-level' script resides).

// 'dialog', based on 'AppDialog()', provides the user interface
// and controls all interactions with 'theApp'.

// 'theApp', based on 'TheApplication()', defines data and methods
// directly related to the operation of the script. 'theApp'has no'knowledge' of
// 'dialog', except for a local reference to the dialog's status_bar.

///////////////////////////////////////////////////////////////////////////////
// Class: AppDialog()
///////////////////////////////////////////////////////////////////////////////

function AppDialog(theApp, root_folder)
{
    this.__base__ = Dialog;
    this.__base__();

    // DIALOG  MEMBER PROPERTIES -----------------------------------------------

    this.script_file_path                     = #__FILE__;
    this.root_folder                          = root_folder;

    this.theApp                               = theApp;

    this.b_abort_requested                    = false;

    this.windowTitle                          = g_control_data.script_title;

    this.b_onResize_pause                     = false;  // Not used at this time

    this.b_show_description                   = true;

    this.top_sizer_height_ratio               = TOP_SIZER_HEIGHT_RATIO;

    this.review_image_width_ratio             = REVIEW_IMAGE_WIDTH_RATIO;
    this.review_image_adjustments_width_ratio = REVIEW_IMAGE_ADJUSTMENTS_WIDTH_RATIO;
    this.description_width_ratio              = DESCRIPTION_WIDTH_RATIO;
    this.settings_width_ratio                 = SETTINGS_WIDTH_RATIO;

    this.description_folder                   = DESCRIPTION_FOLDER;

    // An array of color modes
    this.dialog_color_mode_array              = ["RGB", "HSV"];
    this.dialog_color_mode                    = "RGB";


    // UTILITY FUNCTIONS --------------------------------------------------------

    this.getIconBitmap = function(icon_name)
    {
        return new Bitmap( this.root_folder + "/System/Icons/" + icon_name );

    };

    //

    this.conditionalAbort = function()
    {
        if (this.b_abort_requested)
        {
            theApp.close();

            this.enabled = true;

            this.cancel();
        }
    };

    // DIALOG UPDATE FUNCTIONS -------------------------------------------------

    this.updateReviewImage = function()
    {
        let previous_background_color = this.review_image_control.backgroundColor;

        this.review_image_control.setBackgroundAlertColor(0xffff8000);

        theApp.recreateReviewImage();
        this.review_image_control.updateImage();
        this.review_image_control.updatePaintSettings();

        this.review_image_control.setBackgroundAlertColor(previous_background_color)
    };

    //

    this.updateColorizedImagePreview = function()
    {
        this.colorized_image_preview_control.updateImage();

        this.colorized_image_preview_control.updatePaintSettings();
    };

    //

    this.updateColorComboBox = function(row)
    {
        let red    = theApp.image_attributes_array[row].red;
        let green  = theApp.image_attributes_array[row].green;
        let blue   = theApp.image_attributes_array[row].blue;
        let alpha  = 1.0;
        let rowColor =  Color.rgbaColorF(red, green, blue, alpha);

        this.attributes_conrol_set_array[row].colorComboBox.setCurrentColor(rowColor);
        this.attributes_conrol_set_array[row].colorComboBox.backgroundColor = rowColor;

    };

    //

    // Sets dialog controls to values in the application's 'theApp.image_attributes_array'
    this.updateDialogControls_RGB_mode = function(row)
    {

        if (theApp.image_attributes_array[row].b_active)
        {
            this.attributes_conrol_set_array[row].view_list.currentView = theApp.view_array[row];
        }
        else
        {
            let current_invalid_view = this.attributes_conrol_set_array[row].view_list.currentView;
            this.attributes_conrol_set_array[row].view_list.remove(current_invalid_view);
            this.attributes_conrol_set_array[row].view_list.reload();
        }

        this.attributes_conrol_set_array[row].filterTypeComboBox.currentItem
               = this.attributes_conrol_set_array[row].filterTypeComboBox.findItem(theApp.image_attributes_array[row].filter, 0, true); // Filter

        this.attributes_conrol_set_array[row].amount_spinbox.value = theApp.image_attributes_array[row].amount * 100.0; // Amount spinbox

        let red   = theApp.image_attributes_array[row].red;   // Normalized floating point value
        let green = theApp.image_attributes_array[row].green; // Normalized floating point value
        let blue  = theApp.image_attributes_array[row].blue;  // Normalized floating point value

        this.attributes_conrol_set_array[row].red_spinbox.value    = red   * this.attributes_conrol_set_array[row].red_spinbox.maxValue;
        this.attributes_conrol_set_array[row].green_spinbox.value  = green * this.attributes_conrol_set_array[row].green_spinbox.maxValue;
        this.attributes_conrol_set_array[row].blue_spinbox.value   = blue  * this.attributes_conrol_set_array[row].blue_spinbox.maxValue;

        this.attributes_conrol_set_array[row].red_slider.value      = red   * this.attributes_conrol_set_array[row].red_slider.maxValue ;
        this.attributes_conrol_set_array[row].green_slider.value    = green * this.attributes_conrol_set_array[row].green_slider.maxValue;
        this.attributes_conrol_set_array[row].blue_slider.value     = blue  * this.attributes_conrol_set_array[row].blue_slider.maxValue;

        // 'Red' controls
        this.attributes_conrol_set_array[row].red_spinbox.backgroundColor    = 0xffc80000;     // Set Red spinbox color
        this.attributes_conrol_set_array[row].red_spinbox.textColor          = 0xffffffff;     // Set Red spinbox text color
        this.attributes_conrol_set_array[row].red_slider.backgroundColor     = 0xffc80000;     // Set Red slider color
        this.attributes_conrol_set_array[row].red_slider.textColor           = 0xffffffff;     // Set Red slider text color

        // 'Green' controls
        this.attributes_conrol_set_array[row].green_spinbox.backgroundColor  = 0xff00c800;     // Set Green spinbox color
        this.attributes_conrol_set_array[row].green_spinbox.textColor        = 0xffffffff;     // Set Green spinbox text color
        this.attributes_conrol_set_array[row].green_slider.backgroundColor   = 0xff00c800;     // Set Green slider color
        this.attributes_conrol_set_array[row].green_slider.textColor         = 0xffffffff;     // Set Green slider text color

        // 'Blue' controls
        this.attributes_conrol_set_array[row].blue_spinbox.backgroundColor   = 0xff0000c8;     // Set Blue spinbox color
        this.attributes_conrol_set_array[row].blue_spinbox.textColor         = 0xffffffff;     // Set Blue spinbox text color
        this.attributes_conrol_set_array[row].blue_slider.backgroundColor    = 0xff0000c8;     // Set Blue slider color
        this.attributes_conrol_set_array[row].blue_slider.textColor          = 0xffffffff;     // Set Blue slider text color

        // 'ComboBox' control

        let colorVal =  Color.rgbaColorF(red, green, blue, 1);
        this.attributes_conrol_set_array[row].colorComboBox.setCurrentColor(colorVal);
        this.attributes_conrol_set_array[row].colorComboBox.backgroundColor = colorVal;
     };

    //

    // Sets dialog controls to values in the application's 'theApp.image_attributes_array'
    this.updateDialogControls_HSV_mode = function(row)
    {

        if (theApp.image_attributes_array[row].b_active)
        {
            this.attributes_conrol_set_array[row].view_list.currentView = theApp.view_array[row];
        }
        else
        {
            let current_invalid_view = this.attributes_conrol_set_array[row].view_list.currentView;
            this.attributes_conrol_set_array[row].view_list.remove(current_invalid_view);
            this.attributes_conrol_set_array[row].view_list.reload();
        }

        this.attributes_conrol_set_array[row].filterTypeComboBox.currentItem
               = this.attributes_conrol_set_array[row].filterTypeComboBox.findItem(theApp.image_attributes_array[row].filter, 0, true); // Filter

        this.attributes_conrol_set_array[row].amount_spinbox.value = theApp.image_attributes_array[row].amount * 100.0; // Amount spinbox

        let red   = theApp.image_attributes_array[row].red;   // Normalized floating point value
        let green = theApp.image_attributes_array[row].green; // Normalized floating point value
        let blue  = theApp.image_attributes_array[row].blue;  // Normalized floating point value

        let hsv_color      = convertRGBtoHSV(red, green, blue);
        let hue            = hsv_color[0];                   // Normalized floating point value
        let hsvSaturation  = hsv_color[1];                   // Normalized floating point value
        let hsv_value      = hsv_color[2];                   // Normalized floating point value

        this.attributes_conrol_set_array[row].red_spinbox.value    = hue           * this.attributes_conrol_set_array[row].red_spinbox.maxValue;
        this.attributes_conrol_set_array[row].green_spinbox.value  = hsvSaturation * this.attributes_conrol_set_array[row].green_spinbox.maxValue;
        this.attributes_conrol_set_array[row].blue_spinbox.value   = hsv_value     * this.attributes_conrol_set_array[row].blue_spinbox.maxValue;

        this.attributes_conrol_set_array[row].red_slider.value      = hue           * this.attributes_conrol_set_array[row].red_slider.maxValue ;
        this.attributes_conrol_set_array[row].green_slider.value    = hsvSaturation * this.attributes_conrol_set_array[row].green_slider.maxValue;
        this.attributes_conrol_set_array[row].blue_slider.value     = hsv_value     * this.attributes_conrol_set_array[row].blue_slider.maxValue;

        let dialog_RGB       = convertHSVtoRGB(hue, 1.0, 1.0 );

        // 'Red' (== Hue) controls
        let dialog_color_val = Color.rgbaColorF(dialog_RGB[0], dialog_RGB[1], dialog_RGB[2], 1);
        this.attributes_conrol_set_array[row].red_spinbox.backgroundColor  = dialog_color_val;  // Set Red spinbox == Hue
        this.attributes_conrol_set_array[row].red_spinbox.textColor        = 0xff000000;        // Set Red slider text color
        this.attributes_conrol_set_array[row].red_slider.backgroundColor   = dialog_color_val;  // Set Red slider  == Hue
        this.attributes_conrol_set_array[row].red_slider.textColor         = 0xff000000;        // Set Red slider text color

        // 'Green' (== Saturation) controls
        dialog_color_val    = Color.rgbaColor(157, 204, 230, 255);  // Semi-arbitrary choice of hue. 'rgbaColor'  inputs and output are in 0-255 form
        let dialog_hue      = Color.hue(dialog_color_val);          // 'Color.hue' inputs are 0-255; returns 'dialog_hue' as a floating point value
        dialog_RGB          = convertHSVtoRGB(dialog_hue, hsvSaturation, 1); // 'convertHSVtoRGB'  inputs are array output are floating point values
        dialog_color_val    = Color.rgbaColorF(dialog_RGB[0], dialog_RGB[1], dialog_RGB[2], 1); // 'rgbaColorF' inputs are floating point values, but output values are in 0-255 form
        this.attributes_conrol_set_array[row].green_spinbox.backgroundColor  = dialog_color_val; // Set Green spinbox == Saturation
        this.attributes_conrol_set_array[row].green_spinbox.textColor        = 0xff000000;       // Set Green spinbox text color
        this.attributes_conrol_set_array[row].green_slider.backgroundColor   = dialog_color_val; // Set Green slider  == Saturation
        this.attributes_conrol_set_array[row].green_slider.textColor         = 0xff000000;       // Set Green slider text color

        // 'Blue' (== Value) controls
        dialog_color_val = Color.rgbaColorF(hsv_value, hsv_value, hsv_value, 1);           // Note: HSV 'Value' entered as RGB (gray) colors
        this.attributes_conrol_set_array[row].blue_spinbox.backgroundColor  = dialog_color_val;  // Set Blue spinbox == Value
        this.attributes_conrol_set_array[row].blue_slider.backgroundColor   = dialog_color_val;  // Set Blue slider  == Value

        // 'ComboBox' control (always RGB, even in HSV mode)

        let colorVal =  Color.rgbaColorF(red, green, blue, 1);
        this.attributes_conrol_set_array[row].colorComboBox.setCurrentColor(colorVal);
        this.attributes_conrol_set_array[row].colorComboBox.backgroundColor = colorVal;

        // Text color for 'Blue' and 'colorComboBox' controls
        if (hsv_value < 0.4 )
        {
            let dialog_color_val2 = Color.rgbaColorF(1-hsv_value, 1-hsv_value, 1-hsv_value, 1);
            this.attributes_conrol_set_array[row].blue_spinbox.textColor        = dialog_color_val2;
            this.attributes_conrol_set_array[row].colorComboBox.textColor       = dialog_color_val2;
        }
        else
        {
            this.attributes_conrol_set_array[row].blue_spinbox.textColor        = 0xff000000;
            this.attributes_conrol_set_array[row].colorComboBox.textColor       = 0xff000000;
        }

    };

    //

    this.updateDialogControls = function(row)
    {
        if (this.dialog_color_mode == "RGB")
        {
            this.updateDialogControls_RGB_mode(row);
        }
        else // this.dialog_color_mode == "HSV"
        {
            this.updateDialogControls_HSV_mode(row);
        }
    };

    //

    this.updateAllDialogControls = function()
    {
        for (let row = 0; row < NUM_DIALOG_ATTRIBUTES_CONTROL_SETS; row++)
        {
            this.updateDialogControls(row);
        }
    };

    // DIALOG EVENT HANDLERS ---------------------------------------------------

    this.onClose = function()
    {
        if (theApp.b_processes_running)
        {
            let message_box = new MessageBox(
                  "<p>Long running process in progress. Close?"+
                  "<p>Press 'YES'  to close afer the process finishes</p>"+
                  "<p>Press 'CANCEL' to contimue as before.</p>",
                  "CLOSE?",
                  StdIcon_Question,
                  StdButton_Yes, StdButton_Cancel);

            message_box.firstButton.text = "Wait";

            let result = message_box.execute();

            if (result == StdButton_Cancel)
            {
                return false;
            }
            else if (result == StdButton_Yes)
            {
                this.b_abort_requested = true;

                return false;
            }
        }

        theApp.close();
        this.dialog.enabled = true;
        return true;

    };

    //

    this.onSettingsChanged = function(b_settings_changed_since_combine = true,
                                      b_update_review_image_alert_color = true)
    {
        theApp.setSettingsChangedFlag(b_settings_changed_since_combine);

        if (b_update_review_image_alert_color)
        {
            let background_color = 0xff000000;

            if (b_settings_changed_since_combine)
            {
                background_color = 0xffff0000;
            }
            else
            {
                background_color = 0xffd0d0d0;
            }

            this.review_image_control.setBackgroundAlertColor(background_color);
        }
    };

    //

    this.onSaveSettings = function()
    {
        theApp.saveSettings();
    };

    //

    this.onLoadSettings = function()
    {
        (new MessageBox(
                  "<p>Both Combine settings and Image Review settings will be updated.</p><p>You will need to execute COMBINE before new Image Review settings will take effect.</p>",
                  "WARNING",
                  StdIcon_Warning,
                  StdButton_Ok
             )).execute();

        theApp.loadSettings();

        this.updateAllDialogControls(true);

        this.onPaletteChange(theApp.palette, false); // false == b_reset_image_attributes_array

        this.palette_auto_select_push_button.clicked = theApp.b_auto_set_palette_for_known_filter;

        if (theApp.b_auto_set_palette_for_known_filter)
        {
             this.palette_auto_select_push_button.backgroundColor = 0xff909090;
        }
        else
        {
            this.palette_auto_select_push_button.backgroundColor = 0xffd0d0d0;
        }

        this.colorize_method_combo_box.currentItem
                = this.colorize_method_combo_box.findItem( theApp.colorize_method, 0, true);

        this.combineMethodComboBox.currentItem
                = this.combineMethodComboBox.findItem( theApp.combine_method, 0, true);

        this.linear_fit_colorized_images_checkBox.checked = theApp.b_linear_fit_colorized_images_before_combine;

        this.rescale_method_combo_box.currentItem
                = this.rescale_method_combo_box.findItem( theApp.rescale_method, 0, true);

        this.rescale_channels_checkBox.checked = theApp.b_rescale_channels_independently;


        this.review_image_auto_stretch_checkbox.checked = theApp.b_auto_STF;

        this.review_image_linked_channels_checkbox.checked = theApp.b_linked_RGB_channels;

        this.review_image_saturation_numeric_control.setValue(theApp.review_image_saturation_boost);

        this.review_image_midtone_adjust_numeric_control.setValue(theApp.review_image_midtone_adjustment);

        this.review_image_shadows_clipping_adjust_numeric_control.setValue(theApp.review_image_shadows_clipping);

        this.onSettingsChanged(true, true); // true == settings changed. true == set alert color

    };

    //

    this.onViewChange = function(view, row)
    {
        // Null views are allowed here to trigger clearing of selection and/or colorized_image_preview
        theApp.addViewToRow(view, row);

        this.onSettingsChanged(true, true); // true == settings changed. true == set alert color

        let filter_name  = theApp.getFilterName(view,
                                                theApp.filter_names,
                                                theApp.filter_names_JWST,
                                                theApp.priority_JWST_filter_array );

        this.onFilterChange(row, filter_name); //  will call this.updateDialogControls(row)

        this.onRowSelectionChange(row, true);  // true == b_clear_selection
        this.updateColorizedImagePreview();

    };

    //

    this.onRowSelectionChange = function(row, b_clear_selection = false)
    {

        theApp.selectRow(row, b_clear_selection);

        for (let j = 0; j < NUM_DIALOG_ATTRIBUTES_CONTROL_SETS; j++)
        {
            if (j == theApp.selected_view_row)
            {
                this.attributes_conrol_set_array[j].row_select_push_button.backgroundColor = 0xff909090;
                this.attributes_conrol_set_array[j].row_select_push_button.icon = new Bitmap( ":/messages/check.png" );
            }
            else
            {
                this.attributes_conrol_set_array[j].row_select_push_button.backgroundColor = 0xffd0d0d0;
                this.attributes_conrol_set_array[j].row_select_push_button.icon = new Bitmap( ":/icons/control-button.png" );
            }

        }
        this.updateColorizedImagePreview();
        this.updateDialogControls(row);
    };

    //

    this.onFilterChange = function(row, filter_name)
    {
        // String case sensitivity approach.
        // Strings are used as keys in filter-palette database, so strings cases
        // are important.
        // In this function, strings cases remain unchanged (from that used by the system)
        // and tests are case-sensitive.

        // Change the filter
        theApp.image_attributes_array[row].filter = filter_name;

        this.onSettingsChanged(true, true); // true == settings changed. true == set alert color

        if (theApp.b_auto_set_palette_for_known_filter)
        {
            // Change Palette and adjust to new filter
            let JWST_index = theApp.filter_names_JWST.indexOf(filter_name);

            if (JWST_index >= 0 )
            {
                console.writeln("");
                console.writeln("JWST filter found. Settimg default JWST palette");
                console.writeln("");

                // No need to use resetImageAttributeArrayRow(row) here, will be called for ALL filter changes
                this.onPaletteChange("JWST", false); // false == resetImageAttributeArrayRow(row);
            }
            else
            {

                let SDSS_index = theApp.filter_names_SDSS.indexOf(filter_name);
                if (SDSS_index >= 0 )
                {
                    console.writeln("");
                    console.writeln("SDSS filter found. Settimg default SDSS palette");
                    console.writeln("");

                    // No need to use resetImageAttributeArrayRow(row) here, will be called for ALL filter changes
                    this.onPaletteChange("SDSS", false); // false == resetImageAttributeArrayRow(row);
                }
            }
        }

        theApp.resetImageAttributeArrayRow(row);
        this.updateDialogControls(row);

        this.onRowSelectionChange(row, true); // true == b_clear_selection
        this.updateColorizedImagePreview();
    };

    //

    this.onDialogColorModeChange = function(dialog_color_mode)
    {
        this.dialog_color_mode                      = dialog_color_mode;
        this.dialog_color_mode_push_button.text     = dialog_color_mode;

        if (dialog_color_mode.toUpperCase() == "RGB")
        {

            this.dialog_color_mode_push_button.backgroundColor = 0xffd0d0d0;
            this.dialog_color_mode_push_button.icon = new Bitmap( ":/toolbar/image-display-rgb.png"  );

            this.dialog_image_column_label_controls_array[8].text  =  "RED [0-255]";  // Red slider   ==  8
            this.dialog_image_column_label_controls_array[10].text = "GREEN [0-255]"; // Green slider == 10
            this.dialog_image_column_label_controls_array[12].text = "BLUE [0-255]";  // Red slider   == 12

            for (let row = 0; row < NUM_DIALOG_ATTRIBUTES_CONTROL_SETS; row++)
            {

                this.attributes_conrol_set_array[row].red_spinbox.minValue = 0;
                this.attributes_conrol_set_array[row].red_spinbox.maxValue = 255;
                this.attributes_conrol_set_array[row].red_spinbox.toolTip  = "<p>Red Value [0-255]</p>";

                this.attributes_conrol_set_array[row].green_spinbox.minValue = 0;
                this.attributes_conrol_set_array[row].green_spinbox.maxValue = 255;
                this.attributes_conrol_set_array[row].green_spinbox.toolTip  = "<p>Green Value [0-255]</p>";

                this.attributes_conrol_set_array[row].blue_spinbox.minValue = 0;
                this.attributes_conrol_set_array[row].blue_spinbox.maxValue = 255;
                this.attributes_conrol_set_array[row].blue_spinbox.toolTip  = "<p>Blue Value [0-255]</p>";

                this.attributes_conrol_set_array[row].red_slider.minValue = 0;
                this.attributes_conrol_set_array[row].red_slider.maxValue = 2550;
                this.attributes_conrol_set_array[row].red_slider.pageSize = 10;
                this.attributes_conrol_set_array[row].red_slider.toolTip = "<p>Red Value [0-255]</p>";

                this.attributes_conrol_set_array[row].green_slider.minValue = 0;
                this.attributes_conrol_set_array[row].green_slider.maxValue = 2550;
                this.attributes_conrol_set_array[row].green_slider.pageSize = 10;
                this.attributes_conrol_set_array[row].green_slider.toolTip  = "<p>Green Value [0-255]</p>";


                this.attributes_conrol_set_array[row].blue_slider.minValue = 0;
                this.attributes_conrol_set_array[row].blue_slider.maxValue = 2550;
                this.attributes_conrol_set_array[row].blue_slider.pageSize = 10;
                this.attributes_conrol_set_array[row].blue_slider.toolTip  = "<p>Blue Value [0-255]</p>";

            }
            this.updateAllDialogControls();
        }
        else // dialog_color_mode == "HSV"
        {
            this.dialog_color_mode_push_button.backgroundColor = 0xffd0d0d0;
            this.dialog_color_mode_push_button.icon = new Bitmap( ":/toolbar/image-display-saturation.png" );

            this.dialog_image_column_label_controls_array[8].text =  "HSV HUE [0-360]";         // Red slider   ==  8
            this.dialog_image_column_label_controls_array[10].text = "HSV SATURATION [0-100]";  // Green slider == 10
            this.dialog_image_column_label_controls_array[12].text = "HSV VALUE [0-100]";       // Blue slider  == 12

            for (let row = 0; row < NUM_DIALOG_ATTRIBUTES_CONTROL_SETS; row++)
            {

                this.attributes_conrol_set_array[row].red_spinbox.minValue = 0;
                this.attributes_conrol_set_array[row].red_spinbox.maxValue = 360;
                this.attributes_conrol_set_array[row].red_spinbox.toolTip  = "<p>HSV Hue [0-360]</p>";

                this.attributes_conrol_set_array[row].green_spinbox.minValue = 0;
                this.attributes_conrol_set_array[row].green_spinbox.maxValue = 100;
                this.attributes_conrol_set_array[row].green_spinbox.toolTip  = "<p>HSV Saturation [0-100]</p>";

                this.attributes_conrol_set_array[row].blue_spinbox.minValue = 0;
                this.attributes_conrol_set_array[row].blue_spinbox.maxValue = 100;
                this.attributes_conrol_set_array[row].blue_spinbox.toolTip  = "<p>HSV Value [0-100]</p>";

                this.attributes_conrol_set_array[row].red_slider.minValue = 0;
                this.attributes_conrol_set_array[row].red_slider.maxValue = 3600;
                this.attributes_conrol_set_array[row].red_slider.pageSize = 10;
                this.attributes_conrol_set_array[row].red_slider.toolTip  = "<p>HSV Hue  [0-360]</p>";

                this.attributes_conrol_set_array[row].green_slider.minValue = 0;
                this.attributes_conrol_set_array[row].green_slider.maxValue = 1000;
                this.attributes_conrol_set_array[row].green_slider.pageSize = 10;
                this.attributes_conrol_set_array[row].green_slider.toolTip  = "<p>HSV Saturation [0-100]</p>";

                this.attributes_conrol_set_array[row].blue_slider.minValue = 0;
                this.attributes_conrol_set_array[row].blue_slider.maxValue = 1000;
                this.attributes_conrol_set_array[row].blue_slider.pageSize = 10;
                this.attributes_conrol_set_array[row].blue_slider.toolTip  = "<p>HSV Value [0-100]</p>";
            }
            this.updateAllDialogControls();
        }
    };

    //

    this.onColorSpread = function()
    {
        theApp.setColorSpread();
        this.onSettingsChanged(true, true); // true == settings changed. true == set alert color
        this.updateAllDialogControls();
    };

    //

    this.onPaletteChange = function(palette, b_reset_image_attributes_array = false)
    {

        if (palette.toUpperCase() == "NONE")
        {
            this.NO_PALETTE_button.backgroundColor = 0xff909090;
            this.NATURAL_button.backgroundColor    = 0xffd0d0d0;
            this.JWST_button.backgroundColor       = 0xffd0d0d0;
            this.SDSS_button.backgroundColor       = 0xffd0d0d0;
            this.CFHT_button.backgroundColor       = 0xffd0d0d0;
            this.HST_button.backgroundColor        = 0xffd0d0d0;
            theApp.setPalette("NONE", b_reset_image_attributes_array );
        }
        if (palette.toUpperCase() == "HST")
        {
            this.NO_PALETTE_button.backgroundColor = 0xffd0d0d0;
            this.NATURAL_button.backgroundColor    = 0xffd0d0d0;
            this.JWST_button.backgroundColor       = 0xffd0d0d0;
            this.SDSS_button.backgroundColor       = 0xffd0d0d0;
            this.CFHT_button.backgroundColor       = 0xffd0d0d0;
            this.HST_button.backgroundColor        = 0xff909090;
            theApp.setPalette("HST", b_reset_image_attributes_array );
        }
        else if (palette.toUpperCase() == "CFHT")
        {
            this.NO_PALETTE_button.backgroundColor = 0xffd0d0d0;
            this.NATURAL_button.backgroundColor    = 0xffd0d0d0;
            this.JWST_button.backgroundColor       = 0xffd0d0d0;
            this.SDSS_button.backgroundColor       = 0xffd0d0d0;
            this.CFHT_button.backgroundColor       = 0xff909090;
            this.HST_button.backgroundColor        = 0xffd0d0d0;
            theApp.setPalette("CFHT", b_reset_image_attributes_array);
        }
        if (palette.toUpperCase() == "SDSS")
        {
            this.NO_PALETTE_button.backgroundColor = 0xffd0d0d0;
            this.NATURAL_button.backgroundColor    = 0xffd0d0d0;
            this.JWST_button.backgroundColor       = 0xffd0d0d0;
            this.SDSS_button.backgroundColor       = 0xff909090;
            this.CFHT_button.backgroundColor       = 0xffd0d0d0;
            this.HST_button.backgroundColor        = 0xffd0d0d0;
            theApp.setPalette("SDSS", b_reset_image_attributes_array);
        }
        if (palette.toUpperCase() == "JWST")
        {
            this.NO_PALETTE_button.backgroundColor = 0xffd0d0d0;
            this.NATURAL_button.backgroundColor    = 0xffd0d0d0;
            this.JWST_button.backgroundColor       = 0xff909090;
            this.SDSS_button.backgroundColor       = 0xffd0d0d0;
            this.CFHT_button.backgroundColor       = 0xffd0d0d0;
            this.HST_button.backgroundColor        = 0xffd0d0d0;
            theApp.setPalette("JWST", b_reset_image_attributes_array);
        }
        else if (palette.toUpperCase() == "NATURAL")
        {
            this.NO_PALETTE_button.backgroundColor = 0xffd0d0d0;
            this.NATURAL_button.backgroundColor    = 0xff909090;
            this.JWST_button.backgroundColor       = 0xffd0d0d0;
            this.SDSS_button.backgroundColor       = 0xffd0d0d0;
            this.CFHT_button.backgroundColor       = 0xffd0d0d0;
            this.HST_button.backgroundColor        = 0xffd0d0d0;
            theApp.setPalette("NATURAL", b_reset_image_attributes_array);
        }

        this.updateAllDialogControls();
    };

    //

    this.adjustTopSizerDimensions = function(dialog_width = -1, dialog_height = -1)
    {

            if (dialog_width == -1 || dialog_height == -1)
            {
                dialog_width  = this.width;
                dialog_height = this.height;
            }

            // Adjust sizes of Groupboxes in top_horizontal_sizer, based on proportions of Dialog size

            let top_sizer_height                             = this.top_sizer_height_ratio                  * dialog_height;

            this.review_image_groupbox.minWidth              = this.review_image_width_ratio                * dialog_width;
            this.review_image_groupbox.width                 = this.review_image_groupbox.minWidth;
            this.review_image_groupbox.minHeight             = top_sizer_height;
            this.review_image_groupbox.height                = top_sizer_height;

            this.review_image_control.minWidth               = this.review_image_groupbox.minWidth;
            this.review_image_control.width                  = this.review_image_control.minWidth;
            this.review_image_control.minHeight              = top_sizer_height;
            this.review_image_control.height                 = top_sizer_height;

            this.review_image_management_groupbox.minWidth   = this.review_image_adjustments_width_ratio    * dialog_width;
            this.review_image_management_groupbox.width      = this.review_image_management_groupbox.minWidth;
            this.review_image_management_groupbox.minHeight  = top_sizer_height;
            this.review_image_management_groupbox.height     = top_sizer_height;

            this.colorized_image_preview_control.minWidth             = this.review_image_management_groupbox.minWidth * 0.9;
            this.colorized_image_preview_control.width                = this.colorized_image_preview_control.minWidth;
            this.colorized_image_preview_control.minHeight            = this.colorized_image_preview_control.minWidth  * 0.9;
            this.colorized_image_preview_control.height               = this.colorized_image_preview_control.width; // Square

            // The order of the next two statements affect whether word wrap works (??)
            this.user_guide_textbox.minWidth                 = this.description_width_ratio                 * dialog_width;
            this.user_guide_textbox.width                    = this.user_guide_textbox.minWidth ;

            this.description_groupbox.minWidth               = this.description_width_ratio                 * dialog_width;
            this.description_groupbox.width                  = this.description_groupbox.minWidth;
            this.description_groupbox.minHeight              = top_sizer_height;
            this.description_groupbox.height                 = top_sizer_height;

            this.settings_groupbox.minWidth                  = this.settings_width_ratio                    * dialog_width;
            this.settings_groupbox.width                     = this.settings_groupbox.minWidth;
            this.settings_groupbox.minHeight                 = top_sizer_height;
            this.settings_groupbox.height                    = top_sizer_height;
    };

    //

    this.enableReviewImageManagementPanel_1 = function(b_enabled = true)
    {
        this.review_image_stretch_settings_label.enabled                    = b_enabled;
        this.review_image_auto_stretch_checkbox.enabled                     = b_enabled;
        this.review_image_linked_channels_checkbox.enabled                  = b_enabled;
        this.review_image_shadows_clipping_adjust_label.enabled             = b_enabled;
        this.review_image_shadows_clipping_adjust_numeric_control.enabled   = b_enabled;
        this.review_image_save_stretch_settings_push_button.enabled         = b_enabled;
        this.review_image_save_and_use_stretch_settings_push_button.enabled = b_enabled;
        this.review_image_use_saved_stretch_settings_push_button.enabled    = b_enabled;
        this.review_image_midtone_adjust_label.enabled                      = b_enabled;
        this.review_image_midtone_adjust_numeric_control.enabled            = b_enabled;
        this.review_image_saturation_label.enabled                          = b_enabled;
        this.review_image_saturation_numeric_control.enabled                = b_enabled;
        this.review_image_refresh_push_button.enabled                       = b_enabled;
    }

    //

    this.onResize = function(new_width = -1, new_height =-1, previous_width = -1, previous_height = -1)
    {
        if (!this.b_onResize_pause)
        {
            // Update the review_image_control (positioned  within review_image_groupbox)

            this.review_image_control.updatePaintSettings();
        }
    };

    //

    this.onExecute = function()
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>DIALOG STARTUP EXECUTION - dialog.onExecute()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        // DIALOG STARTUP EXECUTION

        this.adjustTopSizerDimensions();

        this.enableReviewImageManagementPanel_1(false);

        // Needed to update the review_image_control to display the splash image
        this.dialog.onSettingsChanged(false, true); // false == settings NOT changed. true == update alert color
        this.review_image_control.updateImage();
        this.review_image_control.updatePaintSettings();

        // Auto-populate image attribute rows with images auto-opened from Input Images folder

        if (theApp.b_auto_populate_row_entries)
        {
            let num_auto_opened_views = theApp.auto_opened_views_array.length;

            if (num_auto_opened_views > 0)
            {
                console.writeln("");
                console.writeln("--- AUTO POPULATING ROWS WITH IMAGES LOADED FROM INPUT FOLDER ---");
                console.writeln("");
            }
            console.writeln("");
            for (let i = 0; i < num_auto_opened_views && i < NUM_VIEW_SLOTS; i++)
            {

                console.writeln("");
                console.writeln("Auto-populating image rows : " + theApp.auto_opened_views_array[i].id);
                console.writeln("");

                this.dialog_status_message_label.update("Updating View: Executing", 0xffff0000);
                this.onViewChange(theApp.auto_opened_views_array[i], i);
                this.dialog_status_message_label.update("Updating View: Done", 0xff000000);

                let JWST_index = theApp.filter_names_JWST.indexOf(theApp.image_attributes_array[i].filter);

                if (JWST_index >= 0)
                {
                    console.writeln("");
                    console.writeln("JWST Filter Found.  Settimg Default JWST Palette");
                    console.writeln("");
                    this.onPaletteChange("JWST", true);
                }
            }
        }

        // Initialises a theApp member variable
        // This the only thing theApp 'knows' about the dialog object
        theApp.status_bar = this.app_status_message_label;

        this.dialog.onSettingsChanged(false, true); // false == settings NOT changed. true == update alert color

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: dialog.onExecute()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    };

    // DIALOG CONTROLS ---------------------------------------------------------

    let emWidth = this.font.width( 'M' );
    let labelWidth1 = this.font.width( "MMMMMMMMMMMM" ) + emWidth;

    //

    this.reviewImageControl = function(control_parent)
    {
        this.__base__ = Frame;
        this.__base__( control_parent );

        this.scaled_width    = 1000; // Initial value only
        this.scaled_height   = 1000; // Initial value only
        this.scaled_origin_x = 0;    // Initial value only
        this.scaled_origin_y = 0;    // Initial value only

        this.bitmap          = null;

        //

        this.updateImage = function()
        {
            if (theApp.review_image_view)
            {
                if (this.bitmap)
                {
                    this.bitmap.clear();  this.bitmap = null;  // Assist with garbage collection
                }
                this.bitmap  = theApp.review_image_view.image.render();
            }
        }

        //

        this.updatePaintSettings = function()
        {

            let image_aspect_ratio   = 1;
            let control_aspect_ratio = 1;

            if (this.bitmap)
            {
                let image_width  = theApp.review_image_view.image.width;
                let image_height = theApp.review_image_view.image.height;
                image_aspect_ratio     = image_width / image_height;
            }
            else
            {
               return;
            }

            if (control_parent != null)
            {
                control_aspect_ratio     = this.width / this.height;

                // Note: in this context, 'scaled' does NOT mean the same as PixInsight 'UI scaling'

                let image_scaled_width  = 1;
                let image_scaled_height = 1;
                let image_origin_x      = 0;
                let image_origin_y      = 0;

                if (image_aspect_ratio > control_aspect_ratio) // i.e. Image is 'more' landscape than the control
                {
                    image_scaled_width  = this.width;
                    image_scaled_height = image_scaled_width / image_aspect_ratio;

                    image_origin_x = 0;
                    image_origin_y = (this.height - image_scaled_height) / 2;

                }
                else // i.e. Image is 'more' portrait than the control
                {
                    image_scaled_height = this.height;
                    image_scaled_width  = image_scaled_height * image_aspect_ratio;

                    image_origin_x = (this.width - image_scaled_width)  / 2;
                    image_origin_y = 0;
                }

                this.scaled_width  = image_scaled_width;
                this.scaled_height = image_scaled_height;

                this.scaled_origin_x = (this.width  - image_scaled_width)  / 2;
                this.scaled_origin_y = (this.height - image_scaled_height) / 2;

            }

            this.update();

        };

        //

        this.onPaint = function()
        {
            if (theApp.review_image_view)
            {
                if (this.bitmap)
                {
                    let graphics   = new Graphics( this );

                    let scaled_bmp = this.bitmap.scaledTo( this.scaled_width, this.scaled_height);

                    graphics.drawBitmap( this.scaled_origin_x, this.scaled_origin_y,  scaled_bmp);
                    scaled_bmp.clear(); scaled_bmp = null; // Assist with garbage collection

                    graphics.end();
                }
            }
        };

        //

        this.onMouseDoubleClick = function(x, y, buttonState, modifiers )
        {
            this.dialog.enabled = false;
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color

            let new_center_global_normalized
                             = theApp.getWorkingImagesCenter(((x - this.scaled_origin_x) / this.scaled_width ) - 0.5,    // Normalized
                                                             ((y - this.scaled_origin_y) / this.scaled_height) - 0.5 );  // Normalized

            this.dialog.dialog_status_message_label.update("Recreate working images (Shrink "
                                                            + theApp.working_images_shrink_mode
                                                            + ") : Executing", 0xffff0000);
            theApp.setWorkingImagesShrinkMode(theApp.working_images_shrink_mode,
                                              new_center_global_normalized);
            this.dialog.dialog_status_message_label.update("Recreate working images : Done", 0xff000000);
            this.dialog.enabled = true;
        };

        //

        this.setBackgroundAlertColor = function(background_color = 0xff000000)
        {
            this.backgroundColor = background_color;
        }

    };

    //

    this.reviewImageControl.prototype = new Frame;

    //

    this.review_image_control = new this.reviewImageControl(this);
    this.review_image_control.frameStyle = FrameStyle_Raised;
    this.review_image_control.lineWidth = 5;
    this.review_image_control.backgroundColor = 0xffd0d0d0;

    //

    this.review_image_groupbox = new GroupBox (this);
    this.review_image_groupbox.title = "Review Image";
    this.review_image_groupbox.visible = true;
    this.review_image_groupbox.sizer = new VerticalSizer;
    this.review_image_groupbox.sizer.add(this.review_image_control);

    //

    // Review Image stretch controls

    this.review_image_stretch_settings_label = new Label( this );
    this.review_image_stretch_settings_label.useRichText = true;
    this.review_image_stretch_settings_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;
    this.review_image_stretch_settings_label.text = "<b>OUTPUT IMAGE</b>";
    this.review_image_stretch_settings_label.minWidth = 40;

    //

    this.review_image_stretch_settings_label2 = new Label( this );
    this.review_image_stretch_settings_label2.useRichText = true;
    this.review_image_stretch_settings_label2.textAlignment = TextAlign_Center| TextAlign_VertCenter;
    this.review_image_stretch_settings_label2.text = "(Display-only settings)";
    this.review_image_stretch_settings_label2.minWidth = 40;

    //

    this.review_image_auto_stretch_checkbox = new CheckBox( this );
    this.review_image_auto_stretch_checkbox.text = "Auto-stretch review image";
    this.review_image_auto_stretch_checkbox.checked = theApp.b_auto_STF;
    this.review_image_auto_stretch_checkbox.toolTip = "<p>Auto-stretch (only) review image</p>";
    this.review_image_auto_stretch_checkbox.onCheck = function( bChecked )
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Refresh Review : Executing", 0xffff0000);
        theApp.b_auto_STF = bChecked;
        if (bChecked)
        {
            this.dialog.review_image_midtone_adjust_numeric_control.slider.toolTip = "Midtone adjustment (relative to display auto stretch settings)";
        }
        else
        {
            this.dialog.review_image_midtone_adjust_numeric_control.slider.toolTip = "Midtone adjustment";
        }
        this.dialog.updateReviewImage();
        this.dialog.dialog_status_message_label.update("Refresh Review : Done", 0xff000000);
        this.dialog.enabled = true;
    };

    //

    this.review_image_linked_channels_checkbox = new CheckBox( this );
    this.review_image_linked_channels_checkbox.text = "Linked RGB channels";
    this.review_image_linked_channels_checkbox.checked = theApp.b_linked_RGB_channels;
    this.review_image_linked_channels_checkbox.toolTip = "<p>Linked RGB channels if 'Auto-stretch review image' is also checked", "</p>";
    this.review_image_linked_channels_checkbox.onCheck = function( bChecked )
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Refresh Review : Executing", 0xffff0000);
        theApp.b_linked_RGB_channels = bChecked;
        this.dialog.updateReviewImage();
        this.dialog.dialog_status_message_label.update("Refresh Review : Done", 0xff000000);
        this.dialog.enabled = true;
    };

    //

    this.review_image_shadows_clipping_adjust_label = new Label( this );
    this.review_image_shadows_clipping_adjust_label.text = "Shadows Clipping";
    this.review_image_shadows_clipping_adjust_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;
    this.review_image_shadows_clipping_adjust_label.minWidth = 40;

    //

    this.review_image_shadows_clipping_adjust_numeric_control = new NumericControl( this );
    this.review_image_shadows_clipping_adjust_numeric_control.real = true;
    this.review_image_shadows_clipping_adjust_numeric_control.precision = 2;
    this.review_image_shadows_clipping_adjust_numeric_control.setRange(-3, -0.01);
    this.review_image_shadows_clipping_adjust_numeric_control.slider.setRange(1, 300);
    this.review_image_shadows_clipping_adjust_numeric_control.slider.scaledMinWidth  = 50;
    this.review_image_shadows_clipping_adjust_numeric_control.slider.toolTip = "STF shadows clipping (sigma units from main histogram peak)";
    this.review_image_shadows_clipping_adjust_numeric_control.setValue(theApp.review_image_shadows_clipping);
    this.review_image_shadows_clipping_adjust_numeric_control.onValueUpdated = function( value )
    {
        theApp.review_image_shadows_clipping = value;
    };
    this.review_image_shadows_clipping_adjust_numeric_control.slider.onMouseDoubleClick = function(x, y, buttonState, modifiers )
    {
        this.dialog.review_image_shadows_clipping_adjust_numeric_control.setValue(SHADOWS_CLIP);
        theApp.review_image_shadows_clipping = SHADOWS_CLIP;
    };

    //

    this.review_image_save_stretch_settings_push_button= new PushButton( this);
    this.review_image_save_stretch_settings_push_button.text            = "Save";
    this.review_image_save_stretch_settings_push_button.toolTip         = "Save current display stretch settings";
    this.review_image_save_stretch_settings_push_button.backgroundColor = 0xffd0d0d0;
    this.review_image_save_stretch_settings_push_button.icon = new Bitmap( ":/icons/item-export.png" );
    this.review_image_save_stretch_settings_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Save current display stretch settings : Executing", 0xffff0000);
        theApp.saved_H   = g_helper_data.H;
        theApp.saved_STF = g_helper_data.STF;
        this.dialog.dialog_status_message_label.update("Save current display stretch settings : Done", 0xff000000);
        this.dialog.enabled = true;
    };

    //

    this.review_image_use_saved_stretch_settings_push_button= new PushButton( this);
    this.review_image_use_saved_stretch_settings_push_button.textColor       = 0xffff0000;
    this.review_image_use_saved_stretch_settings_push_button.text            = "Use?";
    this.review_image_use_saved_stretch_settings_push_button.toolTip         = "Use saved stretch settings";
    this.review_image_use_saved_stretch_settings_push_button.icon = new Bitmap( ":/icons/item-import.png" );
    this.review_image_use_saved_stretch_settings_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Selection: Use saved STF settings (" + theApp.b_use_saved_histogram + ") : Executing", 0xffff0000);
        if (!theApp.b_use_saved_histogram)
        {
            theApp.b_use_saved_histogram = true;
            this.dialog.review_image_auto_stretch_checkbox.enabled = false;
            this.dialog.review_image_linked_channels_checkbox.enabled = false;
            this.dialog.review_image_shadows_clipping_adjust_numeric_control.enabled  = false;
            this.dialog.review_image_use_saved_stretch_settings_push_button.textColor = 0xff228b22;
            this.dialog.review_image_use_saved_stretch_settings_push_button.text      = "Using. Stop?";
            this.dialog.review_image_save_and_use_stretch_settings_push_button.textColor = 0xff228b22;
            this.dialog.review_image_save_and_use_stretch_settings_push_button.text      = "Using. Stop?";
            this.dialog.updateReviewImage();
        }
        else
        {
            theApp.b_use_saved_histogram = false;
            this.dialog.review_image_auto_stretch_checkbox.enabled = true;
            this.dialog.review_image_linked_channels_checkbox.enabled = true;
            this.dialog.review_image_shadows_clipping_adjust_numeric_control.enabled     = true;
            this.dialog.review_image_use_saved_stretch_settings_push_button.textColor    = 0xffff0000;
            this.dialog.review_image_use_saved_stretch_settings_push_button.text         = "Use?";
            this.dialog.review_image_save_and_use_stretch_settings_push_button.textColor = 0xffff0000;
            this.dialog.review_image_save_and_use_stretch_settings_push_button.text      = "Save and use settings?";
            this.dialog.updateReviewImage();
        }
        this.dialog.dialog_status_message_label.update("Selection: Use/Do not use saved STF settings : Done", 0xff000000);
        this.dialog.enabled = true;
    }

    //

    this.review_image_save_and_use_stretch_settings_push_button= new PushButton( this);
    this.review_image_save_and_use_stretch_settings_push_button.textColor       = 0xffff0000;
    this.review_image_save_and_use_stretch_settings_push_button.text            = "Save and use settings?";
    this.review_image_save_and_use_stretch_settings_push_button.toolTip         = "Save and use current display stretch settings?";
    this.review_image_save_and_use_stretch_settings_push_button.backgroundColor = 0xffd0d0d0;
    this.review_image_save_and_use_stretch_settings_push_button.icon = new Bitmap( ":/icons/item.png");
    this.review_image_save_and_use_stretch_settings_push_button.onClick = function()
    {

        if (!theApp.b_use_saved_histogram)
        {
            this.dialog.enabled = false;
            this.dialog.dialog_status_message_label.update("Save and use current display stetch settings : Executing", 0xffff0000);
            theApp.saved_H   = g_helper_data.H;
            theApp.saved_STF = g_helper_data.STF;
            theApp.b_use_saved_histogram = true;
            this.dialog.review_image_auto_stretch_checkbox.enabled = false;
            this.dialog.review_image_linked_channels_checkbox.enabled = false;
            this.dialog.review_image_shadows_clipping_adjust_numeric_control.enabled  = false;
            this.dialog.review_image_save_and_use_stretch_settings_push_button.textColor = 0xff228b22;
            this.dialog.review_image_save_and_use_stretch_settings_push_button.text      = "Using. Stop?";
            this.dialog.review_image_use_saved_stretch_settings_push_button.textColor    = 0xff228b22;
            this.dialog.review_image_use_saved_stretch_settings_push_button.text         = "Using. Stop?";
            this.dialog.updateReviewImage();
            this.dialog.dialog_status_message_label.update("Save current display stetch settings : Done, now using", 0xff000000);
            this.dialog.enabled = true;
        }
        else
        {
            this.dialog.enabled = false;
            this.dialog.dialog_status_message_label.update("Save and use current display stetch settings : Executing", 0xffff0000);
            theApp.b_use_saved_histogram = false;
            this.dialog.review_image_auto_stretch_checkbox.enabled = true;
            this.dialog.review_image_linked_channels_checkbox.enabled = true;
            this.dialog.review_image_shadows_clipping_adjust_numeric_control.enabled     = true;
            this.dialog.review_image_save_and_use_stretch_settings_push_button.text      = "Save and use settings?";
            this.dialog.review_image_save_and_use_stretch_settings_push_button.textColor = 0xffff0000;
            this.dialog.review_image_use_saved_stretch_settings_push_button.textColor    = 0xffff0000;
            this.dialog.review_image_use_saved_stretch_settings_push_button.text         = "Use?";
            this.dialog.updateReviewImage();
            this.dialog.dialog_status_message_label.update("Save current display stetch settings : Done, not using", 0xff000000);
            this.dialog.enabled = true;
        }



    }

    //

    this.review_image_save_use_horizontal_sizer = new HorizontalSizer;
    this.review_image_save_use_horizontal_sizer.add( this.review_image_save_stretch_settings_push_button);
    this.review_image_save_use_horizontal_sizer.add( this.review_image_use_saved_stretch_settings_push_button);

    //

    this.review_image_midtone_adjust_label = new Label( this );
    this.review_image_midtone_adjust_label.useRichText = true;
    this.review_image_midtone_adjust_label.text = "<b>Midtone Adjustment</b>";
    this.review_image_midtone_adjust_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;
    this.review_image_midtone_adjust_label.minWidth = 40;

    //

    this.review_image_midtone_adjust_numeric_control = new NumericControl( this );
    this.review_image_midtone_adjust_numeric_control.real = true;
    this.review_image_midtone_adjust_numeric_control.precision = 8;
    this.review_image_midtone_adjust_numeric_control.setRange(0.00000001, 0.99999999);
    this.review_image_midtone_adjust_numeric_control.slider.setRange(1, 99999999);
    this.review_image_midtone_adjust_numeric_control.slider.scaledMinWidth  = 50;
    // Change slider toolTip depending on the state of theApp.b_auto_STF)
    if (theApp.b_auto_STF)
    {
         this.review_image_midtone_adjust_numeric_control.slider.toolTip = "Midtone adjustment (relative to display stretch settings)";
    }
    else
    {
         this.review_image_midtone_adjust_numeric_control.slider.toolTip = "Midtone adjustment";
    }
    this.review_image_midtone_adjust_numeric_control.setValue(0.50);
    this.review_image_midtone_adjust_numeric_control.onValueUpdated = function( value )
    {
        theApp.review_image_midtone_adjustment = value;
    };
    this.review_image_midtone_adjust_numeric_control.slider.onMouseDoubleClick = function(x, y, buttonState, modifiers )
    {
        this.dialog.review_image_midtone_adjust_numeric_control.setValue(0.50);
        theApp.review_image_midtone_adjustment = 0.5;
    };

    //

    this.review_image_saturation_label = new Label( this );
    this.review_image_saturation_label.useRichText = true;
    this.review_image_saturation_label.text = "<b>Saturation Adjustment</b>";
    this.review_image_saturation_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;
    this.review_image_saturation_label.minWidth = 40;

    //

    this.review_image_saturation_numeric_control = new NumericControl( this );
    this.review_image_saturation_numeric_control.real = true;
    this.review_image_saturation_numeric_control.precision = 2;
    this.review_image_saturation_numeric_control.setRange(-10.00, 10.00);
    this.review_image_saturation_numeric_control.slider.setRange(0, 1000);
    this.review_image_saturation_numeric_control.slider.scaledMinWidth  = 50;
    this.review_image_saturation_numeric_control.slider.toolTip  = "Saturation adjustment [-10 to +10]";
    this.review_image_saturation_numeric_control.setValue(theApp.review_image_saturation_boost );
    this.review_image_saturation_numeric_control.onValueUpdated = function( value )
    {
        theApp.review_image_saturation_boost = value;
    };
    this.review_image_saturation_numeric_control.slider.onMouseDoubleClick = function(x, y, buttonState, modifiers )
    {
        this.dialog.review_image_saturation_numeric_control.setValue(0.0);
        theApp.review_image_saturation_boost = 0.0;
    };

    //

    this.review_image_refresh_push_button= new PushButton( this);
    this.review_image_refresh_push_button.toolTip         = "Refresh Review Image";
    this.review_image_refresh_push_button.text            = "Refresh Review Image";
    this.review_image_refresh_push_button.backgroundColor = 0xffd0d0d0;
    this.review_image_refresh_push_button.foregroundColor = 0xffff0000;
    this.review_image_refresh_push_button.icon = new Bitmap( ":/icons/refresh.png" );
    this.review_image_refresh_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Refresh Review Image : Executing", 0xffff0000);
        this.dialog.updateReviewImage();
        this.dialog.dialog_status_message_label.update("Refresh Review Image : Done", 0xff000000);
        this.dialog.enabled = true;
        this.dialog.conditionalAbort();
    };

    //

    this.colorizedImagePreviewControl =  function(control_parent)
    {
        this.__base__ = Frame;
        this.__base__( control_parent );

        this.scaled_width          = 1000; // Initial value only
        this.scaled_height         = 1000; // Initial value only
        this.scaled_origin_x       = 0;    // Initial value only
        this.scaled_origin_y       = 0;    // Initial value only

        this.bitmap          = null;

        //

        this.updateImage = function()
        {

            if (theApp.colorized_image_preview.view)
            {
                if (this.bitmap)
                {
                    this.bitmap.clear();  this.bitmap = null;  // Assist with garbage collection
                }
                this.bitmap  = theApp.colorized_image_preview.view.image.render() ;
             }
        }

        //

        this.updatePaintSettings = function()
        {

            let image_aspect_ratio   = 1;
            let control_aspect_ratio = 1;

            if (theApp.review_image_view)
            {
                let image_width  = theApp.review_image_view.image.width;
                let image_height = theApp.review_image_view.image.height;
                image_aspect_ratio     = image_width / image_height;
            }
            else
            {
               return;
            }

            if (control_parent != null)
            {

                control_aspect_ratio     = this.width / this.height;

                // Note: in this context, 'scaled' does NOT mean the same as PixInsight 'UI scaling'

                let image_scaled_width  = 1;
                let image_scaled_height = 1;
                let image_origin_x      = 0;
                let image_origin_y      = 0;

                if (image_aspect_ratio > control_aspect_ratio) // i.e. Image is 'more' landscape than the control
                {
                    image_scaled_width  = this.width;
                    image_scaled_height = image_scaled_width / image_aspect_ratio;

                    image_origin_x = 0;
                    image_origin_y = (this.height - image_scaled_height) / 2;

                }
                else // i.e. Image is 'more' portrait than the control
                {
                    image_scaled_height = this.height;
                    image_scaled_width  = image_scaled_height * image_aspect_ratio;

                    image_origin_x = (this.width - image_scaled_width)  / 2;
                    image_origin_y = 0;

                }

                this.scaled_width  = image_scaled_width;
                this.scaled_height = image_scaled_height;

                this.scaled_origin_x = (this.width  - image_scaled_width)  / 2;
                this.scaled_origin_y = (this.height - image_scaled_height) / 2;

            }

            this.update();
        };


        //

        this.onPaint = function()
        {
            if (theApp.review_image_view)
            {
                if (this.bitmap)
                {
                    let graphics   = new Graphics( this );

                    let scaled_bmp = this.bitmap.scaledTo( this.scaled_width, this.scaled_height);

                    graphics.drawBitmap( this.scaled_origin_x, this.scaled_origin_y,  scaled_bmp);
                    scaled_bmp.clear(); scaled_bmp = null; // Assist with garbage collection

                    graphics.end();
                }
            }
        };

    };

    //

    this.colorizedImagePreviewControl.prototype = new Frame;

    //

    this.colorized_image_preview_label = new Label( this );
    this.colorized_image_preview_label.useRichText = true;
    this.colorized_image_preview_label.text = "<b>COLORIZED INPUT IMAGE</b>";
    this.colorized_image_preview_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;
    this.colorized_image_preview_label.minWidth = 40;

    //

    this.colorized_image_preview_label2 = new Label( this );
    this.colorized_image_preview_label2.useRichText = true;
    this.colorized_image_preview_label2.text = "<b>Preview</b>";
    this.colorized_image_preview_label2.textAlignment = TextAlign_Center| TextAlign_VertCenter;
    this.colorized_image_preview_label2.minWidth = 40;

    //

    this.colorized_image_preview_control = new this.colorizedImagePreviewControl(this);
    this.colorized_image_preview_control.frameStyle = FrameStyle_Raised;
    this.colorized_image_preview_control.lineWidth = 5;
    this.colorized_image_preview_control.backgroundColor = 0xffd0d0d0;
    this.colorized_image_preview_control.toolTip = "Select a checkbox on the left of a specific grayscale image to preview colorized version. Deselect to cancel.";

    //

    this.colorized_input_preview_midtone_adjust_label = new Label( this );
    this.colorized_input_preview_midtone_adjust_label.useRichText = true;
    this.colorized_input_preview_midtone_adjust_label.text = "<b>Colorized Input Midtone</b>";
    this.colorized_input_preview_midtone_adjust_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;
    this.colorized_input_preview_midtone_adjust_label.minWidth = 40;

    //

    this.colorized_image_preview_midtone_adjust_numeric_control = new NumericControl( this );
    this.colorized_image_preview_midtone_adjust_numeric_control.real = true;
    this.colorized_image_preview_midtone_adjust_numeric_control.precision = 8;
    this.colorized_image_preview_midtone_adjust_numeric_control.setRange(0.00000001, 0.99999999);
    this.colorized_image_preview_midtone_adjust_numeric_control.slider.setRange(1, 99999999);
    this.colorized_image_preview_midtone_adjust_numeric_control.slider.scaledMinWidth  = 50;
    this.colorized_image_preview_midtone_adjust_numeric_control.slider.toolTip = "Colorized input image Midtone adjustment (relative to display stretch settings)";
    this.colorized_image_preview_midtone_adjust_numeric_control.setValue(0.50);
    this.colorized_image_preview_midtone_adjust_numeric_control.onValueUpdated = function( value )
    {
        theApp.colorized_image_preview_midtone_adjustment = value;
    };
    this.colorized_image_preview_midtone_adjust_numeric_control.slider.onMouseDoubleClick = function(x, y, buttonState, modifiers )
    {
        this.dialog.colorized_image_preview_midtone_adjust_numeric_control.setValue(0.50);
        theApp.colorized_image_preview_midtone_adjustment = 0.5;
    };

    //

    this.review_image_management_groupbox = new GroupBox (this);
    this.review_image_management_groupbox.title = "Review";
    this.review_image_management_groupbox.minWidth  = 210;
    this.review_image_management_groupbox.sizer = new VerticalSizer;
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.review_image_stretch_settings_label);
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.review_image_stretch_settings_label2);
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.review_image_auto_stretch_checkbox);
    this.review_image_management_groupbox.sizer.add( this.review_image_linked_channels_checkbox);
    this.review_image_management_groupbox.sizer.addSpacing(6);
    this.review_image_management_groupbox.sizer.add( this.review_image_shadows_clipping_adjust_label);
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.review_image_shadows_clipping_adjust_numeric_control);
    this.review_image_management_groupbox.sizer.addSpacing(6);
    this.review_image_management_groupbox.sizer.add( this.review_image_save_use_horizontal_sizer);
    this.review_image_management_groupbox.sizer.addSpacing(6);
    this.review_image_management_groupbox.sizer.add( this.review_image_save_and_use_stretch_settings_push_button);
    this.review_image_management_groupbox.sizer.addSpacing(6);
    this.review_image_management_groupbox.sizer.add( this.review_image_midtone_adjust_label);
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.review_image_midtone_adjust_numeric_control);
    this.review_image_management_groupbox.sizer.addSpacing(6);
    this.review_image_management_groupbox.sizer.add( this.review_image_saturation_label);
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.review_image_saturation_numeric_control);
    this.review_image_management_groupbox.sizer.addSpacing(6);
    this.review_image_management_groupbox.sizer.add( this.review_image_refresh_push_button);
    this.review_image_management_groupbox.sizer.addSpacing(10);
    this.review_image_management_groupbox.sizer.add( this.colorized_image_preview_label);
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.colorized_image_preview_label2);
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.colorized_image_preview_control);
    this.review_image_management_groupbox.sizer.addSpacing(6);
    this.review_image_management_groupbox.sizer.add( this.colorized_input_preview_midtone_adjust_label);
    this.review_image_management_groupbox.sizer.addSpacing(3);
    this.review_image_management_groupbox.sizer.add( this.colorized_image_preview_midtone_adjust_numeric_control);
    this.review_image_management_groupbox.sizer.addSpacing(6);
    this.review_image_management_groupbox.sizer.addStretch();

    //

    // Description/User Guide label control

    this.description_label               = new Label( this );
    this.description_label.frameStyle    = FrameStyle_Flat;
    this.description_label.margin        = 4;
    this.description_label.textAlignment = TextAlign_Left| TextAlign_VertCenter;
    this.description_label.minWidth      = 100;
    this.description_label.useRichText   = true;
    this.description_label.wordWrapping  = true;
    this.description_label.font          = new Font("Arial", 9) ;
    this.description_label.text          = "<b>" + g_control_data.script_title.toUpperCase() + " v" + g_control_data.script_version + "</b>"
                                                 + readTextFile("DESCRIPTION.txt",
                                                                this.dialog.root_folder + "/" + this.description_folder);
    //

    this.user_guide_textbox               = new TextBox( this );
    this.user_guide_textbox.frameStyle    = FrameStyle_Flat;
    this.user_guide_textbox.readOnly      = true;
    this.user_guide_textbox.margin        = 4;
    this.user_guide_textbox.minWidth      = 800;
    this.user_guide_textbox.wordWrapping  = true;
    this.user_guide_textbox.useRichText   = true;
    this.user_guide_textbox.font          = new Font("Arial", 9) ;
    this.user_guide_textbox.text          = readTextFile("USER GUIDE.txt",
                                                         this.dialog.root_folder  + "/" + this.description_folder);
    this.user_guide_textbox.home();

    this.user_guide_groupbox = new GroupBox (this);
    this.user_guide_groupbox.title = "USER GUIDE";
    this.user_guide_groupbox.minStretchedWidth =  800;
    this.user_guide_groupbox.sizer = new VerticalSizer;
    this.user_guide_groupbox.sizer.add( this.user_guide_textbox);

    //

    this.description_groupbox = new GroupBox (this);
    this.description_groupbox.title = "";
    this.description_groupbox.minStretchedWidth =  800;
    this.description_groupbox.sizer = new VerticalSizer;
    this.description_groupbox.sizer.add( this.description_label);
    this.description_groupbox.sizer.addSpacing(6);
    this.description_groupbox.sizer.add( this.user_guide_groupbox );

    //

    // Settings controls

    this.show_description_push_button= new PushButton( this);
    this.show_description_push_button.toolTip         = "Show/hide description";
    this.show_description_push_button.text            = "Hide Description";
    this.show_description_push_button.backgroundColor = 0xffd0d0d0;
    this.show_description_push_button.foregroundColor = 0xff000000;
    this.show_description_push_button.icon = new Bitmap( ":/auto-hide/floating.png" );
    this.show_description_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        if ( this.dialog.b_show_description == true)
        {
            this.dialog.b_show_description = false;
            this.text = "Show Description";
            this.dialog.description_label.hide();
        }
        else
        {
            this.dialog.b_show_description = true;
            this.text = "Hide Description";
            this.dialog.description_label.show();
        }
        this.dialog.description_label.ensureLayoutUpdated();

        this.dialog.enabled = true;
    };

    //

    this.settings_label = new Label( this );
    this.settings_label.useRichText = true;
    this.settings_label.text = "<b>Settings</b>";
    this.settings_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;

    //

    this.save_settings_push_button = new PushButton( this);
    this.save_settings_push_button.toolTip         = "Save settings";
    this.save_settings_push_button.text            = "Save";
    this.save_settings_push_button.minWidth        = 25;
    this.save_settings_push_button.backgroundColor = 0xffd0d0d0;
    this.save_settings_push_button.foregroundColor = 0xff0000000;
    this.save_settings_push_button.icon = new Bitmap( ":/toolbar/process-favorites-save.png" );
    this.save_settings_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Save settings : Executing", 0xffff0000);
        this.dialog.onSaveSettings();
        this.dialog.dialog_status_message_label.update("Save settings : Done", 0xff000000);
        this.dialog.enabled = true;
    };

    //

    this.load_settings_push_button = new PushButton( this);
    this.load_settings_push_button.toolTip         = "Load settings";
    this.load_settings_push_button.text            = "Load";
    this.load_settings_push_button.minWidth        = 25;
    this.load_settings_push_button.backgroundColor = 0xffd0d0d0;
    this.load_settings_push_button.foregroundColor = 0xff000000;
    this.load_settings_push_button.icon = new Bitmap( ":/icons/settings.png" );
    this.load_settings_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Load settings : Executing", 0xffff0000);
        this.dialog.onLoadSettings();
        this.dialog.dialog_status_message_label.update("Load settings : Done", 0xff000000);
        this.dialog.enabled = true;
    };

    //

    this.load_save_settings_sizer = new HorizontalSizer;
    this.load_save_settings_sizer.add(this.save_settings_push_button);
    this.load_save_settings_sizer.addSpacing(5);
    this.load_save_settings_sizer.add(this.load_settings_push_button);

    //

    this.default_palettes_label = new Label( this );
    this.default_palettes_label.useRichText = true;
    this.default_palettes_label.text = "<b>Default palettes</b>";
    this.default_palettes_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;

    //

    this.update_default_palettes_push_button = new PushButton( this);
    this.update_default_palettes_push_button.toolTip         = "Update Palettes";
    this.update_default_palettes_push_button.text            = "Update";
    this.update_default_palettes_push_button.minWidth        = 25;
    this.update_default_palettes_push_button.backgroundColor = 0xffd0d0d0;
    this.update_default_palettes_push_button.foregroundColor = 0xff0000000;
    this.update_default_palettes_push_button.icon = new Bitmap( ":/icons/documents-import.png" );
    this.update_default_palettes_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Update default palettes : Executing", 0xffff0000);
        theApp.updateDefaultPalettes();
        this.dialog.dialog_status_message_label.update("Update default palettes : Done", 0xff000000);
        this.dialog.enabled = true;
    };

    //

    this.reset_default_palettes_push_button = new PushButton( this);
    this.reset_default_palettes_push_button.toolTip         = "Reset to system default palettes";
    this.reset_default_palettes_push_button.text            = "Reset";
    this.reset_default_palettes_push_button.minWidth        = 25;
    this.reset_default_palettes_push_button.backgroundColor = 0xffd0d0d0;
    this.reset_default_palettes_push_button.foregroundColor = 0xff000000;
    this.reset_default_palettes_push_button.icon = new Bitmap( ":/icons/reload.png" );
    this.reset_default_palettes_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Restore default palettes : Executing", 0xffff0000);
        theApp.resetDefaultSystemPalettes();
        this.dialog.dialog_status_message_label.update("Restore default palettes : Done", 0xff000000);
        this.dialog.enabled = true;
    };

    this.update_restore_default_palettes_sizer = new HorizontalSizer;
    this.update_restore_default_palettes_sizer.add(this.update_default_palettes_push_button);
    this.update_restore_default_palettes_sizer.addSpacing(5);
    this.update_restore_default_palettes_sizer.add(this.reset_default_palettes_push_button);

    //

    this.save_diagnostics_label = new Label( this );
    this.save_diagnostics_label.useRichText = true;
    this.save_diagnostics_label.text = "<b>Diagnostics</b>";
    this.save_diagnostics_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;

    //

    this.save_diagnostics_push_button = new PushButton( this);
    this.save_diagnostics_push_button.toolTip  = "Save  ?";
    if (theApp.b_save_diagnostics == true)
    {
        this.save_diagnostics_push_button.text            = "Save diagnostics";
        this.save_diagnostics_push_button.backgroundColor = 0xffd0d0d0;
        this.save_diagnostics_push_button.textColor       = 0xff228b22;
    }
    else
    {
        this.save_diagnostics_push_button.text            = "Do not save diagnostics";
        this.save_diagnostics_push_button.backgroundColor = 0xffd0d0d0;
        this.save_diagnostics_push_button.textColor       = 0xffff0000;
    }
    this.save_diagnostics_push_button.icon = new Bitmap( ":/icons/document-save.png" );
    this.save_diagnostics_push_button.onClick = function()
    {
        this.dialog.enabled = false;

        if (theApp.b_save_diagnostics == true)
        {
            theApp.b_save_diagnostics = false;
            this.dialog.save_diagnostics_push_button.text            = "Do not save diagnostics";
            this.dialog.save_diagnostics_push_button.textColor       = 0xffff0000;
        }
        else
        {
            theApp.b_save_diagnostics = true;
            this.dialog.save_diagnostics_push_button.text            = "Save diagnostics";
            this.dialog.save_diagnostics_push_button.textColor       = 0xff228b22;
        }

        this.dialog.enabled = true;
    };

    //

    this.colorize_method_label = new Label( this );
    this.colorize_method_label.useRichText = true;
    this.colorize_method_label.text = "<b>Colorize Method</b>";
    this.colorize_method_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;

    //

    this.colorize_method_combo_box = new ComboBox( this);
    for (let i = 0; i < theApp.colorize_methods_array.length; i++ )
    {
        this.colorize_method_combo_box.addItem( theApp.colorize_methods_array[i] );
    }
    this.colorize_method_combo_box.currentItem
        = this.colorize_method_combo_box.findItem( theApp.colorize_method, 0, true);
    this.colorize_method_combo_box.toolTip = "Colorize method  ";

    this.colorize_method_combo_box.currentItem.font = new Font("Arial", 9) ;
    this.colorize_method_combo_box.onItemSelected = function( index )
    {
        this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
        theApp.colorize_method = this.dialog.colorize_method_combo_box.itemText(index);
    };

    //

    this.combine_method_label = new Label( this );
    this.combine_method_label.useRichText = true;
    this.combine_method_label.text = "<b>Combine Method</b>";
    this.combine_method_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;

    //

    this.combineMethodComboBox = new ComboBox( this);
    for ( let i = 0; i < theApp.combine_methods.length; i++ )
    {
        this.combineMethodComboBox.addItem( theApp.combine_methods[i] );
    }
    this.combineMethodComboBox.currentItem
        = this.combineMethodComboBox.findItem(theApp.combine_method, 0, true);
    this.combineMethodComboBox.toolTip = "Image combination method";
    this.combineMethodComboBox.onItemSelected = function( index )
    {
        this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
        theApp.combine_method = this.dialog.combineMethodComboBox.itemText(index);
    };

    //

    this.linear_fit_colorized_images_checkBox = new CheckBox( this );
    this.linear_fit_colorized_images_checkBox.text = "Linear Fit colorized images?";
    this.linear_fit_colorized_images_checkBox.checked = theApp.b_linear_fit_colorized_images_before_combine;
    this.linear_fit_colorized_images_checkBox.toolTip = "<p>Linear Fit colorized images</p>";
    this.linear_fit_colorized_images_checkBox.onCheck = function( bChecked )
    {
        this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
        theApp.b_linear_fit_colorized_images_before_combine = bChecked
    };

    //

    this.rescale_label = new Label( this );
    this.rescale_label.useRichText = true;
    this.rescale_label.text = "<b>Rescale Method</b>";
    this.rescale_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;

    //

    this.rescale_method_combo_box = new ComboBox( this);
    for ( let i = 0; i < theApp.rescale_methods.length; i++ )
    {
        this.rescale_method_combo_box.addItem( theApp.rescale_methods[i] );
    }
    this.rescale_method_combo_box.currentItem
        = this.rescale_method_combo_box.findItem(theApp.rescale_method, 0, true);
     this.rescale_method_combo_box.currentItem.font = new Font("Arial", 9) ;
    this.rescale_method_combo_box.toolTip = "Image rescale method  ";
    this.rescale_method_combo_box.onItemSelected = function( index )
    {
        this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
        theApp.rescale_method = this.dialog.rescale_method_combo_box.itemText(index);
    };

    //

    this.rescale_channels_checkBox = new CheckBox( this );
    this.rescale_channels_checkBox.text = "Channels independently?";
    this.rescale_channels_checkBox.checked = theApp.b_rescale_channels_independently;
    this.rescale_channels_checkBox.toolTip = "<p>Rescale each channel independently?</p>";
    this.rescale_channels_checkBox.onCheck = function( bChecked )
    {
        this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
        theApp.b_rescale_channels_independently = bChecked
    };

    //

    this.save_colorized_input_images_label = new Label( this );
    this.save_colorized_input_images_label.useRichText = true;
    this.save_colorized_input_images_label.text = "<b>Save colorized images?</b>";
    this.save_colorized_input_images_label.textAlignment = TextAlign_Center| TextAlign_VertCenter;

    //

    this.save_colorized_input_images_push_button = new PushButton( this);
    this.save_colorized_input_images_push_button.toolTip             = "Save colorized images?";
    if (theApp.b_save_colorized_images == true)
    {
        this.save_colorized_input_images_push_button.text            = "Save images";
        this.save_colorized_input_images_push_button.backgroundColor = 0xffd0d0d0;
        this.save_colorized_input_images_push_button.textColor       = 0xff228b22;
    }
    else
    {
        this.save_colorized_input_images_push_button.text            = "Do not save images";
        this.save_colorized_input_images_push_button.backgroundColor = 0xffd0d0d0;
        this.save_colorized_input_images_push_button.textColor       = 0xffff0000;
    }
    this.save_colorized_input_images_push_button.icon = new Bitmap( ":/icons/document-save.png" );
    this.save_colorized_input_images_push_button.onClick = function()
    {
        this.dialog.enabled = false;

        if (theApp.b_save_colorized_images == true)
        {
            theApp.b_save_colorized_images = false;
            this.dialog.save_colorized_input_images_push_button.text      = "Do not save images";
            this.dialog.save_colorized_input_images_push_button.textColor = 0xffff0000;
        }
        else
        {
            theApp.b_save_colorized_images = true;
            this.dialog.save_colorized_input_images_push_button.text      = "Save images";
            this.dialog.save_colorized_input_images_push_button.textColor = 0xff228b22;
        }

        this.dialog.enabled = true;
    };

    //

    this.colorized_image_file_extension_combobox = new ComboBox( this);
    for ( let i = 0; i < theApp.file_extensions_array.length; i++ )
    {
        this.colorized_image_file_extension_combobox.addItem( theApp.file_extensions_array[i] );
    }
    this.colorized_image_file_extension_combobox.currentItem
        = this.colorized_image_file_extension_combobox.findItem(theApp.colorized_image_file_extension, 0, true);
    this.colorized_image_file_extension_combobox.toolTip = "Colorized image file extension";
    this.colorized_image_file_extension_combobox.onItemSelected = function( index )
    {
        theApp.colorized_image_file_extension = this.dialog.colorized_image_file_extension_combobox.itemText(index);
    };

    //

    this.colorized_image_autostretch_combobox = new ComboBox( this);
    let colorized_input_image_autostretch_options_array = ["No Autostretch", "With Autostretch"];
    for ( let i = 0; i < colorized_input_image_autostretch_options_array.length; i++ )
    {
        this.colorized_image_autostretch_combobox.addItem(colorized_input_image_autostretch_options_array[i] );
    }
    if (theApp.b_save_colorized_images_with_autostretch)
    {
        this.colorized_image_autostretch_combobox.currentItem = 1;
        this.colorized_image_autostretch_combobox.textColor = 0xff228b22;
    }
    else
    {
        this.colorized_image_autostretch_combobox.currentItem = 0;
        this.colorized_image_autostretch_combobox.textColor = 0xffff0000;
    }

    this.colorized_image_autostretch_combobox.onItemSelected = function( index )
    {
        if (index == 0)
        {
            theApp.b_save_colorized_images_with_autostretch = false;
            this.dialog.colorized_image_autostretch_combobox.textColor = 0xffff0000;
        }
        else
        {
            theApp.b_save_colorized_images_with_autostretch = true;
            this.dialog.colorized_image_autostretch_combobox.textColor = 0xff228b22;
        }

    };

    //

    this.settings_groupbox = new GroupBox (this);
    this.settings_groupbox.title = "Settings";
    this.settings_groupbox.minStretchedWidth =  200;
    this.settings_groupbox.sizer = new VerticalSizer;
    this.settings_groupbox.sizer.add(this.show_description_push_button);
    this.settings_groupbox.sizer.addStretch();
    this.settings_groupbox.sizer.spacing = 6;
    this.settings_groupbox.sizer.add( this.settings_label );
    this.settings_groupbox.sizer.add( this.load_save_settings_sizer );
    this.settings_groupbox.sizer.spacing = 6;
    this.settings_groupbox.sizer.add( this.default_palettes_label );
    this.settings_groupbox.sizer.spacing = 6;
    this.settings_groupbox.sizer.add( this.update_restore_default_palettes_sizer );
    this.settings_groupbox.sizer.addStretch();
    this.settings_groupbox.sizer.add( this.save_diagnostics_label );
    this.settings_groupbox.sizer.add( this.save_diagnostics_push_button );
    this.settings_groupbox.sizer.addStretch();
    this.settings_groupbox.sizer.add( this.colorize_method_label );
    this.settings_groupbox.sizer.add( this.colorize_method_combo_box );
    this.settings_groupbox.sizer.add( this.linear_fit_colorized_images_checkBox );
    this.settings_groupbox.sizer.addSpacing(6);
    this.settings_groupbox.sizer.add( this.save_colorized_input_images_label );
    this.settings_groupbox.sizer.add( this.save_colorized_input_images_push_button );
    this.settings_groupbox.sizer.add( this.colorized_image_file_extension_combobox );
    this.settings_groupbox.sizer.add( this.colorized_image_autostretch_combobox );
    this.settings_groupbox.sizer.addSpacing(10);
    this.settings_groupbox.sizer.add( this.combine_method_label );
    this.settings_groupbox.sizer.add( this.combineMethodComboBox );
    this.settings_groupbox.sizer.addSpacing(10);
    this.settings_groupbox.sizer.add( this.rescale_label );
    this.settings_groupbox.sizer.add( this.rescale_method_combo_box );
    this.settings_groupbox.sizer.add( this.rescale_channels_checkBox );

    this.settings_groupbox.sizer.addSpacing(20);


    // Top horizonal sizer control

    this.top_horizontal_sizer = new HorizontalSizer;
    this.top_horizontal_sizer.add( this.review_image_groupbox, 150 );
    this.top_horizontal_sizer.addSpacing(5);
    this.top_horizontal_sizer.add( this.review_image_management_groupbox, 45 );
    this.top_horizontal_sizer.addStretch();
    this.top_horizontal_sizer.add( this.description_groupbox, 45 );
    this.top_horizontal_sizer.addSpacing(5);
    this.top_horizontal_sizer.add( this.settings_groupbox, 45 );

    //

    function AttributesControlSet(attributes_conrol_parent, theApp, controlSet)
    {
        // This dialog maintains 'dialog.attributes_conrol_set_array'
        // of size NUM_DIALOG_ATTRIBUTES_CONTROL_SETS.
        // An 'AttributesControlSet' is recorded in each row of the array,
        // storing dialog controls for each image.
        // This mirors the 'theApp.image_attributes_array' that
        // stores the image settings (ImageAttributesSet) desired by the user
        // There must be one AttributesControlSet for each ImageAttributesSet

        // This function declares the basic controls for each image in the array
        this.__base__ = Object;
        this.__base__();

        //

        let dialog_image_row_height = 25;

        //

        this.row_select_push_button= new PushButton( attributes_conrol_parent);
        this.row_select_push_button.scaledMinHeight = dialog_image_row_height;
        this.row_select_push_button.scaledMaxHeight = dialog_image_row_height;
        this.row_select_push_button.height = dialog_image_row_height;
        this.row_select_push_button.toolTip = "Use checkbox to select/deselect a specific grayscale image as a reference for a linear-fit, or to preview a colorized version";
        this.row_select_push_button.text            = "";
        this.row_select_push_button.backgroundColor = 0xffd0d0d0;
        this.row_select_push_button.foregroundColor = 0xffff0000;
        this.row_select_push_button.icon = new Bitmap( ":/icons/control-button.png" );
        this.row_select_push_button.onClick = function()
        {
            this.dialog.enabled = false;
            this.dialog.dialog_status_message_label.update("Select image : Executing", 0xffff0000);
            this.dialog.onRowSelectionChange(controlSet);
            this.dialog.dialog_status_message_label.update("Select image : Done", 0xff000000);
            this.dialog.enabled = true;
        };

        //

        this.filterTypeComboBox = new ComboBox( attributes_conrol_parent);
        this.filterTypeComboBox.scaledMinHeight = dialog_image_row_height;
        this.filterTypeComboBox.height = dialog_image_row_height;
        for ( let i = 0; i < theApp.filter_names.length; ++i )
        {
            this.filterTypeComboBox.addItem( theApp.filter_names[i] );
        }
        this.filterTypeComboBox.currentItem
           = this.filterTypeComboBox.findItem(theApp.image_attributes_array[controlSet].filter); // Filter
        this.filterTypeComboBox.toolTip = "Filter";
        this.filterTypeComboBox.onItemSelected = function( index )
        {
            this.dialog.enabled = false;
            this.dialog.onFilterChange(controlSet,
                attributes_conrol_parent.dialog.attributes_conrol_set_array[controlSet].filterTypeComboBox.itemText(index));
            this.dialog.enabled = true;
        };

        //

        this.row_load_image_button= new PushButton( attributes_conrol_parent);
        this.row_load_image_button.scaledMinHeight = dialog_image_row_height;
        this.row_load_image_button.scaledMaxHeight = dialog_image_row_height;
        this.row_load_image_button.height = dialog_image_row_height;
        this.row_load_image_button.toolTip = "Open and load image from file system";
        this.row_load_image_button.text            = "";
        this.row_load_image_button.backgroundColor = 0xffd0d0d0;
        this.row_load_image_button.foregroundColor = 0xffff0000;
        this.row_load_image_button.icon = new Bitmap( ":/browser/select-file-1.png" );
        this.row_load_image_button.onClick = function()
        {
            this.dialog.enabled = false;
            let new_view = theApp.selectAndOpenImage();

            if (new_view)
            {
                this.dialog.dialog_status_message_label.update("Updating View: Executing", 0xffff0000);
                this.dialog.onViewChange(new_view, controlSet);
                this.dialog.dialog_status_message_label.update("Updating View: Done", 0xff000000);

            }
            this.dialog.enabled = true;
        };

        //

        this.view_list = new ViewList( attributes_conrol_parent );
        this.view_list.scaledMinWidth = 400;
        this.view_list.scaledMinHeight = dialog_image_row_height;
        this.view_list.scaledMaxHeight = dialog_image_row_height;
        this.view_list.height = dialog_image_row_height;
        this.view_list.excludeIdentifiersPattern = "COMBINED_IMAGE|COMBINED_IMAGE_FINALIZED|REVIEW_IMAGE|SPLASH_IMAGE|WORKING_DUPLICATE|COLORIZED_INPUT_IMAGE";
        this.view_list.getMainViews();
        this.view_list.toolTip = "File/View";
        this.view_list.onViewSelected = function( view )
        {
            this.dialog.enabled = false;
            this.dialog.dialog_status_message_label.update("Updating View: Executing", 0xffff0000);
            this.dialog.onViewChange(view, controlSet);
            this.dialog.dialog_status_message_label.update("Updating View: Done", 0xff000000);
            this.dialog.enabled = true;
        };

        //

        this.palette_reset_push_button = new ToolButton(attributes_conrol_parent);
        this.palette_reset_push_button.scaledMinHeight = dialog_image_row_height;
        this.palette_reset_push_button.height = dialog_image_row_height;
        this.palette_reset_push_button.width          = 6;
        this.palette_reset_push_button.toolTip = "Reset image/filter settings to default palette";
        this.palette_reset_push_button.text            = "Reset";
        this.palette_reset_push_button.backgroundColor = 0xffff0000;
        this.palette_reset_push_button.foregroundColor = 0xff000000;
        this.palette_reset_push_button.icon = new Bitmap( ":/icons/reload.png" );
        this.palette_reset_push_button.onClick = function()
        {
            this.dialog.enabled = false;
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
            theApp.resetImageAttributeArrayRow(controlSet);
            attributes_conrol_parent.updateDialogControls(controlSet);
            this.dialog.enabled = true;
         };

        //

        this.amount_spinbox = new SpinBox( attributes_conrol_parent );
        this.amount_spinbox.scaledMinHeight = dialog_image_row_height;
        this.amount_spinbox.height = dialog_image_row_height;
        this.amount_spinbox.minValue = 0;
        this.amount_spinbox.maxValue = 100;
        this.amount_spinbox.value    =  theApp.image_attributes_array[controlSet].amount * 100; // Amount
        this.amount_spinbox.toolTip  = "<p>Amount [0-100%] - The percentage of these RGB values that will contribute to the combined image</p>";
        this.amount_spinbox.onValueUpdated = function( value )
        {
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
            theApp.image_attributes_array[controlSet].amount = this.normalizedValue
        };

        //

        this.colorComboBox = new ColorComboBox( attributes_conrol_parent);
        this.colorComboBox.scaledMinHeight = dialog_image_row_height;
        this.colorComboBox.height = dialog_image_row_height;
        this.colorComboBox.currentItem
            = this.colorComboBox.findItem( "White", 0, true);
        this.colorComboBox.toolTip = "Pick color from list   ";
        this.colorComboBox.onColorSelected = function(rgba)
        {
            this.dialog.enabled = false;

            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color

            theApp.image_attributes_array[controlSet].red   = Color.red(rgba)   / 255.0;
            theApp.image_attributes_array[controlSet].green = Color.green(rgba) / 255.0;
            theApp.image_attributes_array[controlSet].blue  = Color.blue(rgba)  / 255.0;

            attributes_conrol_parent.dialog.updateDialogControls(controlSet);

            this.dialog.enabled = true;
        };

        //

        this.red_slider = new Slider( attributes_conrol_parent );
        this.red_slider.scaledMinWidth = 220;
        this.red_slider.scaledMinHeight = dialog_image_row_height;
        this.red_slider.height = dialog_image_row_height;
        this.red_slider.minValue = 0;
        this.red_slider.maxValue = 2550;
        this.red_slider.pageSize = 10;
        this.red_slider.value    =  theApp.image_attributes_array[controlSet].red * 2550.0; // red
        this.red_slider.backgroundColor = 0xffc80000;
        this.red_slider.textColor = 0xffffffff;
        this.red_slider.toolTip = "<p>Red Value [0-255]</p>";
        this.red_slider.onValueUpdated = function( new_value )
        {
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color

            if (this.dialog.dialog_color_mode == "RGB")
            {
                theApp.image_attributes_array[controlSet].red   = this.normalizedValue;

                attributes_conrol_parent.dialog.updateDialogControls_RGB_mode(controlSet);
            }
            else // this.dialog_color_mode == "HSV", Red slider adjusts Hue
            {

                let red   = theApp.image_attributes_array[controlSet].red;
                let green = theApp.image_attributes_array[controlSet].green;
                let blue  = theApp.image_attributes_array[controlSet].blue;

                let hsv_color = convertRGBtoHSV(red, green, blue);

                let new_RGB = [ 0, 0 , 0 ];
                if (this.normalizedValue < 1)
                {
                    new_RGB   = convertHSVtoRGB(this.normalizedValue, hsv_color[1], hsv_color[2]);
                }
                else
                {
                    new_RGB   = convertHSVtoRGB(1 - 0.00001, hsv_color[1], hsv_color[2]);
                }

                theApp.image_attributes_array[controlSet].red   = new_RGB[0];
                theApp.image_attributes_array[controlSet].green = new_RGB[1];
                theApp.image_attributes_array[controlSet].blue  = new_RGB[2];

                attributes_conrol_parent.dialog.updateDialogControls_HSV_mode(controlSet);

            }
        };

        //

        this.red_spinbox = new SpinBox( attributes_conrol_parent );
        this.red_spinbox.scaledMinHeight = dialog_image_row_height;
        this.red_spinbox.height = dialog_image_row_height;
        this.red_spinbox.minValue = 0;
        this.red_spinbox.maxValue = 255;
        this.red_spinbox.value    =  theApp.image_attributes_array[controlSet].red * 255.0; // red
        this.red_spinbox.backgroundColor = 0xffc80000;
        this.red_spinbox.textColor = 0xffffffff;
        this.red_spinbox.toolTip = "<p>Red Value [0-255]</p>";
        this.red_spinbox.onValueUpdated = function( new_value )
        {
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color

            if (this.dialog.dialog_color_mode == "RGB")
            {
                theApp.image_attributes_array[controlSet].red   = this.normalizedValue;

                attributes_conrol_parent.dialog.updateDialogControls_RGB_mode(controlSet);
            }
            else // this.dialog.dialog_color_mode == "HSV", Red spinbox adjusts Hue
            {

                let red   = theApp.image_attributes_array[controlSet].red;
                let green = theApp.image_attributes_array[controlSet].green;
                let blue  = theApp.image_attributes_array[controlSet].blue;

                let hsv_color = convertRGBtoHSV(red, green, blue);

                let new_RGB = [ 0, 0 , 0 ];
                if (this.normalizedValue < 1)
                {
                    new_RGB   = convertHSVtoRGB(this.normalizedValue, hsv_color[1], hsv_color[2]);
                }
                else
                {
                   // Subtract a small number to avoid exactly one (1.0)  'Hue', and therby avoid cycling back to zero
                    new_RGB   = convertHSVtoRGB(1 - 0.00001, hsv_color[1], hsv_color[2]);
                }

                theApp.image_attributes_array[controlSet].red   = new_RGB[0];
                theApp.image_attributes_array[controlSet].green = new_RGB[1];
                theApp.image_attributes_array[controlSet].blue  = new_RGB[2];

                attributes_conrol_parent.dialog.updateDialogControls_HSV_mode(controlSet);

            }
        };

        //

        this.green_slider = new Slider( attributes_conrol_parent );
        this.green_slider.scaledMinWidth = 220;
        this.green_slider.scaledMinHeight = dialog_image_row_height;
        this.green_slider.height = dialog_image_row_height;
        this.green_slider.minValue = 0;
        this.green_slider.maxValue = 2550;
        this.green_slider.pageSize = 10;
        this.green_slider.value    =  theApp.image_attributes_array[controlSet].green * 2550.0; // green
        this.green_slider.backgroundColor = 0xff00c800;
        this.green_slider.textColor = 0xffffffff;
        this.green_slider.toolTip = "<p>Green Value [0-255]</p>";
        this.green_slider.onValueUpdated = function( new_value )
        {
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color

            if (this.dialog.dialog_color_mode == "RGB")
            {
                theApp.image_attributes_array[controlSet].green   = this.normalizedValue;

                attributes_conrol_parent.dialog.updateDialogControls_RGB_mode(controlSet);
            }
            else // this.dialog.dialog_color_mode == "HSV", Green slider adjusts Saturation
            {

                let red   = theApp.image_attributes_array[controlSet].red;
                let green = theApp.image_attributes_array[controlSet].green;
                let blue  = theApp.image_attributes_array[controlSet].blue;

                let hsv_color = convertRGBtoHSV(red, green, blue);

                let new_RGB = [ 0, 0 , 0 ];
                if (this.normalizedValue > 0)
                {
                    new_RGB   = convertHSVtoRGB(hsv_color[0], this.normalizedValue, hsv_color[2]);
                }
                else
                {
                    // Add a small number to avoid exactly zero 'Saturation', and therby avoid locking 'Hue'
                    new_RGB   = convertHSVtoRGB(hsv_color[0], this.normalizedValue + 0.00001, hsv_color[2]);
                }

                theApp.image_attributes_array[controlSet].red   = new_RGB[0];
                theApp.image_attributes_array[controlSet].green = new_RGB[1];
                theApp.image_attributes_array[controlSet].blue  = new_RGB[2];

                attributes_conrol_parent.dialog.updateDialogControls_HSV_mode(controlSet);

            }
        };

        //

        this.green_spinbox = new SpinBox( attributes_conrol_parent );
        this.green_spinbox.scaledMinHeight = dialog_image_row_height;
        this.green_spinbox.height = dialog_image_row_height;
        this.green_spinbox.minValue = 0;
        this.green_spinbox.maxValue = 255;
        this.green_spinbox.value    =  theApp.image_attributes_array[controlSet].green * 255.0; // green
        this.green_spinbox.backgroundColor = 0xff00c800;
        this.green_spinbox.textColor = 0xffffffff;
        this.green_spinbox.toolTip = "<p>Green Value [0-255]</p>";
        this.green_spinbox.onValueUpdated = function( new_value )
        {
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color

            if (this.dialog.dialog_color_mode == "RGB")
            {
                theApp.image_attributes_array[controlSet].green   = this.normalizedValue;

                attributes_conrol_parent.dialog.updateDialogControls_RGB_mode(controlSet);
            }
            else // this.dialog.dialog_color_mode == "HSV", Green spinbox adjusts Saturation
            {

                let red   = theApp.image_attributes_array[controlSet].red;
                let green = theApp.image_attributes_array[controlSet].green;
                let blue  = theApp.image_attributes_array[controlSet].blue;

                let hsv_color = convertRGBtoHSV(red, green, blue);

                let new_RGB = [ 0, 0 , 0 ];
                if (this.normalizedValue > 0)
                {
                    new_RGB   = convertHSVtoRGB(hsv_color[0], this.normalizedValue, hsv_color[2]);
                }
                else
                {
                    // Add a small number to avoid exactly zero 'Saturation', and therby avoid locking 'Hue'
                    new_RGB   = convertHSVtoRGB(hsv_color[0], this.normalizedValue + 0.00001, hsv_color[2]);
                }

                theApp.image_attributes_array[controlSet].red   = new_RGB[0];
                theApp.image_attributes_array[controlSet].green = new_RGB[1];
                theApp.image_attributes_array[controlSet].blue  = new_RGB[2];

                attributes_conrol_parent.dialog.updateDialogControls_HSV_mode(controlSet);

            }
        };

        //

        this.blue_slider = new Slider( attributes_conrol_parent );
        this.blue_slider.scaledMinWidth = 220;
        this.blue_slider.scaledMinHeight = dialog_image_row_height;
        this.blue_slider.height = dialog_image_row_height;
        this.blue_slider.minValue = 0;
        this.blue_slider.maxValue = 2550;
        this.blue_slider.pageSize = 10;
        this.blue_slider.value    =  theApp.image_attributes_array[controlSet].blue * 2550.0 // blue
        this.blue_slider.scaledMinWidth = 220;
        this.blue_slider.backgroundColor = 0xff0000c8;
        this.blue_slider.textColor = 0xffffffff;
        this.blue_slider.toolTip = "<p>Blue Value [0-255]</p>";
        this.blue_slider.onValueUpdated = function( new_value )
        {
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color

            if (this.dialog.dialog_color_mode == "RGB")
            {
                theApp.image_attributes_array[controlSet].blue   = this.normalizedValue;

                attributes_conrol_parent.dialog.updateDialogControls_RGB_mode(controlSet);
            }
            else // this.dialog.dialog_color_mode == "HSV", Blue slider adjusts Value
            {

                let red   = theApp.image_attributes_array[controlSet].red;
                let green = theApp.image_attributes_array[controlSet].green;
                let blue  = theApp.image_attributes_array[controlSet].blue;

                let hsv_color = convertRGBtoHSV(red, green, blue);

                let new_RGB = [ 0, 0 , 0 ];
                if (this.normalizedValue > 0)
                {
                    new_RGB   = convertHSVtoRGB(hsv_color[0], hsv_color[1], this.normalizedValue);
                }
                else
                {
                    // Add a small number to avoid exactly zero 'Value', and therby avoid locking 'Saturation' and 'Hue'
                    new_RGB   = convertHSVtoRGB(hsv_color[0], hsv_color[1], this.normalizedValue  + 0.00001 );
                }

                theApp.image_attributes_array[controlSet].red   = new_RGB[0];
                theApp.image_attributes_array[controlSet].green = new_RGB[1];
                theApp.image_attributes_array[controlSet].blue  = new_RGB[2];

                attributes_conrol_parent.dialog.updateDialogControls_HSV_mode(controlSet);

            }
        };

        this.blue_spinbox = new SpinBox( attributes_conrol_parent );
        this.blue_spinbox.scaledMinHeight = dialog_image_row_height;
        this.blue_spinbox.height = dialog_image_row_height;
        this.blue_spinbox.minValue = 0;
        this.blue_spinbox.maxValue = 255;
        this.blue_spinbox.value    =  theApp.image_attributes_array[controlSet].blue * 255.0; // blue
        this.blue_spinbox.backgroundColor = 0xff0000c8;
        this.blue_spinbox.textColor = 0xffffffff;
        this.blue_spinbox.toolTip = "<p>Blue Value [0-255]</p>";
        this.blue_spinbox.onValueUpdated = function( new_value )
        {
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color

            if (this.dialog.dialog_color_mode == "RGB")
            {
                theApp.image_attributes_array[controlSet].blue   = this.normalizedValue;

                attributes_conrol_parent.dialog.updateDialogControls_RGB_mode(controlSet);
            }
            else // this.dialog.dialog_color_mode == "HSV", Blue spinbox adjusts Value
            {

                let red   = theApp.image_attributes_array[controlSet].red;
                let green = theApp.image_attributes_array[controlSet].green;
                let blue  = theApp.image_attributes_array[controlSet].blue;

                let hsv_color = convertRGBtoHSV(red, green, blue);

                let new_RGB = [ 0, 0 , 0 ];
                if (this.normalizedValue > 0)
                {
                    new_RGB   = convertHSVtoRGB(hsv_color[0], hsv_color[1], this.normalizedValue);
                }
                else
                {
                   // Add a small number to avoid exactly zero 'Value', and therby avoid locking 'Saturation' and 'Hue'
                    new_RGB   = convertHSVtoRGB(hsv_color[0], hsv_color[1], this.normalizedValue  + 0.00001 );
                }

                theApp.image_attributes_array[controlSet].red   = new_RGB[0];
                theApp.image_attributes_array[controlSet].green = new_RGB[1];
                theApp.image_attributes_array[controlSet].blue  = new_RGB[2];

                attributes_conrol_parent.dialog.updateDialogControls_HSV_mode(controlSet);

            }
        };
    };

    AttributesControlSet.prototype = new Object;

    //

    // attributes_conrol_set_array, one row for each set of image dialog controls

    this.attributes_conrol_set_array = new Array;

    for (let i = 0; i < NUM_DIALOG_ATTRIBUTES_CONTROL_SETS; i++)
    {
        let newDialogAttributeSet
              = new AttributesControlSet(this, theApp, i);
        this.attributes_conrol_set_array.push(newDialogAttributeSet);
    }

    // An set of horizontal and vertical sizers to arrange data entry rows for each image.
    // The  arrangement is quite complicated to make sure labels continue to align
    // with columns (of dialog rows) when resizing the Dialog

    let num_columns = 13;
    let dialog_image_column_names_array = ["SELECT", "FILTER", "", "IMAGE", "", "AMOUNT", "COMBINED COLOR (RGB)", "", "RED", "", "GREEN", "", "BLUE"];
    this.dialog_image_column_label_controls_array = [];  // For access throughout the Dialog

    this.dialog_image_horizontal_sizer = new HorizontalSizer;

    let dialog_image_vertical_sizer_array = [];

    for (let column = 0; column < num_columns; column++)
    {

        let dialog_image_vertical_sizer = new VerticalSizer;
        dialog_image_vertical_sizer.spacing = 6;

        dialog_image_vertical_sizer_array.push(dialog_image_vertical_sizer);

        let label =  new Label( this );
        label.text = dialog_image_column_names_array[column];
        label.textAlignment = TextAlign_Center| TextAlign_VertCenter;

        this.dialog_image_column_label_controls_array.push(label); // For access throughout the Dialog

        dialog_image_vertical_sizer.add(label);

        for (let image_row = 0; image_row < NUM_DIALOG_ATTRIBUTES_CONTROL_SETS; image_row++)
        {
            switch (column)
            {
               case  0: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].row_select_push_button);
                        break;
               case  1: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].filterTypeComboBox);
                        break;
               case  2: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].row_load_image_button);
                        break;
               case  3: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].view_list);
                        break;
               case  4: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].palette_reset_push_button);
                        break;
               case  5: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].amount_spinbox);
                        break;
               case  6: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].colorComboBox);
                        break;
               case  7: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].red_spinbox);
                        break;
               case  8: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].red_slider);
                        break;
               case  9: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].green_spinbox);
                        break;
               case 10: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].green_slider);
                        break;
               case 11: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].blue_spinbox);
                        break;
               case 12: dialog_image_vertical_sizer.add(this.attributes_conrol_set_array[image_row].blue_slider);
                        break;
           }
        }
        switch (column)
        {
            case 4: this.dialog_image_horizontal_sizer.addSpacing(10);
               break;
        }

        this.dialog_image_horizontal_sizer.add(dialog_image_vertical_sizer_array[column]);

    };

    //

    // Color Mode Controls

    this.dialog_color_mode_label = new Label( this );
    this.dialog_color_mode_label.useRichText = true;
    this.dialog_color_mode_label.text = "<b>RGB/HSV: </b>";
    this.dialog_color_mode_label.textAlignment = TextAlign_Right| TextAlign_VertCenter;

    //

    this.dialog_color_mode_push_button = new PushButton( this);
    this.dialog_color_mode_push_button.toolTip = "Select RGB or HSV mode";
    this.dialog_color_mode_push_button.text            = this.dialog_color_mode;
    this.dialog_color_mode_push_button.backgroundColor = 0xffd0d0d0;
    this.dialog_color_mode_push_button.foregroundColor = 0xffff0000;
    this.dialog_color_mode_push_button.icon = new Bitmap( ":/toolbar/image-display-rgb.png" );
    this.dialog_color_mode_push_button.minWidth = 80;
    this.dialog_color_mode_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        if (this.dialog.dialog_color_mode.toUpperCase()  == "RGB")
        {
           this.dialog.dialog_color_mode = "HSV"
        }
        else
        {
           this.dialog.dialog_color_mode = "RGB"
        }
        this.dialog.onDialogColorModeChange(this.dialog.dialog_color_mode);
        this.dialog.enabled = true;
    };

    //

    this.dialog_color_mode_groupbox = new GroupBox (this);
    this.dialog_color_mode_groupbox.title = "Color Mode";
    this.dialog_color_mode_groupbox.sizer = new HorizontalSizer;
    this.dialog_color_mode_groupbox.sizer.addSpacing(4);
    this.dialog_color_mode_groupbox.sizer.add( this.dialog_color_mode_label );
    this.dialog_color_mode_groupbox.sizer.addSpacing(4);
    this.dialog_color_mode_groupbox.sizer.add( this.dialog_color_mode_push_button );
    this.dialog_color_mode_groupbox.sizer.addSpacing(4);

    //

    this.color_wheel_button = new PushButton( this );
    this.color_wheel_button.text            = "Color Spread";
    this.color_wheel_button.backgroundColor = 0xffd0d0d0;
    this.color_wheel_button.foregroundColor = 0xffff0000;
    this.color_wheel_button.icon = new Bitmap( ":/icons/color-wheel.png" );
    this.color_wheel_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.onColorSpread();
        this.dialog.enabled = true;
    };

    this.special_colors_groupbox = new GroupBox (this);
    this.special_colors_groupbox.title = "Special Colors";
    this.special_colors_groupbox.sizer = new HorizontalSizer;
    this.special_colors_groupbox.sizer.add( this.color_wheel_button );

    //  Palette Selection Controls

    this.palette_auto_select_push_button = new PushButton(this);
    this.palette_auto_select_push_button.toolTip = "Automatically set palette for known filter";
    this.palette_auto_select_push_button.text            = "Auto";
    if (theApp.b_auto_set_palette_for_known_filter)
    {
       this.palette_auto_select_push_button.backgroundColor = 0xff909090;
    }
    else
    {
       this.palette_auto_select_push_button.backgroundColor = 0xffd0d0d0;
    }
    this.palette_auto_select_push_button.foregroundColor = 0xff000000;
    this.palette_auto_select_push_button.icon = new Bitmap( ":/image-container/image-container-icon.png" );
    this.palette_auto_select_push_button.onClick = function()
    {
        this.dialog.enabled = false;
        theApp.b_auto_set_palette_for_known_filter = ! theApp.b_auto_set_palette_for_known_filter;
        if (theApp.b_auto_set_palette_for_known_filter)
        {
            this.dialog.palette_auto_select_push_button.backgroundColor = 0xff909090;
        }
        else
        {
            this.dialog.palette_auto_select_push_button.backgroundColor = 0xffd0d0d0;
        }

        this.dialog.enabled = true;
     };

    //

    this.NO_PALETTE_button = new PushButton( this );
    this.NO_PALETTE_button.text            = "NONE";
    this.NO_PALETTE_button.backgroundColor = 0xffd0d0d0;
    this.NO_PALETTE_button.foregroundColor = 0xffff0000;
    this.NO_PALETTE_button.icon = new Bitmap( ":/image-container/image-container-icon.png" );
    this.NO_PALETTE_button.onClick = function()
    {
        this.dialog.enabled = false;
        theApp.b_auto_set_palette_for_known_filter = false;
        this.dialog.palette_auto_select_push_button.backgroundColor = 0xffd0d0d0;
        this.dialog.onPaletteChange("NONE", false);  // false == do not reset previous settings
        this.dialog.enabled = true;
    };

    //

    this.HST_button = new PushButton( this );
    this.HST_button.text            = "HST";
    this.HST_button.backgroundColor = 0xffd0d0d0;
    this.HST_button.foregroundColor = 0xff0000ff;
    this.HST_button.icon = new Bitmap( ":/image-container/image-container-icon.png" );
    this.HST_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.onPaletteChange("HST", false);  // false == do not reset previous settings
        this.dialog.enabled = true;
    };

    //

    this.CFHT_button = new PushButton( this );
    this.CFHT_button.text = "CFHT";
    this.CFHT_button.backgroundColor = 0xffd0d0d0;
    this.CFHT_button.foregroundColor = 0xff0000ff;
    this.CFHT_button.icon = new Bitmap( ":/image-container/image-container-icon.png" );
    this.CFHT_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.onPaletteChange("CFHT", false);  // false == do not reset previous settings
        this.dialog.enabled = true;
    };

    //

    this.SDSS_button = new PushButton( this );
    this.SDSS_button.text = "SDSS";
    this.SDSS_button.backgroundColor = 0xffd0d0d0;
    this.SDSS_button.foregroundColor = 0xff0000ff;
    this.SDSS_button.icon = new Bitmap( ":/image-container/image-container-icon.png" );
    this.SDSS_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.onPaletteChange("SDSS", false);  // false == do not reset previous settings
        this.dialog.enabled = true;
    };

    //

    this.JWST_button = new PushButton( this );
    this.JWST_button.text = "JWST";
    this.JWST_button.backgroundColor = 0xffd0d0d0;
    this.JWST_button.foregroundColor = 0xff0000ff;
    this.JWST_button.icon = new Bitmap( ":/image-container/image-container-icon.png" );
    this.JWST_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.onPaletteChange("JWST", false);  // false == do not reset previous settings
        this.dialog.enabled = true;
    };

    //

    this.NATURAL_button = new PushButton( this );
    this.NATURAL_button.text = "Natural";
    this.NATURAL_button.backgroundColor = 0xff909090;
    this.NATURAL_button.foregroundColor = 0xffff7f00;
    this.NATURAL_button.icon = new Bitmap( ":/image-container/image-container-icon.png" );
    this.NATURAL_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.onPaletteChange("NATURAL", false);  // false == do not reset previously set images
        this.dialog.enabled = true;
    };

    //

    this.palette_select_groupbox = new GroupBox(this);
    this.palette_select_groupbox.title = "Palette";
    this.palette_select_groupbox.sizer = new HorizontalSizer;
    this.palette_select_groupbox.sizer.add( this.palette_auto_select_push_button );
    this.palette_select_groupbox.sizer.add( this.HST_button );
    this.palette_select_groupbox.sizer.add( this.CFHT_button );
    this.palette_select_groupbox.sizer.add( this.SDSS_button );
    this.palette_select_groupbox.sizer.add( this.JWST_button );
    this.palette_select_groupbox.sizer.add( this.NATURAL_button );
    this.palette_select_groupbox.sizer.add( this.NO_PALETTE_button );

    // Modes horizontal sizer control

    this.modes_horizontal_sizer = new HorizontalSizer;
    this.modes_horizontal_sizer.spacing = 6;
    this.modes_horizontal_sizer.addStretch();
    this.modes_horizontal_sizer.add( this.dialog_color_mode_groupbox);
    this.modes_horizontal_sizer.addSpacing(40);
    this.modes_horizontal_sizer.add( this.special_colors_groupbox);
    this.modes_horizontal_sizer.addSpacing(10);
    this.modes_horizontal_sizer.add( this.palette_select_groupbox);

    //  Execution controls

    this.preprocessing_method_label = new Label( this );
    this.preprocessing_method_label.useRichText = true;
    this.preprocessing_method_label.text = "<b>Method: </b>";
    this.preprocessing_method_label.textAlignment = TextAlign_Left| TextAlign_VertCenter;

    //

    this.preprocessing_method_combo_box = new ComboBox( this);
    for (let i = 0; i < theApp.preprocessing_methods.length; i++ )
    {
        this.preprocessing_method_combo_box.addItem( theApp.preprocessing_methods[i] );
    }
    this.preprocessing_method_combo_box.currentItem
        = this.preprocessing_method_combo_box.findItem( theApp.preprocessing_method, 0, true);
    this.preprocessing_method_combo_box.toolTip = "Optional preprocessing methods";
    this.preprocessing_method_combo_box.onItemSelected = function( index )
    {
        theApp.preprocessing_method = this.dialog.preprocessing_method_combo_box.itemText(index);
    };

    //

    this.execute_preprocess_button = new PushButton( this );
    this.execute_preprocess_button.text = "Execute Preprocess";
    this.execute_preprocess_button.icon = this.scaledResource( ":/icons/ok.png" );
    this.execute_preprocess_button.onClick = function()
    {
        if (theApp.preprocessing_method != "NONE")
        {

            this.dialog.enabled = false;
            this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
            this.dialog.dialog_status_message_label.update("Preprocess Images (" +  theApp.preprocessing_method + ") : Executing", 0xffff0000);
            theApp.preprocessAllImages();
            this.dialog.dialog_status_message_label.update("Preprocess Images : Done", 0xff000000);
            this.dialog.enabled = true;
            this.dialog.conditionalAbort();
        }
    };

    //

    this.preprocess_groupbox = new GroupBox (this);
    this.preprocess_groupbox.title = "Preprocessing";
    this.preprocess_groupbox.sizer = new HorizontalSizer;
    this.preprocess_groupbox.sizer.spacing = 6;
    this.preprocess_groupbox.sizer.add( this.preprocessing_method_label);
    this.preprocess_groupbox.sizer.add( this.preprocessing_method_combo_box );
    this.preprocess_groupbox.sizer.add( this.execute_preprocess_button );


    //

    this.dialog_status_message_label = new Label( this );
    this.dialog_status_message_label.minWidth = 450;
    this.dialog_status_message_label.text = "";
    this.dialog_status_message_label.textAlignment = TextAlign_Left| TextAlign_VertCenter;
    this.dialog_status_message_label.update = function(text = "", color = 0xff000000)
    {
        this.dialog.dialog_status_message_label.textColor = color;
        this.dialog.dialog_status_message_label.text = text;
        this.dialog.app_status_message_label.clear();
    }
    this.dialog_status_message_label.clear = function(color = 0xff000000)
    {
        this.dialog.dialog_status_message_label.textColor = color;
        this.dialog.dialog_status_message_label.text = "";
        this.dialog.app_status_message_label.clear();
    }

    //

    // This the only thing theApp 'knows' about the dialog object
    this.app_status_message_label = new Label( this );
    this.app_status_message_label.minWidth = 370;
    this.app_status_message_label.text = "";
    this.app_status_message_label.textAlignment = TextAlign_Left| TextAlign_VertCenter;
    this.app_status_message_label.update = function(text = "", color = 0xff000000)
    {
        this.dialog.app_status_message_label.textColor = color;
        this.dialog.app_status_message_label.text = text;
    }
    this.dialog.app_status_message_label.clear = function(color = 0xff000000)
    {
        this.dialog.app_status_message_label.textColor = color;
        this.dialog.app_status_message_label.text = "";
    }

    //

    this.status_groupbox = new GroupBox (this);
    this.status_groupbox.title = "Status";
    this.status_groupbox.sizer = new HorizontalSizer;
    this.status_groupbox.sizer.spacing = 6;
    this.status_groupbox.sizer.addSpacing(6);
    this.status_groupbox.sizer.add( this.dialog_status_message_label);
    this.status_groupbox.sizer.add( this.app_status_message_label);

    //

    this.combine_button = new PushButton( this );
    this.combine_button.text = "Combine";
    this.combine_button.icon = this.scaledResource( ":/icons/execute.png" );
    this.combine_button.onClick = function()
    {

        if (theApp.b_linear_fit_colorized_images_before_combine & theApp.selected_view_row  <  0)
        {
             (new MessageBox("<p>Linear Fit requires a reference image. </p>",
                        "ERROR", StdIcon_Error, StdButton_Ok)).execute();
        }
        else
        {
            this.dialog.enabled = false;
            this.dialog.dialog_status_message_label.update("Combine Images ("
                                                        + theApp.colorize_method
                                                        + ", "
                                                        + theApp.combine_method
                                                        + ", "
                                                        + theApp.rescale_method
                                                        + ") : Executing",
                                                        0xffff0000);
            if (theApp.combineImages())
            {
                this.dialog.b_onResize_pause = false;
                this.dialog.updateReviewImage();
                this.dialog.enableReviewImageManagementPanel_1(true);
                this.dialog.onSettingsChanged(false, true); // true == settings NOT changed. true == update alert color
            }
            this.dialog.dialog_status_message_label.update("Combine Images : Done", 0xff000000);
            this.dialog.enabled = true;
        }
        this.dialog.conditionalAbort();
    };

    //

    this.working_images_shrink_mode_combo_box = new ComboBox( this);
    for (let i = 0; i < theApp.working_images_shrink_mode_array.length; i++ )
    {
        this.working_images_shrink_mode_combo_box.addItem( theApp.working_images_shrink_mode_array[i] );
    }
    this.working_images_shrink_mode_combo_box.currentItem
        = this.working_images_shrink_mode_combo_box.findItem( (theApp.working_images_shrink_mode), 0, true);
    this.working_images_shrink_mode_combo_box.toolTip = "Enter shrink mode for preview";
    this.working_images_shrink_mode_combo_box.onItemSelected = function( index )
    {
        this.dialog.enabled = false;
        this.dialog.onSettingsChanged(true, true); // true == settings changed. true == set alert color
        let working_images_shrink_mode = this.dialog.working_images_shrink_mode_combo_box.itemText(index);
        this.dialog.dialog_status_message_label.update("Create working images (Shrink "
                                                        + working_images_shrink_mode
                                                        + ") : Executing", 0xffff0000);
        let working_images_shrink_mode = this.dialog.working_images_shrink_mode_combo_box.itemText(index);
        theApp.setWorkingImagesShrinkMode(working_images_shrink_mode);
        this.dialog.dialog_status_message_label.update("Create working images: Done", 0xff000000);
        this.dialog.enabled = true;
    };

    this.resample_percentage_label = new Label( this );
    this.resample_percentage_label.minWidth = 15;
    this.resample_percentage_label.text = "%";
    this.resample_percentage_label.textAlignment = TextAlign_Left| TextAlign_VertCenter;

    //

    this.save_button = new PushButton( this );
    this.save_button.text = "Save";
    this.save_button.icon = this.scaledResource( ":/icons/picture-export.png" );
    this.save_button.onClick = function()
    {
        this.dialog.enabled = false;
        this.dialog.dialog_status_message_label.update("Save Combined Image : Executing", 0xffff0000);
        theApp.saveCombinedImage();
        this.dialog.dialog_status_message_label.update("Save Combined Image : Done", 0xff000000);
        this.dialog.enabled = true;
    };

    //

    this.cancel_button = new PushButton( this );
    this.cancel_button.text = "Exit";
    this.cancel_button.icon = this.scaledResource( ":/icons/power.png" );
    this.cancel_button.onClick = function()
    {
        theApp.close();
        this.dialog.enabled = true;
        this.dialog.cancel();
    };

    //

    this.execution_horizontal_sizer = new HorizontalSizer;
    this.execution_horizontal_sizer.spacing = 6;
    this.execution_horizontal_sizer.add( this.preprocess_groupbox );
    this.execution_horizontal_sizer.addStretch();
    this.execution_horizontal_sizer.add( this.status_groupbox );
    this.execution_horizontal_sizer.addStretch();
    this.execution_horizontal_sizer.add( this.combine_button );
    this.execution_horizontal_sizer.add( this.working_images_shrink_mode_combo_box );
    this.execution_horizontal_sizer.add( this.resample_percentage_label );
    this.execution_horizontal_sizer.add( this.save_button );
    this.execution_horizontal_sizer.add( this.cancel_button );

    //
    // Add all horizontal sizers to a single VerticalSizer
    this.sizer = new VerticalSizer;
    this.sizer.margin  = 6;
    this.sizer.spacing = 6;
    this.sizer.add( this.top_horizontal_sizer );
    this.sizer.addSpacing( 4 );
    this.sizer.add( this.dialog_image_horizontal_sizer );
    this.sizer.addSpacing( 4 );
    this.sizer.add( this.modes_horizontal_sizer );
    this.sizer.addSpacing( 4 );
    this.sizer.add( this.execution_horizontal_sizer );

};

//------------------------------------------------------------------------------

// Dialog inherits properties and methods from the parent Dialog object.
AppDialog.prototype = new Dialog;

//------------------------------------------------------------------------------

#endif   // __DC_CombineImages_dialog_js

///////////////////////////////////////////////////////////////////////////////
// EOF DC_CombineImages_dialog.js
///////////////////////////////////////////////////////////////////////////////
