///////////////////////////////////////////////////////////////////////////////
// DC_CombineImages_the_app.js
///////////////////////////////////////////////////////////////////////////////
//
// DC_CombineImages: Version 0.61
//
// A script to combine multiple grayscale images to form a single color image
// with user-defined colors
//
// Copyright (c) 2018-2022 Dean Carr
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __DC_CombineImages_the_app_js
#define __DC_CombineImages_the_app_js

//------------------------------------------------------------------------------

#include <pjsr/DataType.jsh>

#include "DC_ProcessHelpers.js"

#include "DC_CombineImages_default_settings.js"

#include "DC_CombineImages_default_system_palettes.js"

///////////////////////////////////////////////////////////////////////////////

// This script maintains three primary objects: 'g_control_data', 'theApp',
// and 'dialog'.

// g_control_data, based on GlobalControlData, records basic 'wrapper'
// information such as 'script_title', 'script_version', and
// 'root_folder' (where the top-level' script resides).

// 'dialog', based on 'AppDialog()', provides the user interface
// and controls all interactions with 'theApp'.

// 'theApp', based on 'TheApplication()', defines data and methods
// directly related to the operation of the script. 'theApp'has no'knowledge' of
// 'dialog', except for a local reference to the dialog's status_bar.

///////////////////////////////////////////////////////////////////////////////
// Class: TheApplication()
///////////////////////////////////////////////////////////////////////////////

function TheApplication(root_folder)
{

    this.__base__ = Object;
    this.__base__();

    this.ImageAttributesSet = function(b_active, filter, amount, red, green, blue)
    {

        // This script maintains the array 'image_attributes_array' of size NUM_VIEW_SLOTS.
        // An 'ImageAttributesSet' is recorded in each row of the array,
        // detailing the following elements: b_active, filter, amount, and RGB values

        // This function declares the basic attribute set for each image in the array
        this.__base__ = Object;
        this.__base__();

        // this.view    = view;
        this.b_active  = b_active;  // Is true only if a view has been selected and recorded in this particular row of the array
        this.filter    = filter;    // The filter assigned to the image, either found from the metadata or assigned by the user.
                                // Used only to provide default RGB color values based on the currently selected pallette
        this.amount    = amount;    // The proportion of the image pixel value to modulate the assigned RGB color channels
        // Colors, expressed as RGB values (0.0-1.0), can be set by selecting a particlar filter/palette combination,
        // or independently by the user.
        this.red       = red;       // The Red channel of the RGB color currently assiged to the image
        this.green     = green;     // The Green channel of the RGB color currentlyassiged to the image
        this.blue      = blue;      // The Blue channel of the RGB color currently assiged to the image
    };

    //--------------------------------------------------------------------------

    this.ImageAttributesSet.prototype = new Object;

    // MEMBER PROPERTIES -------------------------------------------------------

    this.script_file_path                    = #__FILE__;
    this.root_folder                         = root_folder;

    this.b_auto_populate_row_entries              = B_AUTO_POPULATE_ROW_ENTRIES;
    this.b_auto_STF                               = B_AUTO_STF;
    this.b_linked_RGB_channels                    = B_LINKED_RGB_CHANNELS;
    this.b_save_diagnostics                       = B_SAVE_DIAGNOSTICS;
    this.b_save_colorized_images                  = B_SAVE_COLORIZED_IMAGES;
    this.b_save_colorized_images_with_autostretch = B_SAVE_COLORIZED_IMAGES_WITH_AUTOSTRETCH;
    this.b_auto_set_palette_for_known_filter      = B_AUTO_SET_KNOWN_PALETTE;

    this.input_images_folder                 = this.root_folder + "/" + INPUT_IMAGES_FOLDER;
    this.saved_combined_images_folder        = this.root_folder + "/" + SAVED_COMBINED_IMAGES_FOLDER;
    this.saved_colorized_images_folder       = this.root_folder + "/" + SAVED_COLORIZED_IMAGES_FOLDER;
    this.saved_settings_folder               = this.root_folder + "/" + SAVED_SETTINGS_FOLDER
    this.saved_diagnostics_folder            = this.root_folder + "/" + SAVED_DIAGNOSTICS_FOLDER;
    this.default_palettes_folder             = this.root_folder + "/" + DEFAULT_PALETTES_FOLDER;

    //Image preprocessing methods
    this.preprocessing_methods    = ["NONE", "BLACKPOINT CLIP", "LINEAR FIT"];
    this.preprocessing_method     = DEFAULT_PREPROCESSING_METHOD;

    // Array of colorize methods
    this.colorize_methods_array =   ["MULTIPLY", "HSI INTENSITY REPLACE", "HSV VALUE REPLACE", "LRGBCOMBINATION"];
    this.colorize_method          = DEFAULT_COLORIZE_METHOD;

    // Image combination methods
    this.combine_methods          = ["MEAN", "MEDIAN", "SUM", "MAX"];
    this.combine_method           = DEFAULT_COMBINE_METHOD;

    this.b_linear_fit_colorized_images_before_combine = false;

    // Rescale methods
    this.rescale_methods          = [ "TRUNCATE",  "RESCALE", "RESCALE_IF_NEEDED"];
    this.rescale_method           = DEFAULT_RESCALE_METHOD;

    this.b_rescale_channels_independently = false;

    // File extensions
    this.file_extensions_array    = [".xisf",".fit",".fits",".tif",".jpg",".png" ];
    this.colorized_image_file_extension = SAVE_COLORIZED_IMAGE_FILE_EXTENSION;

    // An array to hold image attributes
    this.image_attributes_array                  = [];

    // An array to hold references to multiple image views
    this.view_array                              = [];
    this.working_view_array                      = [];   // Working copies of grayscale images, reduced in size for faster processing
    this.working_view_for_colorized_images_array = [];   // Special working copies of grayscale images, greatly reduced in size for processing colorized_image_preview
    this.auto_opened_views_array                 = [];  // For images auto-opened by the script

    // References to various image views,
    this.combined_image_view           = null;
    this.review_image_view             = null;
    this.colorized_image_preview       = null;
    this.splash_image                  = null;

    // Working image settings
    this.working_images_shrink_mode_array = [ "FULL SIZE",
                                              "RESAMPLE_0.50", "RESAMPLE_0.25", "RESAMPLE_0.10",
                                              "CROP_0.50", "CROP_0.25", "CROP_0.10"
                                            ];
    this.working_images_shrink_mode         =   "FULL SIZE";
    this.working_images_size                =   [-1,   -1 ];
    this.colorized_image_preview_size       =   [-1,   -1 ];
    this.working_images_center_normalized   =   [0.5, 0.5 ];

    // Palettes
    this.palette_names = [ "NONE", "NATURAL", "HST", "CFHT", "JWST", "SDSS" ];
    this.palette       = DEFAULT_PALETTE;
    this.palettes      = null;

    // Filters
    this.filter_names  =  [];

    this.filter_names_JWST = [ // MIRI
                               "F560W",  "F770W",  "F1000W", "F1130W", "F1280W",
                               "F1500W", "F1800W", "F2100W", "F2550W", "F2550WR",
                               // NIRI
                               "F070W",  "F090W",  "F115W",  "F140M",  "F150W",
                               "F150W2", "F162M",  "F164N",  "F182M",  "F187N",
                               "F200W",  "F210M",  "F212N",  "F250M",  "F277W",
                               "F300M",  "F322W2", "F323N",  "F335M",  "F360M",
                               "F356W",  "F405N",  "F410M",  "F430M",  "F444W",
                               "F460M",  "F466N",  "F470N",  "F480M",
                             ];

    this.priority_JWST_filter_array = ["F162M", "F164N", "F323N", "F405N", "F466N", "F470N"];

    this.filter_names_SDSS = ["SDSS-R",  "SDSS-G",  "SDSS-I", "SDSS-U", "SDSS-Z",
                              "U", "Z"];

    this.selected_view_row                  = -1;

    this.review_image_saturation_boost      = 0.0;
    this.review_image_midtone_adjustment    = 0.5;
    this.review_image_shadows_clipping      = SHADOWS_CLIP;

    this.colorized_image_preview_midtone_adjustment = 0.5;

    this.saved_STF = [ // c0, c1, m, r0, r1
                         [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                         [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                         [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                         [0.00000, 1.00000, 0.50000, 0.00000, 1.00000]
                     ];

    this.saved_H   = [ // c0, m, c1, r0, r1
                         [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                         [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                         [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                         [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                         [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000]
                     ];

    this.b_use_saved_histogram                = false;

    // Set true when ANY setting is changed, otherwise false until a change is made
    this.b_settings_changed_since_combine     = false;

    // Must be initialised by the Dialog.
    // This the only thing theApp 'knows' about the dialog object
    this.status_bar                           = null;

    this.b_processes_running                  = false;

    // MEMBER METHODS ----------------------------------------------------------

    this.initialize = function()
    {

        this.b_processes_running = true;

        this.initialize_implementation();

        gc();   // Garbage collection

        this.b_processes_running = false;

    };

    //

    this.clearStatus = function()
    {
        this.clearStatus_implementation()
    };

    //

    this.updateStatus = function(text = "", color = 0xff000000)
    {
        this.updateStatus_implementation(text, color)
    };

    //

    this.initializeDiagnostics = function()
    {
        this.initializeDiagnostics_implementation();
    };

    //

    this.saveDiagnostics = function()
    {
        this.b_processes_running = true;

        this.saveDiagnostics_implementation(this.b_save_diagnostics,
                                            this.saved_diagnostics_folder);
        this.b_processes_running = false;
    };

    //

    this.loadSplashImage = function()
    {
        this.b_processes_running = true;

        let result = this.loadSplashImage_implementation();

        this.b_processes_running = false;

        return result;

    };

    //

    this.setSettingsChangedFlag = function(b_settings_changed_since_combine = true)
    {
        this.setSettingsChangedFlag_implementation(b_settings_changed_since_combine);
    };

    //

    this.preprocessAllImages = function()
    {
        this.b_processes_running = true;

        this.preprocessAllImages_implementation(this.view_array,
                                                this.image_attributes_array,
                                                this.preprocessing_method,
                                                this.selected_view_row,
                                                true);  // true  == rebuild working images
        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.blackpointClipImage = function(view)
    {
        this.b_processes_running = true;

        this.blackpointClipImage_implementation(view);

        gc();

        this.b_processes_running = false;
    };

    //

    this.linearFitImage = function(view, reference_view,
                                   b_protect_null_RGB_channels = false,
                                   b_continue_on_error         = false,
                                   b_silent_on_error           = false)
    {
        this.b_processes_running = true;

        this.linearFitImage_implementation(view, reference_view,
                                           b_protect_null_RGB_channels,
                                           b_continue_on_error,
                                           b_silent_on_error);

        gc();

        this.b_processes_running = false;
    };

    //

    this.combineImages = function()
    {
        // This function adds the RGB channel values assigned by the user
        // to each active grayscale image and creates a new/combined color image.

        this.b_processes_running = true;

        console.writeln( "" );
        console.writeln( "<b>", "----------------------------------------------------------------------",                            "</b>" );
        console.writeln( "<b>", "combineImages()",                                                                                   "</b>" );
        console.writeln( ""                                                                                                                 );
        console.writeln( "<b>", "Colorize Method                            : " + this.colorize_method,                              "</b>" );
        console.writeln( "<b>", "Combine Method                             : " + this.combine_method,                               "</b>" );
        console.writeln( "<b>", "Linear fit colorized images before_combine : " + this.b_linear_fit_colorized_images_before_combine, "</b>" );
        console.writeln( "<b>", "Rescale Method                             : " + this.rescale_method,                               "</b>" );
        console.writeln( "<b>", "Rescale channels separately                : " + this.b_rescale_channels_independently,             "</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------",                             "</b>" );
        console.writeln( ""                                                                                                                 );

        let result = false;

        let combine_target_array = this.getCombineTargetArray_implementation();

        // Alternative version #1 of combineImages_implementation only works with MULTIPLY colorize method
        // - Fast and memory efficient, but not as flexible.  Does not support b_linear_fit_colorized_images_before_combine

        if (this.colorize_method == "MULTIPLY" && ! this.b_linear_fit_colorized_images_before_combine)
        {
            // This alternative version of combineImages_implementation() only works with MULTIPLY coloring method
            // - not as flexible, but faster

            try
            {

                result = this.combineImages_multiply_implementation(this.image_attributes_array,
                                                                    combine_target_array,
                                                                    this.colorize_method,
                                                                    this.b_save_colorized_images,
                                                                    this.b_save_colorized_images_with_autostretch,
                                                                    this.saved_colorized_images_folder,
                                                                    this.colorized_image_file_extension,
                                                                    this.combine_method,
                                                                    this.rescale_method,
                                                                    this.b_rescale_channels_independently);
            }
            catch (exception)
            {
                console.writeln("");
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                console.writeln( "<b>","Aborting: " + exception.message,                                         "</b>" );
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                console.writeln("");
                return false;
            }
        }
        else if (this.combine_method != "MEDIAN" )
        {
            // Alternative version #2 of combineImages_implementation
            // - Medium flexibility, speed and memory efficiencey,
            // but does not support Median combination

            try
            {
                result = this.combineImages_incremental_implementation(this.image_attributes_array,
                                                                       combine_target_array,
                                                                       this.colorize_method,
                                                                       this.b_save_colorized_images,
                                                                       this.b_save_colorized_images_with_autostretch,
                                                                       this.saved_colorized_images_folder,
                                                                       this.colorized_image_file_extension,
                                                                       this.b_linear_fit_colorized_images_before_combine,
                                                                       this.combine_method,
                                                                       this.rescale_method,
                                                                       this.b_rescale_channels_independently);
            }
            catch (exception)
            {
                console.writeln("");
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                console.writeln( "<b>","Aborting: " + exception.message,                                         "</b>" );
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                console.writeln("");
                return false;
            }
        }
        else
        {
            // Alternative version #3 of combineImages_implementation
            // - Most flexible (supports Median combination ),
            // but slower and less memory efficient

            try
            {
                result = this.combineImages_implementation(this.image_attributes_array,
                                                           combine_target_array,
                                                           this.colorize_method,
                                                           this.b_save_colorized_images,
                                                           this.b_save_colorized_images_with_autostretch,
                                                           this.saved_colorized_images_folder,
                                                           this.colorized_image_file_extension,
                                                           this.b_linear_fit_colorized_images_before_combine,
                                                           this.combine_method,
                                                           this.rescale_method,
                                                           this.b_rescale_channels_independently);
            }
            catch (exception)
            {
                console.writeln("");
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                console.writeln( "<b>","Aborting: " + exception.message,                                         "</b>" );
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                console.writeln("");
                return false;
            }
         }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: combineImages()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        gc();

        this.b_processes_running = false;

        return result;

    };

    //

    this.recreateReviewImage = function()
    {
        this.b_processes_running = true;

        try
        {
            this.recreateReviewImage_implementation(this.review_image_view,
                                                    this.combined_image_view,
                                                    this.b_settings_changed_since_combine,
                                                    this.b_use_saved_histogram,
                                                    this.b_auto_STF,
                                                    this.b_linked_RGB_channels,
                                                    this.review_image_midtone_adjustment,
                                                    this.review_image_shadows_clipping,
                                                    this.review_image_saturation_boost);
        }
        catch (exception)
        {
            console.writeln("");
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>","Aborting: " + exception.message,                                         "</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln("");
            return;
        }

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.saveCombinedImage = function()
    {
        this.b_processes_running = true;

        this.saveCombinedImage_implementation(this.combined_image_view,
                                              this.b_auto_STF, this.b_linked_RGB_channels);

        this.b_processes_running = false;
    };

    //

    this.saveSettings = function()
    {
        this.b_processes_running = true;

        this.saveSettings_implementation(this.b_auto_STF,  this.b_linked_RGB_channels,
                                         this.review_image_saturation_boost,
                                         this.review_image_midtone_adjustment,
                                         this.review_image_shadows_clipping,
                                         this.image_attributes_array,
                                         this.view_array,
                                         this.palette,
                                         this.b_auto_set_palette_for_known_filter,
                                         this.colorize_method,
                                         this.combine_method,
                                         this.b_linear_fit_colorized_images_before_combine,
                                         this.rescale_method,
                                         this.b_rescale_channels_independently,
                                         this.saved_settings_folder);

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.loadSettings = function()
    {
        this.b_processes_running = true;

        this.loadSettings_implementation(this.saved_settings_folder);

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.loadDefaultUserPalettes_JSON = function()
    {
        this.b_processes_running = true;

        this.loadDefaultUserPalettes_JSON_implementation(this.palettes, this.default_palettes_folder);

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.saveDefaultUserPalettes_JSON = function()
    {
        this.b_processes_running = true;

        this.saveDefaultUserPalettes_JSON_implementation(this.palettes, this.default_palettes_folder);

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.updateDefaultPalettes = function()
    {
        this.b_processes_running = true;

        this.updateDefaultPalettes_implementation(this.palettes, this.default_palettes_folder);

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.resetDefaultSystemPalettes = function()
    {
        this.b_processes_running = true;

        this.resetDefaultSystemPalettes_implementation(this.palettes, this.default_palettes_folder);

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.selectAndOpenImage = function()
    {
        this.b_processes_running = true;

        let result = this.selectAndOpenImage_implementation(this.input_images_folder,
                                                            false);  // false ==  not b_multiple_images
        this.b_processes_running = false;

        return result;
    };

    //

    this.autoOpenInputImages = function()
    {
        this.b_processes_running = true;

        this.autoOpenInputImages_implementation();

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.getCombineTargetArray = function()
    {
        this.b_processes_running = true;

        let result = this.getCombineTargetArray_implementation();

        this.b_processes_running = false;

        return result;
    };

    //

    this.getWorkingImagesCenter = function(local_new_center_normalized_x,
                                           local_new_center_normalized_y)
    {
        this.b_processes_running = true;

        let result = this.getWorkingImagesCenter_implementation(local_new_center_normalized_x,
                                                                local_new_center_normalized_y);
        this.b_processes_running = false;

        return result;
    };

    //

    this.getWorkingView = function(view,
                                   working_images_shrink_mode,
                                   center_normalized = [0.5, 0.5],
                                   name_prefix = "")
    {
        this.b_processes_running = true;

        let result = this.getWorkingView_implementation(view,
                                                        working_images_shrink_mode,
                                                        center_normalized,
                                                        name_prefix);
        this.b_processes_running = false;

        return result;
    };

    //

    this.setWorkingImagesShrinkMode = function(working_images_shrink_mode,
                                               working_images_center_normalized = [0.5, 0.5])
    {
        this.b_processes_running = true;

        this.setWorkingImagesShrinkMode_implementation(working_images_shrink_mode,
                                                       working_images_center_normalized);

        gc();   // Garbage collection

        this.b_processes_running = false;
    };

    //

    this.createWorkingImages = function(working_images_shrink_mode,
                                        working_images_center_normalized = [0.5, 0.5])
    {

        this.b_processes_running = true;

        this.createWorkingImages_implementation(working_images_shrink_mode,
                                                working_images_center_normalized);

        gc();   // Garbage collection

        this.b_processes_running = false;

    };

    //

    this.createWorkingImagesForColorizedImagePreview = function()
    {

        this.b_processes_running = true;

        this.createWorkingImagesForColorizedImagePreview_implementation();

        gc();   // Garbage collection

        this.b_processes_running = false;

    };

    //

    this.setColorSpread = function()
    {
        this.b_processes_running = true;

        this.setColorSpread_implementation();

        this.b_processes_running = false;
    };

    //

    this.updateColorizedImagePreview = function(row)
    {
        this.b_processes_running = true;

        this.updateColorizedImagePreview_implementation(row);

        this.b_processes_running = false;
    };

    //

    this.selectRow = function(row, b_clear_selection = false)
    {
        this.b_processes_running = true;

        let result = this.selectRow_implementation(row, b_clear_selection);

        this.b_processes_running = false;

        return result;
    }

    //

    this.setPalette = function(palette,
                               b_reset_image_attributes_array = false)
    {
        this.b_processes_running = true;

        this.setPalette_implementation(palette,
                                       b_reset_image_attributes_array);
        this.b_processes_running = false;
    }

    //

    this.appendView = function(view)
    {
        this.b_processes_running = true;

        this.appendView_implementation(view);

        gc();

        this.b_processes_running = false;
    };

    //

    this.addViewToRow = function(view, row = -1)
    {
        this.b_processes_running = true;

        this.addViewToRow_implementation(view, row);

        gc();

        this.b_processes_running = false;

    };

    //

    this.saveColorizedImage = function(color_image, name,
                                       b_with_auto_stretch,
                                       folder_path, file_extension = ".xisf")
    {
        this.b_processes_running = true;

        this.saveColorizedImage_implementation(color_image, name,
                                               b_with_auto_stretch,
                                               folder_path, file_extension);

        gc();

        this.b_processes_running = false;
    };

    //

   this.resetImageAttributeArrayRow = function(row)
    {
        this.b_processes_running = true;

        this.resetImageAttributeArrayRow_implementation(row)

        gc();

        this.b_processes_running = false;
    };

    //

    this.resetImageAttributeArray = function()
    {
        this.b_processes_running = true;

        this.resetImageAttributeArray_implementation()

        gc();

        this.b_processes_running = false;
    };


    //

    this.checkNewViewIsGrayscale = function(view)
    {
        this.b_processes_running = true;

        this.checkNewViewIsGrayscale_implementation(view)

        gc();

        this.b_processes_running = false;
    };

    //

    this.checkNewViewSize = function(view, row)
    {
        this.b_processes_running = true;

        this.checkNewViewSize_implementation(view, row)

        gc();

        this.b_processes_running = false;
    };

    //

    this.getFilterName = function(view,
                                  filter_names,
                                  filter_names_JWST,
                                  priority_JWST_filter_array)
    {
        this.b_processes_running = true;

        let result = this.getFilterName_implementation(view,
                                                       filter_names,
                                                       filter_names_JWST,
                                                       priority_JWST_filter_array);

        this.b_processes_running = false;

        return result;
    };

    //

    this.conditionalRescale = function(view, b_individual_channels = false)
    {
        this.b_processes_running = true;

        this.conditionalRescale_implementation(view, b_individual_channels)

        gc();

        this.b_processes_running = false;
    };

    //

    this.closeCombinedImage = function(optionalText = "")
    {
        this.b_processes_running = true;

        this.closeCombinedImage_implementation(optionalText);

        gc();

        this.b_processes_running = false;
    };

    //

    this.closeWorkingImages = function()
    {
        this.b_processes_running = true;

        this.closeWorkingImages_implementation();

        gc();

        this.b_processes_running = false;
    };

    //

    this.close = function()
    {
        this.close_implementation();
    };

    // MEMBER METHODS (IMPLEMENTATION)  ----------------------------------------

    this.initialize_implementation = function()
    {

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>initialize_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        this.initializeDiagnostics_implementation();

        this.palettes = new Palettes();

        // Create a backup of the default SYSTEM palette structure in JSON file
        this.saveDefaultSystemPalettes_JSON_implementation(this.palettes, this.default_palettes_folder);
        // Read the default USER palette from a JSON file if it exists
        this.loadDefaultUserPalettes_JSON_implementation(this.palettes, this.default_palettes_folder);

        this.filter_names =  Object.getOwnPropertyNames(this.palettes.filter_palettes);

        // Build  arrays to store image attributes and null image references,
        // one row for each potential image view
        for (let row = 0; row < NUM_VIEW_SLOTS; row++)
        {                                                        // Active  Filter Amount  RGB Values
            let newImageAttributeSet = new this.ImageAttributesSet(false, "NONE", 1.0,   1.0, 1.0, 1.0);

            this.image_attributes_array.push(newImageAttributeSet);

            this.view_array[row] = null;

            this.working_view_array[row] = null;

            this.working_view_for_colorized_images_array[row] = null;

        }

        // Create a colorized_image_preview
        // Still needs initialisation to match size of grayscale working images for colorized images
        this.colorized_image_preview = new ColorizedImage(this, "COLORIZED_INPUT_IMAGE");

         // Load and prepare Splash Image

        this.combined_image_view = this.loadSplashImage();

        let b_previous_settings_changed_since_combine = this.b_settings_changed_since_combine;
        let b_previous_auto_STF = this.b_auto_STF;
        this.setSettingsChangedFlag_implementation(false);
        this.b_auto_STF = false;
        this.recreateReviewImage();
        this.b_settings_changed_since_combine = b_previous_settings_changed_since_combine;
        this.b_auto_STF = b_previous_auto_STF;

        this.setSettingsChangedFlag_implementation(true);

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: initialize_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    };

    //

    this.updateStatus_implementation = function(text = "", color = 0xff000000)
    {
        if (this.status_bar != null && typeof this.status_bar !== 'undefined')
        {
            this.status_bar.update(text, color);
        }
    };

    //

    this.clearStatus_implementation = function()
    {
        if (this.status_bar != null && typeof this.status_bar !== 'undefined')
        {
            this.status_bar.clear();
        }
    };

    //

    this.initializeDiagnostics_implementation = function()
    {

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>initializeDiagnostics_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        // Must be initialized, even if not used at first. User can change setting later

        let function_name = "initializeDiagnostics_implementation()"

        try
        {
            console.beginLog();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: initializeDiagnostics_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

       }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: initializeDiagnostics_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
        }
    };

    //

    this.saveDiagnostics_implementation = function(b_save_diagnostics, folder_path)
    {
        let function_name = "saveDiagnostics_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>saveDiagnostics_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        if (b_save_diagnostics)
        {
            try
            {

                let log_data = console.endLog();

                let diagnostics_date = new Date;
                let diagnostic_dtg_string = format("%04d-%02d-%02d__%02dh%02dm%02ds",
                             diagnostics_date.getUTCFullYear(),
                             diagnostics_date.getUTCMonth() + 1,
                             diagnostics_date.getUTCDate(),
                             diagnostics_date.getUTCHours(),
                             diagnostics_date.getUTCMinutes(),
                             diagnostics_date.getUTCSeconds() );

                let logName = "log_" + diagnostic_dtg_string + ".log";

                saveFile(log_data, logName, folder_path);

                console.writeln( "" );
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                console.writeln( "<b>END: saveDiagnostics_implementation()</b>" );
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            }
            catch (exception)
            {
                console.writeln( "" );
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                console.writeln( "<b>END: saveDiagnostics_implementation()</b>" );
                console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

                errorProcessing_standard(function_name, exception.message);
            }
        }
    };

    //

    this.loadSplashImage_implementation = function()
    {
        let function_name = "loadSplashImage_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>loadSplashImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            let script_file_path    = #__FILE__;
            let drive  = File.extractDrive(script_file_path);
            let folder = File.extractDirectory(script_file_path);
            let root_folder  = drive + folder + "/../";
            let input_files_path = root_folder + "System" ;

            let input_files_array = FindFiles(input_files_path,
                                              [ "SPLASH_IMAGE.xisf",
                                                "SPLASH_IMAGE.fit",
                                                "SPLASH_IMAGE.fits",
                                                "SPLASH_IMAGE.tif",
                                                "SPLASH_IMAGE.jpg",
                                                "SPLASH_IMAGE.png"
                                              ],
                                              false );

            let num_input_files_subarrays = input_files_array.length;

            let total_num_input_files = 0;

            for (let i = 0; i < num_input_files_subarrays; i++)
            {
                for (let j = 0; j < input_files_array[i].length; j++)
                {
                    let file_path = input_files_array[i][j];

                    if (!File.exists(file_path ))
                    {
                         throw new Error("File not found: " + file_path);
                    }

                    let imageWin = ImageWindow.open(file_path);

                    if(imageWin.isNull)
                    {
                         throw Error( "File not found: " + file_path);

                         return null;
                    }
                    else
                    {

                        this.splash_image = imageWin[0].mainView;

                        console.writeln( "" );
                        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                        console.writeln( "<b>END: loadSplashImage_implementation()</b>" );
                        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

                        return this.splash_image;
                    }

                    console.writeln( "" );
                    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                    console.writeln( "<b>END: loadSplashImage_implementation()</b>" );
                    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

                    return null;

               }
            }

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: loadSplashImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return null;
        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: loadSplashImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);

            return null;
        }

    };

    //

    this.setSettingsChangedFlag_implementation = function(b_settings_changed_since_combine = true)
    {
        this.b_settings_changed_since_combine = b_settings_changed_since_combine;
    };

    //

    this.preprocessAllImages_implementation = function(view_array, image_attributes_array,
                                                       preprocessing_method, selected_view_row,
                                                       b_rebuild_working_images = true)
    {
        console.writeln( "" );
        console.writeln( "<b>", "----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>", "preprocessAllImages_implementation()", "</b>" );
        console.writeln( "" );
        console.writeln( "<b>", "Method : " + preprocessing_method, "</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        if (preprocessing_method == "NONE")
        {
           return;
        }

        if (preprocessing_method == "LINEAR FIT")
        {
            if (selected_view_row < 0)
            {
                (new MessageBox("<p>Linear Fit - Must select a reference image</p>",
                          "ERROR", StdIcon_Error, StdButton_Ok)).execute();
                return;
            }

            let reference_view = this.view_array[selected_view_row];

            for (let testRow = 0; testRow < NUM_VIEW_SLOTS; testRow++)
            {
                if (testRow != selected_view_row && this.image_attributes_array[testRow].b_active )
                {
                    let testView   = this.view_array[testRow];
                    let testWidth  = testView.image.width;
                    let testHeight = testView.image.height;

                    if ( reference_view.image.width !=  testWidth || reference_view.image.height !=  testHeight)
                    {
                         (new MessageBox("<p>Linear Fit - Images are not the same size!</p> <p>Cannot proceed.</p>",
                            "ERROR", StdIcon_Error, StdButton_Ok)).execute();
                         return;
                    }
                }
            }
        }

        // Cycle though all active images and preprocess
        // If one fails, continue with the others
       let result = false;

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (image_attributes_array[i].b_active)
            {
                if (preprocessing_method == "BLACKPOINT CLIP")
                {

                   this.updateStatus_implementation("Blackpoint clipping images");

                   this.blackpointClipImage(view_array[i]);
                }
               else if (preprocessing_method == "LINEAR FIT")
                {
                    this.updateStatus_implementation("Linear fitting images");

                    // All images must be grayscale
                    // false, false, true == b_protect_null_RGB_channels, b_continue_on_error, b_silent_on_error
                    result = this.linearFitImage_implementation(view_array[i],
                                                                view_array[selected_view_row],
                                                                false, false, true);

                    // A failed linear fit is more serious - so we stop
                    if (!result)
                    {
                        return;
                    }
                }
            }
        }

        this.updateStatus_implementation("Rebuilding working images");

        if (b_rebuild_working_images)
        {
            this.setWorkingImagesShrinkMode(this.working_images_shrink_mode);
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: preprocessAllImages_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    };

    //

    this.combineImages_multiply_implementation = function(image_attributes_array,
                                                          source_array,
                                                          colorize_method,
                                                          b_save_colorized_images,
                                                          b_save_colorized_images_with_autostretch,
                                                          colorized_images_folder_path,
                                                          colorized_images_file_extension,
                                                          combine_method,
                                                          rescale_method,
                                                          b_rescale_channels_independently)
    {
         // Alternative version #1 of combineImages_implementation only works with MULTIPLY colorize method
         // - Fast and memory efficient, but not as flexible.  Does not support b_linear_fit_colorized_images_before_combine

         // This function adds the RGB channel values assigned by the user
         // to each active grayscale image and creates a new/combined color image.

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>combineImages_multiply_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let function_name = "combineImages_multiply_implementation";

        let num_active_views  =  0;

        let firstActiveView = -1;

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
           if (image_attributes_array[i].b_active)
           {
              num_active_views++;
              if (num_active_views == 1)
              {
                  firstActiveView = i;
              }
           }
        }

        if (num_active_views < 1)
        {
            (new MessageBox("Need to select at least one image for MULTIPLY or MEAN combination methods, or two images for MAX !",
                            "COMBINE IMAGES - ERROR", StdIcon_Error, StdButton_Ok )).execute();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: combineImages_multiply_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }
        else if (num_active_views < 2 && combine_method.toUpperCase() == "MAX")
        {
           (new MessageBox( "Need to select at least two images for MAX combination method !",
                            "COMBINE IMAGES - ERROR", StdIcon_Error, StdButton_Ok )).execute();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: combineImages_multiply_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }

        this.updateStatus_implementation("Closing previous combined image");

        this.closeCombinedImage_implementation();

        let b_use_single_expression = false;
        let b_rescale               = false;
        let b_create_new_image      = false;
        let colorSpace              = PixelMath.prototype.RGB;
        let b_show_new_image        = false;
        let new_image_id            = "COMBINED_IMAGE";

        let expression  = "";
        let expression1 = "";
        let expression2 = "";
        let expression3 = "";
        let symbols     = "";

        let source_view = source_array[firstActiveView];

        this.updateStatus_implementation("Creating new combined image");

        let combined_image_window
                        = new ImageWindow(source_view.image.width,
                                          source_view.image.height,
                                          3,                                              // 3 channels only
                                          Math.max(32, source_view.window.bitsPerSample), // Min 32 bitsPerSample,
                                          true,                                           // floatSample
                                          true,                                           // Colour image
                                          "COMBINED_IMAGE");
        let combined_image_view = null;
        combined_image_view = combined_image_window.mainView;
        combined_image_view.visible = false;

        this.updateStatus_implementation("Building PixelMath expression");

        console.writeln("");
        console.writeln("--- Building PixelMath expression ---");
        console.writeln("");

        let num_active_views_processed  = 0;

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (image_attributes_array[i].b_active)
            {
                if (! source_array[i] || source_array[i].isNull )
                {
                    throw new Error(function_name + " - source_array[i] is Null ");
                }

                console.writeln("");

                console.writeln("Image: " + source_array[i].id  + " - color channels");

                console.writeln("");

                console.writeln("image_attributes_array[i].amount : " + image_attributes_array[i].amount
                                                                      + "  ("
                                                                      + image_attributes_array[i].amount * 100
                                                                      + ")" ) ;

                console.writeln("image_attributes_array[i].red    : " + image_attributes_array[i].red
                                                                      + "  ("
                                                                      + image_attributes_array[i].red   * 255
                                                                      + ")" ) ;
                console.writeln("image_attributes_array[i].green  : " + image_attributes_array[i].green
                                                                      + "  ("
                                                                      + image_attributes_array[i].green * 255
                                                                      + ")" ) ;
                console.writeln("image_attributes_array[i].blue   : " + image_attributes_array[i].blue
                                                                      + "  ("
                                                                      + image_attributes_array[i].blue  * 255
                                                                      + ")" ) ;
                console.writeln("");

                if (b_save_colorized_images)
                {
                    if (! source_array[i] || source_array[i].isNull )
                    {
                        throw new Error(function_name + " - source_array[i] is Null ");
                    }

                    let new_colorized_image = new ColorizedImage(this);
                    new_colorized_image.reset(source_array[i].image.width, source_array[i].image.height);

                    new_colorized_image.colorize(source_array[i],
                                                 image_attributes_array[i].red,
                                                 image_attributes_array[i].green,
                                                 image_attributes_array[i].blue,
                                                 colorize_method)

                    new_colorized_image.save(source_array[i].id,
                                             b_save_colorized_images_with_autostretch,
                                             colorized_images_folder_path,
                                             colorized_images_file_extension);

                    new_colorized_image.close();
                    new_colorized_image = null;

                }

                let combine_factor_red     = image_attributes_array[i].amount * image_attributes_array[i].red;
                let combine_factor_green   = image_attributes_array[i].amount * image_attributes_array[i].green;
                let combine_factor_blue    = image_attributes_array[i].amount * image_attributes_array[i].blue;

                num_active_views_processed++;

                if (! source_array[i] || source_array[i].isNull )
                {
                    throw new Error(function_name + " - source_array[i] is Null ");
                }

                if (num_active_views_processed == 1)
                {
                    expression =  "(" + source_array[i].id + " * " + combine_factor_red   + ")";
                    expression1 = "(" + source_array[i].id + " * " + combine_factor_green + ")";
                    expression2 = "(" + source_array[i].id + " * " + combine_factor_blue  + ")";
                }
                else
                {

                    if (combine_method.toUpperCase() == "MEAN" || combine_method.toUpperCase() == "SUM")
                    {
                        expression = expression
                             + "+"  // Modulate using 'multiplication'
                             + "(" + source_array[i].id + " * " + combine_factor_red   + ")";
                        expression1 = expression1
                              + "+"   // Modulate using 'multiplication'
                             + "(" + source_array[i].id + " * " + combine_factor_green + ")";
                        expression2 = expression2
                             + "+"   // Modulate using 'multiplication'
                             + "(" + source_array[i].id + " * " + combine_factor_blue  + ")";

                    }
                    else if (combine_method.toUpperCase() == "MAX" || combine_method.toUpperCase() == "MEDIAN")
                    {

                        expression = expression
                             + "," // Modulate using 'multiplication'
                             + "(" + source_array[i].id + " * " + combine_factor_red   + ")";
                        expression1 = expression1
                             + "," // Modulate using 'multiplication'
                             + "(" + source_array[i].id + " * " + combine_factor_green + ")";
                        expression2 = expression2
                             + "," // Modulate using 'multiplication'
                             + "(" + source_array[i].id + " * " + combine_factor_blue  + ")";

                    }
               }
            }
        }

        // If combination method is Max, add the max() operator
        // If combination method is Mean, divde by the number of active views

        if (combine_method.toUpperCase() == "MAX")
        {
            // Arithmetic maximum
            expression   = "max(" + expression   + ")";
            expression1  = "max(" + expression1  + ")";
            expression2  = "max(" + expression2  + ")";
        }
        else if (combine_method.toUpperCase() == "MEDIAN")
        {
            // Arithmetic maximum
            expression   = "med(" + expression   + ")";
            expression1  = "med(" + expression1  + ")";
            expression2  = "med(" + expression2  + ")";
        }
        else if (combine_method.toUpperCase() == "MEAN")
        {
            // Arithmetic mean
            expression  = "(" + expression  + ")" + "/" + num_active_views_processed;
            expression1 = "(" + expression1 + ")" + "/" + num_active_views_processed;
            expression2 = "(" + expression2 + ")" + "/" + num_active_views_processed;
        }

        console.writeln("");
        console.writeln("");
        console.writeln("Final PixelMath Expressions for Combination");
        console.writeln("");
        console.writeln("Expression  : " + expression);
        console.writeln("Expression1 : " + expression1);
        console.writeln("Expression2 : " + expression2);
        console.writeln("");
        console.writeln("");

        this.updateStatus_implementation("Combining and rescaling");

        console.writeln("");
        console.writeln("--- Combining and rescaling ---");
        console.writeln("");

        // Rescale or truncate as necessary

        // RESCALE or TRUNCATE
        // Normal operation is to use 'standard'' applyPixelMath
        if (rescale_method.toUpperCase() == "TRUNCATE")
        {
            applyPixelMath(combined_image_view,
                           expression, expression1, expression2, expression3,
                           b_use_single_expression,
                           symbols, false, b_create_new_image,   // rescale = false
                           new_image_id, colorSpace, b_show_new_image);
        }
        else if (rescale_method.toUpperCase()  == "RESCALE")
        {
            applyPixelMath_scary(combined_image_view,
                                 expression, expression1, expression2, expression3,
                                 b_use_single_expression,
                                 symbols, false, b_create_new_image,                 // No Rescale
                                 new_image_id, colorSpace, b_show_new_image, false); // No Truncation

            let P = new Rescale;

            if (b_rescale_channels_independently)
            {
                P.mode = Rescale.prototype.RGBK_Individual;
            }
            else
            {
                P.mode = Rescale.prototype.RGBK;
            }
            combined_image_view.beginProcess(UndoFlag_NoSwapFile);
            P.executeOn( combined_image_view );
            combined_image_view.endProcess();
        }
        else // Rescale only when required
        {
             // This version of applyPixelMath_scary
            // allows the disabling of both rescaling and truncation
            applyPixelMath_scary(combined_image_view,
                                 expression, expression1, expression2, expression3,
                                 b_use_single_expression,
                                 symbols, false, b_create_new_image,                 // No Rescale
                                 new_image_id, colorSpace, b_show_new_image, false); // No Truncation

            combined_image_view = this.conditionalRescale_implementation(combined_image_view,
                                                                         b_rescale_channels_independently);
        }

        this.combined_image_view = combined_image_view;
        combined_image_view = null;

        this.setSettingsChangedFlag_implementation(false);

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: combineImages_multiply_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return true;
    };

    //

    this.combineImages_incremental_implementation = function(image_attributes_array,
                                                             source_array,
                                                             colorize_method,
                                                             b_save_colorized_images,
                                                             b_save_colorized_images_with_autostretch,
                                                             colorized_images_folder_path,
                                                             colorized_images_file_extension,
                                                             b_linear_fit_colorized_images_before_combine,
                                                             combine_method,
                                                             rescale_method,
                                                             b_rescale_channels_independently)
    {
         // Alternative version #2 of combineImages_implementation
         // - Medium flexibility, speed and memory efficiencey,
         // but does not support Median combination or b_linear_fit_colorized_images_before_combine

        // This function adds the RGB channel values assigned by the user
        // to each active grayscale image and creates a new/combined color image.

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>combineImages_incremental_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let function_name = "combineImages_incremental_implementation";

        let num_active_views  =  0;

        let firstActiveView = -1;

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (image_attributes_array[i].b_active)
            {
                num_active_views++;
                if (num_active_views == 1)
                {
                    firstActiveView = i;
                }
            }
        }

        if (num_active_views < 1)
        {
           (new MessageBox( "Need to select at least one image for SUM or MEAN combination methods, or two images for MAX !",
                            "COMBINE IMAGES - ERROR", StdIcon_Error, StdButton_Ok )).execute();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>","END: combineImages_incremental_implementation()", "</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }
        else if (num_active_views < 2 && combine_method.toUpperCase() == "MAX")
        {
           (new MessageBox( "Need to select at least two images for MAX combination method !",
                            "COMBINE IMAGES - ERROR", StdIcon_Error, StdButton_Ok )).execute();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>","END: combineImages_incremental_implementation()", "</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }

        this.updateStatus_implementation("Closing previous combined image");

        this.closeCombinedImage_implementation();

        let b_use_single_expression = true;
        let b_rescale               = false;
        let b_create_new_image      = false;
        let colorSpace              = PixelMath.prototype.RGB;
        let b_show_new_image        = false;
        let new_image_id            = "COMBINED_IMAGE";

        let expression  = "";
        let expression1 = "";
        let expression2 = "";
        let expression3 = "";
        let symbols     = "";

        let source_view = source_array[firstActiveView];

        this.updateStatus_implementation("Creating new combined image");

        console.writeln("");
        console.writeln("--- Creating new combined image ---");
        console.writeln("");

        let combined_image_window
                        = new ImageWindow(source_view.image.width,
                                          source_view.image.height,
                                          3,                                              // 3 channels only
                                          Math.max(32, source_view.window.bitsPerSample), // Min 32 bitsPerSample,
                                          true,                                           // floatSample
                                          true,                                           // Colour image
                                          "COMBINED_IMAGE");

        let combined_image_view = combined_image_window.mainView;
        combined_image_view.visible = false;

        let num_active_views_processed  = 0;

        this.updateStatus_implementation("Generating colorized images for incremental processing");

        console.writeln("");
        console.writeln("--- Generating colorized images for incremental processing ---");
        console.writeln("");

        let colorized_image      = new ColorizedImage(this);
        let reference_view       = null;
        let reference_view_index = this.selected_view_row;

        let b_skip_linear_fits = false;

        // Special case
        // To linear fit in an incremental way, the first image to be processed needs to the reference image
        if (b_linear_fit_colorized_images_before_combine && ! b_skip_linear_fits)
        {
            if (reference_view_index >= 0)
            {
                if (source_array[reference_view_index] && ! source_array[reference_view_index].isNull)
                {
                    this.updateStatus_implementation("Colorizing : " + source_array[reference_view_index].id);
                }
                else
                {
                    throw new Error(function_name + " - source_array[reference_view_index] is Null");
                }

                console.writeln("");

                console.writeln("Image: " + source_array[reference_view_index].id  + " - color channels");

                console.writeln("");

                console.writeln("image_attributes_array[reference_view_index].amount : " + image_attributes_array[reference_view_index].amount
                                                                      + "  ("
                                                                      + image_attributes_array[reference_view_index].amount * 100
                                                                      + ")" ) ;

                console.writeln("image_attributes_array[reference_view_index].red    : " + image_attributes_array[reference_view_index].red
                                                                      + "  ("
                                                                      + image_attributes_array[reference_view_index].red   * 255
                                                                      + ")" ) ;
                console.writeln("image_attributes_array[reference_view_index].green  : " + image_attributes_array[reference_view_index].green
                                                                      + "  ("
                                                                      + image_attributes_array[reference_view_index].green * 255
                                                                      + ")" ) ;
                console.writeln("image_attributes_array[reference_view_index].blue   : " + image_attributes_array[reference_view_index].blue
                                                                      + "  ("
                                                                      + image_attributes_array[reference_view_index].blue  * 255
                                                                      + ")" ) ;
                num_active_views_processed++;


                if (source_array[reference_view_index] && ! source_array[reference_view_index].isNull)
                {
                    colorized_image.reset(source_array[reference_view_index].image.width, source_array[reference_view_index].image.height);
                }
                else
                {
                   throw new Error(function_name + " - source_array[reference_view_index] is Null");
                }

                if (source_array[reference_view_index] && ! source_array[reference_view_index].isNull)
                {

                    colorized_image.colorize(source_array[reference_view_index],
                                             image_attributes_array[reference_view_index].red,
                                             image_attributes_array[reference_view_index].green,
                                             image_attributes_array[reference_view_index].blue,
                                             colorize_method)
                }
                else
                {
                    throw new Error(function_name + " - source_array[reference_view_index] is Null");
                }

                reference_view = createDuplicate(colorized_image.view, "LINEAR_FIT_REFERENCE");

                if (b_save_colorized_images && source_array[reference_view_index] && ! source_array[reference_view_index].isNull)
                {

                    colorized_image.save(source_array[reference_view_index].id,
                                         b_save_colorized_images_with_autostretch,
                                         colorized_images_folder_path,
                                         colorized_images_file_extension)
                }

                // For the first active view only, simply set the combined_image_view as the
                // colorised image multiplied by the 'amount'

                expression =  colorized_image.view.id + " * " + image_attributes_array[reference_view_index].amount;

                applyPixelMath(combined_image_view,
                               expression, expression1, expression2, expression3,
                               b_use_single_expression,
                               symbols, false, b_create_new_image,   // rescale = false
                               new_image_id, colorSpace, b_show_new_image);

            }
        }

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (image_attributes_array[i].b_active && !(b_linear_fit_colorized_images_before_combine && reference_view_index == i))
            {
                if (source_array[i] && ! source_array[i].isNull)
                {
                    this.updateStatus_implementation("Colorizing : " + source_array[i].id);
                }
                else
                {
                    throw new Error(function_name + " - source_array[i] is Null");
                }

                console.writeln("");

                console.writeln("Image: " + source_array[i].id  + " - color channels");

                console.writeln("");

                console.writeln("image_attributes_array[i].amount : " + image_attributes_array[i].amount
                                                                      + "  ("
                                                                      + image_attributes_array[i].amount * 100
                                                                      + ")" ) ;

                console.writeln("image_attributes_array[i].red    : " + image_attributes_array[i].red
                                                                      + "  ("
                                                                      + image_attributes_array[i].red   * 255
                                                                      + ")" ) ;
                console.writeln("image_attributes_array[i].green  : " + image_attributes_array[i].green
                                                                      + "  ("
                                                                      + image_attributes_array[i].green * 255
                                                                      + ")" ) ;
                console.writeln("image_attributes_array[i].blue   : " + image_attributes_array[i].blue
                                                                      + "  ("
                                                                      + image_attributes_array[i].blue  * 255
                                                                      + ")" ) ;
                console.writeln("");

                num_active_views_processed++;


                if (source_array[i] && ! source_array[i].isNull)
                {
                    colorized_image.reset(source_array[i].image.width, source_array[i].image.height);
                }
                else
                {
                   throw new Error(function_name + " - source_array[i] is Null ");
                }

                if (source_array[i] && ! source_array[i].isNull)
                {
                    colorized_image.colorize(source_array[i],
                                             image_attributes_array[i].red,
                                             image_attributes_array[i].green,
                                             image_attributes_array[i].blue,
                                             colorize_method)
                }
                else
                {
                    throw new Error(function_name + " - source_array[i]  is Null ");
                }

                if (b_save_colorized_images && source_array[i] && ! source_array[i].isNull)
                {
                    colorized_image.save(source_array[i].id,
                                         b_save_colorized_images_with_autostretch,
                                         colorized_images_folder_path,
                                         colorized_images_file_extension)
                }

                // Linear Fit if required - and only if a Reference image is selected

                if (b_linear_fit_colorized_images_before_combine && ! b_skip_linear_fits)
                {
                    if (reference_view_index >= 0)
                    {
                        let result = false ;

                        this.updateStatus_implementation("Linear fitting colorized image");

                        console.writeln("");
                        console.writeln("--- Linear fitting colorized image ---");
                        console.writeln("");

                        let target_view  = colorized_image.view;

                        // All images are RGB color
                        // true, true, true == b_protect_null_RGB_channels, b_continue_on_error, b_silent_on_error

                        if (target_view    && ! target_view.isNull &&
                            reference_view && ! reference_view.isNull)
                        {
                            result = this.linearFitImage_implementation(target_view, reference_view, true, true, true);
                        }
                        else
                        {
                            result = false;
                        }
                        if (!result)
                        {
                             console.writeln("");
                             console.writeln("A linear fit has failed. Check diagnostics log for details.  Skipping new linear fits");
                             console.writeln("");

                             (new MessageBox("<p>A linear fit has failed. Check diagnostics log for details.</p> <p>Will continue processing, skipping new linear fits</p>",
                                             "WARNING", StdIcon_Warning, StdButton_Ok)).execute();

                             // Setting b_linear_fit_colorized_images_before_combine to false is not an option
                             // because it would require knowledge of, and ability to change, Dialog settings
                             // So set a local flag
                             b_skip_linear_fits = true;
                         }
                    }
                    else
                    {
                        (new MessageBox("<p>Linear fit has failed. Check diagnostics log for details.</p> <p>Will continue processing, skipping linear fits</p>",
                                        "WARNING", StdIcon_Warning, StdButton_Ok)).execute();
                    }
                }

                this.updateStatus_implementation("Combining incrementally");

                console.writeln("");
                console.writeln("--- Combining incrementally ---");
                console.writeln("");

                if (num_active_views_processed == 1)
                {
                    // For the first active view only, simply set the combined_image_view as the
                    // colorised image multiplied by the 'amount'

                    expression =  colorized_image.view.id + " * " + image_attributes_array[i].amount;

                    applyPixelMath(combined_image_view,
                                   expression, expression1, expression2, expression3,
                                   b_use_single_expression,
                                   symbols, false, b_create_new_image,   // rescale = false
                                   new_image_id, colorSpace, b_show_new_image);

                }
                else
                {
                    // For all other active views ...
                    // Build the combine expression according to the intended combination method;
                    // which determines operator or comma delimited expressions

                    expression =  "(" + colorized_image.view.id + " * " + image_attributes_array[i].amount   + ")";

                    if (combine_method.toUpperCase() == "MEAN" || combine_method.toUpperCase() == "SUM")
                    {

                        expression = combined_image_view.id + "+" + expression

                        applyPixelMath_scary(combined_image_view,
                                             expression, expression1, expression2, expression3,
                                             b_use_single_expression,
                                             symbols, false, b_create_new_image,                 // No Rescale
                                             new_image_id, colorSpace, b_show_new_image, false); // No Truncation
                    }
                    else if (combine_method.toUpperCase() == "MAX" )
                    {
                        expression = "max( " + combined_image_view.id + "," + expression + " )";
                        applyPixelMath(combined_image_view,
                                       expression, expression1, expression2, expression3,
                                       b_use_single_expression,
                                       symbols, false, b_create_new_image,   // false == no rescale
                                       new_image_id, colorSpace, b_show_new_image);
                    }
                }
            }
        }

        // If combination method is Mean, divde by the number of active views

        if (combine_method.toUpperCase() == "MEAN")
        {
            // Arithmetic mean
            expression  =  combined_image_view.id + "/" + num_active_views_processed;

            applyPixelMath(combined_image_view,
                           expression, expression1, expression2, expression3,
                           b_use_single_expression,
                           symbols, false, b_create_new_image,   // false == no rescale
                           new_image_id, colorSpace, b_show_new_image);
        }

        this.updateStatus_implementation("Rescaling/truncating as necessary");

        console.writeln("");
        console.writeln("--- Rescaling/truncating as necessary ---");
        console.writeln("");

        // RESCALE or TRUNCATE
        if (rescale_method.toUpperCase() == "TRUNCATE") // Only needed for SUM combination
        {
           expression = combined_image_view.id;

           applyPixelMath(combined_image_view,
                          expression, expression1, expression2, expression3,
                          b_use_single_expression,
                          symbols, false, b_create_new_image,   // rescale = false
                          new_image_id, colorSpace, b_show_new_image);
        }
        else if (rescale_method.toUpperCase()  == "RESCALE")
        {

            let P = new Rescale;

            if (b_rescale_channels_independently)
            {
                P.mode = Rescale.prototype.RGBK_Individual;
            }
            else
            {
               P.mode = Rescale.prototype.RGBK;
            }
            combined_image_view.beginProcess(UndoFlag_NoSwapFile);
            P.executeOn( combined_image_view );
            combined_image_view.endProcess();
        }
        else  // Rescale only when required
        {
            combined_image_view = this.conditionalRescale_implementation(combined_image_view,
                                                                         b_rescale_channels_independently);
        }

        this.combined_image_view = combined_image_view;
        combined_image_view      = null;

        this.updateStatus_implementation("Closing working colorized image");

        console.writeln("");
        console.writeln("--- Closing working colorized image ---");
        console.writeln("");

        if (colorized_image.view)
        {
            colorized_image.close();
            colorized_image = null;
        }

        if (reference_view)
        {
            closeWindow(reference_view);
            reference_view = null;
        }

        this.setSettingsChangedFlag_implementation(false);

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: combineImages_incremental_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return true;

    };

    //

    this.combineImages_implementation = function(image_attributes_array,
                                                 source_array,
                                                 colorize_method,
                                                 b_save_colorized_images,
                                                 b_save_colorized_images_with_autostretch,
                                                 colorized_images_folder_path,
                                                 colorized_images_file_extension,
                                                 b_linear_fit_colorized_images_before_combine,
                                                 combine_method,
                                                 rescale_method,
                                                 b_rescale_channels_independently)
    {
         // Alternative version #3 of combineImages_implementation
         // - Most flexible (supports Median combination and b_linear_fit_colorized_images_before_combine),
         // but slower and less memory efficient

        // This function adds the RGB channel values assigned by the user
        // to each active grayscale image and creates a new/combined color image.

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>combineImages_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let function_name = "combineImages_implementation";

        let num_active_views  =  0;

        let firstActiveView = -1;

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (image_attributes_array[i].b_active)
            {
                num_active_views++;
                if (num_active_views == 1)
                {
                    firstActiveView = i;
                }
            }
        }

        if (num_active_views < 1)
        {
           (new MessageBox( "Need to select at least one image for SUM or MEAN combination methods, or two images for MAX !",
                            "COMBINE IMAGES - ERROR", StdIcon_Error, StdButton_Ok )).execute();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>","END: combineImages_implementation()", "</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }
        else if (num_active_views < 2 && combine_method.toUpperCase() == "MAX")
        {
           (new MessageBox( "Need to select at least two images for MAX combination method !",
                            "COMBINE IMAGES - ERROR", StdIcon_Error, StdButton_Ok )).execute();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>","END: combineImages_implementation()", "</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }

        this.updateStatus_implementation("Closing previous combined image");

        this.closeCombinedImage_implementation();

        let b_use_single_expression = true;
        let b_rescale               = false;
        let b_create_new_image      = false;
        let colorSpace              = PixelMath.prototype.RGB;
        let b_show_new_image        = false;
        let new_image_id            = "COMBINED_IMAGE";

        let expression  = "";
        let expression1 = "";
        let expression2 = "";
        let expression3 = "";
        let symbols     = "";

        let source_view = source_array[firstActiveView];

        this.updateStatus_implementation("Creating new combined image");

        let combined_image_window
                        = new ImageWindow(source_view.image.width,
                                          source_view.image.height,
                                          3,                                              // 3 channels only
                                          Math.max(32, source_view.window.bitsPerSample), // Min 32 bitsPerSample,
                                          true,                                           // floatSample
                                          true,                                           // Colour image
                                          "COMBINED_IMAGE");

        let combined_image_view = null;
        combined_image_view     = combined_image_window.mainView;
        combined_image_view.visible = false;

        let num_active_views_processed  = 0;
        combined_image_view.visible = false;

        this.updateStatus_implementation("Generating colorized images");

        console.writeln("");
        console.writeln("--- Generating colorized images ---");
        console.writeln("");

        let colorized_image_array = [];
        // Just to make sure :-)
        for (let row = 0; row < NUM_VIEW_SLOTS; row++)
        {
            colorized_image_array[row] = null;
        }

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (image_attributes_array[i].b_active)
            {
                if (source_array[i] && ! source_array[i].isNull)
                {
                    this.updateStatus_implementation("Colorizing : " + source_array[i].id);
                }
                else
                {
                    throw new Error(function_name + " - source_array[i] is Null");
                }

                console.writeln("");

                console.writeln("Image: " + source_array[i].id  + " - color channels");

                console.writeln("");

                console.writeln("image_attributes_array[i].amount : " + image_attributes_array[i].amount
                                                                      + "  ("
                                                                      + image_attributes_array[i].amount * 100
                                                                      + ")" ) ;

                console.writeln("image_attributes_array[i].red    : " + image_attributes_array[i].red
                                                                      + "  ("
                                                                      + image_attributes_array[i].red   * 255
                                                                      + ")" ) ;
                console.writeln("image_attributes_array[i].green  : " + image_attributes_array[i].green
                                                                      + "  ("
                                                                      + image_attributes_array[i].green * 255
                                                                      + ")" ) ;
                console.writeln("image_attributes_array[i].blue   : " + image_attributes_array[i].blue
                                                                      + "  ("
                                                                      + image_attributes_array[i].blue  * 255
                                                                      + ")" ) ;
                console.writeln("");

                let new_colorized_image = new ColorizedImage(this);

                if (source_array[i] && ! source_array[i].isNull)
                {
                    new_colorized_image.reset(source_array[i].image.width, source_array[i].image.height);
                }
                else
                {
                    throw new Error(function_name + " - source_array[i] is Null ");
                }

                if (source_array[i] && ! source_array[i].isNull)
                {
                    new_colorized_image.colorize(source_array[i],
                                                 image_attributes_array[i].red,
                                                 image_attributes_array[i].green,
                                                 image_attributes_array[i].blue,
                                                 colorize_method)
                }
                else
                {
                    throw new Error(function_name + " - source_array[i] is Null ");
                }

                colorized_image_array[i] = new_colorized_image; // Maintain same indexes as for view_array[] etc

                if (b_save_colorized_images && source_array[i] && ! source_array[i].isNull)
                {
                    new_colorized_image.save(source_array[i].id,
                                             b_save_colorized_images_with_autostretch,
                                             colorized_images_folder_path,
                                             colorized_images_file_extension)
                }

                num_active_views_processed++;
            }
        }

        // Linear Fit if required - and only if a Reference image is selected

        if (b_linear_fit_colorized_images_before_combine )
        {
            let reference_view_index = this.selected_view_row;

            if (reference_view_index >= 0)
            {
                let result = false ;

                this.updateStatus_implementation("Linear fitting colorized images");

                console.writeln("");
                console.writeln("--- Linear fitting colorized images ---");
                console.writeln("");

                for (let i = 0; i < NUM_VIEW_SLOTS; i++)
                {

                    if (image_attributes_array[i].b_active)
                    {
                        console.writeln("");
                        console.writeln("Linear Fit Reference");

                        let target_view    = colorized_image_array[i].view;
                        let reference_view = colorized_image_array[reference_view_index].view;

                        if (target_view && ! target_view.isNull)
                        {
                            this.updateStatus_implementation(target_view.id);
                        }

                        // All images are RGB color
                        // true, true, true == b_protect_null_RGB_channels, b_continue_on_error, b_silent_on_error
                        if (target_view    && ! target_view.isNull &&
                            reference_view && ! reference_view.isNull)
                        {
                            result = this.linearFitImage_implementation(target_view, reference_view, true, true, true);
                        }
                        else
                        {
                            result = false;
                        }

                        if (!result)
                        {
                            console.writeln("");
                            console.writeln("A linear fit has failed. Check diagnostics log for details. Skipping new linear fits");
                            console.writeln("");

                            (new MessageBox("<p>A linear fit has failed. Check diagnostics log for details.</p> <p>Will continue processing, skipping new linear fits</p>",
                                             "WARNING", StdIcon_Warning, StdButton_Ok)).execute();

                            break;
                        }
                    }
                }
            }
            else
            {
                (new MessageBox("<p>Linear fit has failed. Check diagnostics log for details.</p> <p>Will continue processing, skipping linear fits</p>",
                                "WARNING", StdIcon_Warning, StdButton_Ok)).execute();
            }
        }

        // Build the combine expression according to the intended combination method;
        // which determines operator or comma delimited lists
        // and modulate their respective base colors

        this.updateStatus_implementation("Building PixelMath expression");

        console.writeln("");
        console.writeln("--- Building PixelMath expression ---");
        console.writeln("");

        let active_view_index = -1 ;

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (image_attributes_array[i].b_active)
            {
                active_view_index++;

                if (active_view_index == 0)
                {
                    expression =  "(" + colorized_image_array[i].view.id + " * " + image_attributes_array[i].amount + ")";
                }
                else
                {
                    if (combine_method.toUpperCase() == "MEAN" || combine_method.toUpperCase() == "SUM")
                    {
                        expression = expression
                         + "+"
                         + "(" + colorized_image_array[i].view.id + " * " + image_attributes_array[i].amount   + ")";
                    }
                    else if (combine_method.toUpperCase() == "MAX" || combine_method.toUpperCase() == "MEDIAN")
                    {

                       expression = expression
                         + ","
                         + "(" + colorized_image_array[i].view.id + " * " + image_attributes_array[i].amount    + ")";
                    }
               }
            }
        }

        // Now execute the combination process

        this.updateStatus_implementation("Combining and rescaling images");

        console.writeln("");
        console.writeln("--- Combining and rescaling images ---");
        console.writeln("");

        // If combination method is Max, add the max() operator
        // If combination method is Mean, divde by the number of active views

        if (combine_method.toUpperCase() == "MAX")
        {
            // Arithmetic maximum
            expression   = "max(" + expression   + ")";
        }
        if (combine_method.toUpperCase() == "MEDIAN")
        {
            // Arithmetic maximum
            expression   = "med(" + expression   + ")";
        }
        else if (combine_method.toUpperCase() == "MEAN")
        {
            // Arithmetic mean
            expression  = "(" + expression  + ")" + "/" + num_active_views_processed;
        }

        console.writeln("");
        console.writeln("");
        console.writeln("Final PixelMath Expressions for Combination");
        console.writeln("");
        console.writeln("Expression  : " + expression);
        console.writeln("Expression1 : " + expression1);
        console.writeln("Expression2 : " + expression2);
        console.writeln("");
        console.writeln("");

        // RESCALE or TRUNCATE
        // Normal operation is to use 'standard'' applyPixelMath
        if (rescale_method.toUpperCase() == "TRUNCATE")
        {
           applyPixelMath(combined_image_view,
                          expression, expression1, expression2, expression3,
                          b_use_single_expression,
                          symbols, false, b_create_new_image,   // rescale = false
                          new_image_id, colorSpace, b_show_new_image);
        }
        else if (rescale_method.toUpperCase()  == "RESCALE")
        {
            applyPixelMath_scary(combined_image_view,
                                 expression, expression1, expression2, expression3,
                                 b_use_single_expression,
                                 symbols, false, b_create_new_image,                 // No Rescale
                                 new_image_id, colorSpace, b_show_new_image, false); // No Truncation

            let P = new Rescale;

            if (b_rescale_channels_independently)
            {
                P.mode = Rescale.prototype.RGBK_Individual;
            }
            else
            {
               P.mode = Rescale.prototype.RGBK;
            }
            combined_image_view.beginProcess(UndoFlag_NoSwapFile);
            P.executeOn( combined_image_view );
            combined_image_view.endProcess();
        }
        else  // Rescale only when required
        {
            // This version of applyPixelMath_scary
            // allows the disabling of both rescaling and truncation
            applyPixelMath_scary(combined_image_view,
                                 expression, expression1, expression2, expression3,
                                 b_use_single_expression,
                                 symbols, false, b_create_new_image,                 // No Rescale
                                 new_image_id, colorSpace, b_show_new_image, false); // No Truncation

            combined_image_view = this.conditionalRescale_implementation(combined_image_view,
                                                                         b_rescale_channels_independently);
        }


        this.combined_image_view = combined_image_view;
        combined_image_view = null;

        this.updateStatus_implementation("Closing working colorized images");

        console.writeln("");
        console.writeln("--- Closing working colorized images ---");
        console.writeln("");

        // Clean up - close colorized images
        for (let j = 0; j < colorized_image_array.length; j++)
        {
            if (colorized_image_array[j])
            {
                colorized_image_array[j].close();
                colorized_image_array[j] = null;
            }
        }
        colorized_image_array = null;

        this.setSettingsChangedFlag_implementation(false);

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: combineImages_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return true;
    };

    //

    this.recreateReviewImage_implementation = function(review_image_view,
                                                       combined_image_view,
                                                       b_settings_changed_since_combine,
                                                       b_use_saved_histogram, b_auto_STF, b_linked_RGB_channels,
                                                       review_image_midtone_adjustment,
                                                       review_image_shadows_clipping,
                                                       review_image_saturation_boost)
    {

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>recreateReviewImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let function_name = "recreateReviewImage_implementation";

        if (b_settings_changed_since_combine)
        {

           (new MessageBox(
              "<p>You have changed at least one setting.</p>" + "<p>The Review Image will be refreshed, but you will need (re)select COMBINE before the image will reflect your new settings.</p>",
              "WARNING",
               StdIcon_Warning,
               StdButton_Ok
             )).execute();
        }

        if (review_image_view && ! review_image_view.isNull)
        {
            closeWindow(review_image_view);
        }
        review_image_view = null;

        let new_review_image_view = null;

        if (combined_image_view && ! combined_image_view.isNull)
        {
            // Special case
            if (this.status_bar != null
                && typeof this.updateStatus_implementation !== 'undefined'
                && this.combined_image_view.id.search("SPLASH_IMAGE") < 0)
            {
                this.updateStatus_implementation("Creating new ReviewImage");
            }

            new_review_image_view = createDuplicate(combined_image_view,
                                                   "REVIEW_IMAGE", false);

            if (b_use_saved_histogram)
            {
                // Special case
                if (this.status_bar != null
                    && typeof this.updateStatus_implementation !== 'undefined'
                    && this.combined_image_view.id.search("SPLASH_IMAGE") < 0)
                {
                    this.updateStatus_implementation("Applying save stretch settings");
                }

                applyHistogramTransformation(new_review_image_view, this.saved_H);
            }
            else if (b_auto_STF)
            {
                console.writeln("");
                console.writeln("b_auto_STF : " + b_auto_STF);
                console.writeln("");

                // Special case
                if (this.status_bar != null
                    && typeof this.updateStatus_implementation !== 'undefined'
                    && this.combined_image_view.id.search("SPLASH_IMAGE") < 0)
                {
                    this.updateStatus_implementation("Applying auto stretch settings");
                }

                // function setAutoSTFparameters(view,
                //                               shadows_clipping  = SHADOWS_CLIP ,
                //                               target_background = TARGET_BKG,
                //                               b_rgb_linked      = RGB_LINK)

                setAutoSTFparameters(new_review_image_view,
                                     review_image_shadows_clipping,
                                     TARGET_BKG, b_linked_RGB_channels );

                applyHistogramTransformation(new_review_image_view, g_helper_data.H);
            }
            else
            {
                // Do not apply applyHistogramTransformation(),
                // but do reset parameters
                g_helper_data.STF =
                [ // c0, c1, m, r0, r1
                     [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                     [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                     [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                     [0.00000, 1.00000, 0.50000, 0.00000, 1.00000]
                ];
                g_helper_data.H =
                [ // c0, m, c1, r0, r1
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000]
                ];
            }

            // Special case
            if (this.status_bar != null
                && typeof this.updateStatus_implementation !== 'undefined'
                && this.combined_image_view.id.search("SPLASH_IMAGE") < 0)
            {
                this.updateStatus_implementation("Applying midtone adjustment");
            }

            // Apply midtone adjustment
            let new_histo_midtone = review_image_midtone_adjustment;
            if (review_image_midtone_adjustment != 0.5)
            {
                if (new_histo_midtone <=  0)
                {
                    new_histo_midtone = 0.0000001 ;
                }
                else if (new_histo_midtone >= 1)
                {
                    new_histo_midtone = 0.9999999 ;
                }
            }
            let H =
                [ // c0, m, c1, r0, r1
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, new_histo_midtone, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000]
                ];

            applyHistogramTransformation(new_review_image_view, H);

            // Special case
            if (this.status_bar != null
                && typeof this.updateStatus_implementation !== 'undefined'
                && this.combined_image_view.id.search("SPLASH_IMAGE") < 0)
            {
                this.updateStatus_implementation("Applying saturation adjustment");
            }

            // Apply saturation adjustment
            if (review_image_saturation_boost != 0)
            {
                let P = new ColorSaturation;
                P.HS =
                [ // x, y
                    [0.00000, review_image_saturation_boost],
                    [1.00000, review_image_saturation_boost]
                ];
                P.HSt = ColorSaturation.prototype.AkimaSubsplines;
                P.hueShift = 0.000;

                executeOn(new_review_image_view, P);
            }
        }

        this.review_image_view = new_review_image_view;

        new_review_image_view = null;

        gc(); // one of the few times gc() is called from implementaion // Probably not necessary

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: recreateReviewImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    };

    //

    this.saveCombinedImage_implementation = function(combined_image_view,
                                                     b_auto_STF = true, b_linked_RGB_channels = true,
                                                     name = "DEFAULT", folder_path = "DEFAULT", file_extension = "DEFAULT")
    {

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>saveCombinedImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );


        if (!combined_image_view || combined_image_view.id == "SPLASH_IMAGE")
        {
            (new MessageBox("<p>Nothing saved - you have not created a combined image yet</p>",
                            "NOTE",
                            StdIcon_Warning,
                            StdButton_Ok
                            )).execute();
            return;
        }

        if (this.b_settings_changed_since_combine)
        {

            let result = (new MessageBox(
              "<p>You have changed at least one setting.</p>" + "<p>If you continue, you will only save the previous combined image.</p>"+
              "<p>If you were using a reduced-size Working Image at the time, that is the copy that will be saved.</p>",
                  "Continue?",
                  StdIcon_Question,
                  StdButton_Yes, StdButton_No
            )).execute();

            if (result == StdButton_No)
            {
               return;
            }
        }
        else if (this.working_images_shrink_mode !=   "FULL SIZE")
        {
            let result = (new MessageBox(
              "<p>You have been using a reduced-size Working Image, and that is the image that will be saved.</p>",
                  "Continue?",
                  StdIcon_Question,
                  StdButton_Ok, StdButton_Cancel
            )).execute();

            if (result == StdButton_Cancel)
            {
               return;
            }
        }

        if (name == "DEFAULT" || name == "")
        {
            name = combined_image_view.id;
        }
        if (folder_path == "DEFAULT")
        {
            folder_path = this.saved_combined_images_folder;
        }
        if (file_extension == "DEFAULT")
        {
            file_extension = ".xisf";
        }

        let duplicate = createDuplicate(combined_image_view, name);

        // If the 'b_auto_STF'is true,
        //     an Auto STF stretch will be applied to any finalized image,
        //     and a permanent stretch will be applied to the COMBINED_IMAGE.

        if (b_auto_STF)
        {
            applyAutoScreenTransferFunction(duplicate, b_linked_RGB_channels);
        }

        // SAVE .....

        let full_file_name = name + file_extension;

        if (!File.directoryExists(folder_path))
        {
            File.createDirectory(folder_path);
        }

        let save_dialog = new SaveFileDialog;
        save_dialog.caption = "Save Combined Image";

        save_dialog.filters = [ [ "xisf Files",      "*.xisf"        ],
                                [ "fit Files",       "*.fit, *.fits" ],
                                [ "jpg Files",       "*.jpg, *.jpeg" ],
                                [ "png Files",       "*.png"         ],
                                [ "All image Files", "*.xisf, *.fit, *.tif, *.jpg, *.png"]
                              ];
        save_dialog.initialPath = folder_path + "/" + full_file_name;;
        save_dialog.overwritePrompt = true;
        if (save_dialog.execute())
        {
            save_dialog.fileName;
            // Boolean ImageWindow.saveAs( String filePath[, Boolean queryOptions=true[, Boolean allowMessages=true[, Boolean strict=true[, Boolean verifyOverwrite=true[, String formatHints='']]]]] )
            duplicate.window.saveAs(save_dialog.fileName, false, false, false, false);
        };

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: saveCombinedImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    };

    //

    this.saveSettings_implementation = function(b_auto_STF,  b_linked_RGB_channels,
                                                review_image_saturation_boost,
                                                review_image_midtone_adjustment,
                                                review_image_shadows_clipping,
                                                image_attributes_array, view_array,
                                                palette,
                                                b_auto_set_palette_for_known_filter,
                                                colorize_method,
                                                combine_method,
                                                b_linear_fit_colorized_images_before_combine,
                                                rescale_method,
                                                b_rescale_channels_independently,
                                                folder_path)
    {

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>saveSettings_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let function_name = "saveSettings_implementation()"

        try
        {

            console.writeln("");
            console.writeln("--- SAVING SETTINGS ---");
            console.writeln("");

            if (!File.directoryExists(folder_path))
            {
                File.createDirectory(folder_path);
            }

            let settings_date = new Date;
            let settings_dtg_string = format("%04d-%02d-%02d__%02dh%02dm%02ds",
                             settings_date.getUTCFullYear(),
                             settings_date.getUTCMonth() + 1,
                             settings_date.getUTCDate(),
                             settings_date.getUTCHours(),
                             settings_date.getUTCMinutes(),
                             settings_date.getUTCSeconds() );

            let settings_file_name = "settings_" + settings_dtg_string + ".json";

            let settings_file_path = folder_path + "/" + settings_file_name;

            let file = File.createFileForWriting( settings_file_path );

            let save_data =
            {
                b_auto_STF                                   : b_auto_STF,
                b_linked_RGB_channels                        : b_linked_RGB_channels,
                review_image_saturation_boost                : review_image_saturation_boost,
                review_image_midtone_adjustment              : review_image_midtone_adjustment,
                review_image_shadows_clipping                : review_image_shadows_clipping,
                colorize_method                              : colorize_method,
                combine_method                               : combine_method,
                b_linear_fit_colorized_images_before_combine : b_linear_fit_colorized_images_before_combine,
                rescale_method                               : rescale_method,
                b_rescale_channels_independently             : b_rescale_channels_independently,
                palette                                      : palette,
                b_auto_set_palette_for_known_filter          : b_auto_set_palette_for_known_filter,
                image_attributes_array                       : image_attributes_array,
            }

            file.write(ByteArray.stringToUTF8(JSON.stringify(save_data, null, 4)));

            file.close();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: saveSettings_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: saveSettings_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
        }

    };

    //

    this.loadSettings_implementation = function(folder_path)
    {

        let function_name = "loadSettings_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>loadSettings_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {

            let setings_file_name = "";

            if (!File.directoryExists(folder_path))
            {
                File.createDirectory(folder_path);
            }

            let file_dialog = new OpenFileDialog;

            file_dialog.initialPath = folder_path;

            file_dialog.filters = [["JSON files", "*.json"], ["All files", "*.*"] ];

            file_dialog.multipleSelections = false;

            file_dialog.caption = "Select saved settings file";

            if (file_dialog.execute())
            {
                setings_file_name = file_dialog.fileNames[0]

                let file = new File;

                file.openForReading( setings_file_name );

                let loaded_data = JSON.parse( file.read( DataType_ByteArray, file.size).toString() );

                this.b_auto_STF                           = loaded_data.b_auto_STF;

                // For backward compatibility ------------------------------------
                if (loaded_data.hasOwnProperty("b_linked_RGB_channels"))
                {
                   this.b_linked_RGB_channels             = loaded_data.b_linked_RGB_channels;
                }
                else if (loaded_data.hasOwnProperty("b_linked_STF"))
                {
                   this.b_linked_RGB_channels             = loaded_data.b_linked_STF;
                }

                if (loaded_data.hasOwnProperty("b_linear_fit_colorized_images_before_combine"))
                {
                   this.b_linear_fit_colorized_images_before_combine  = loaded_data.b_linear_fit_colorized_images_before_combine;
                }
                else if (loaded_data.hasOwnProperty("b_linear_fit_swatches_before_combine"))
                {
                   this.b_linear_fit_colorized_images_before_combine   = loaded_data.b_linear_fit_swatches_before_combine;
                }
                // --------------------------------------------------------------------

                this.review_image_saturation_boost        = loaded_data.review_image_saturation_boost;
                this.review_image_midtone_adjustment      = loaded_data.review_image_midtone_adjustment;
                this.review_image_shadows_clipping        = loaded_data.review_image_shadows_clipping;
                this.colorize_method                      = loaded_data.colorize_method;
                this.combine_method                       = loaded_data.combine_method;
                this.rescale_method                       = loaded_data.rescale_method;
                this.b_rescale_channels_independently     = loaded_data.b_rescale_channels_independently;
                this.palette                              = loaded_data.palette;
                this.b_auto_set_palette_for_known_filter  = loaded_data.b_auto_set_palette_for_known_filter;

                for (let i = 0; i < (loaded_data.image_attributes_array.length); i++)
                {

                    let loaded_image_attributes_array_row =  loaded_data.image_attributes_array[i];

                    let b_active =  loaded_image_attributes_array_row.b_active;
                    let filter   =  loaded_image_attributes_array_row.filter;
                    let amount   =  loaded_image_attributes_array_row.amount;
                    let red      =  loaded_image_attributes_array_row.red;
                    let green    =  loaded_image_attributes_array_row.green;
                    let blue     =  loaded_image_attributes_array_row.blue;

                    for (let row = 0; row < NUM_VIEW_SLOTS; row++)
                    {
                        let b_row_updated = false;

                        if (this.image_attributes_array[row].b_active)
                        {
                            if ((filter == this.image_attributes_array[row].filter) && (!b_row_updated))
                            {

                                this.image_attributes_array[row].amount = amount;
                                this.image_attributes_array[row].red    = red;
                                this.image_attributes_array[row].green  = green;
                                this.image_attributes_array[row].blue   = blue;

                                this.setSettingsChangedFlag_implementation(true);

                                b_row_updated = true;

                                continue;
                            }
                        }
                    }

               }

                file.close();

            }

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: loadSettings_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: loadSettings_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
        }
    };

    //

    this.saveDefaultUserPalettes_JSON_implementation = function(palettes, folder_path)
    {

        let function_name = "saveDefaultUserPalettes_JSON_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>saveDefaultUserPalettes_JSON_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            console.writeln("");
            console.writeln("--- SAVING DEFAULT USER PALETTES ---");
            console.writeln("");

            let file_name = "DC_CombineImages_default_user_palettes_JSON.json";

            let file_path = folder_path + "/" + file_name;

            let file = File.createFileForWriting( file_path );

            file.write(ByteArray.stringToUTF8(JSON.stringify(palettes, null, 4)));

            file.close();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: saveDefaultUserPalettes_JSON_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: saveDefaultUserPalettes_JSON_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
        }

    };

    //

    this.saveDefaultSystemPalettes_JSON_implementation = function(palettes, folder_path)
    {

        let function_name = "saveDefaultSystemPalettes_JSON_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>saveDefaultSystemPalettes_JSON_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        // This function does almost the same as 'saveDefaultUserPalettes_JSON_implementation()',
        // except it is saved to a System JSON file.
        // It should be executed immediately after creating the Palettes structure
        // and before loading the user palettes to establish backup System Palette defaults

        try
        {
            console.writeln("");
            console.writeln("--- SAVING (BACKUP) DEFAULT SYSTEM PALETTES (JSON) ---");
            console.writeln("");

            let file_name = "DC_CombineImages_default_system_palettes_JSON.json";

            let file_path = folder_path + "/" + file_name;

            let file = File.createFileForWriting( file_path );

            file.write(ByteArray.stringToUTF8(JSON.stringify(palettes, null, 4)));

            file.close();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: saveDefaultSystemPalettes_JSON_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: saveDefaultSystemPalettes_JSON_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
        }

    };

    //

    this.loadDefaultUserPalettes_JSON_implementation = function(palettes, folder_path)
    {

        let function_name = "loadDefaultUserPalettes_JSON_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>loadDefaultUserPalettes_JSON_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            console.writeln("");
            console.writeln("--- LOADING DEFAULT USER PALETTES (JSON) ---");
            console.writeln("");

            let filename = "DC_CombineImages_default_user_palettes_JSON.json";

            let input_files_array = FindFiles( folder_path, [ filename ], false );

            let num_files_found = input_files_array[0].length;

            if (num_files_found > 0)
            {
                let data_string = readTextFile(filename, folder_path);

                palettes.filter_palettes = JSON.parse(data_string).filter_palettes;
            }

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: loadDefaultUserPalettes_JSON_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: loadDefaultUserPalettes_JSON_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
       }

    };

    //

    this.updateDefaultPalettes_implementation = function(palettes, folder_path)
    {

        let function_name = "updateDefaultPalettes_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>updateDefaultPalettes_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            console.writeln("");
            console.writeln("--- UPDATING DEFAULT USER PALETTES ---");
            console.writeln("");

            let result = (new MessageBox(
                  "<p>" + "This will update user default palette settings, based on currently active images.</p> <p>Do you wish to continue?</p>",
                  "Update User Default Palette Settings",
                  StdIcon_Question,
                  StdButton_Yes, StdButton_No
            )).execute();

            if (result == StdButton_No)
            {
                return;
            }

            let palette = this.palette;

            for (let row = 0; row < NUM_VIEW_SLOTS; row++)
            {
                if (this.image_attributes_array[row].b_active )
                {
                     let filter_name = this.image_attributes_array[row].filter;

                     this.palettes.filter_palettes[filter_name][palette][0] = this.image_attributes_array[row].amount * 100 ;
                     this.palettes.filter_palettes[filter_name][palette][1] = this.image_attributes_array[row].red    * 255 ;
                     this.palettes.filter_palettes[filter_name][palette][2] = this.image_attributes_array[row].green  * 255 ;
                     this.palettes.filter_palettes[filter_name][palette][3] = this.image_attributes_array[row].blue   * 255 ;
                }
            }

            this.saveDefaultUserPalettes_JSON_implementation(palettes, folder_path);

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: updateDefaultPalettes_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: updateDefaultPalettes_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
        }

    };

    //

    this.resetDefaultSystemPalettes_implementation = function(palettes, folder_path)
    {

        let function_name = "resetDefaultSystemPalettes_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>resetDefaultSystemPalettes_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            console.writeln("");
            console.writeln("--- RESETTING PALETTES TO SYSTEM DEFAULTS ---");
            console.writeln("");

            let result = (new MessageBox(
                  "<p>" + "This will reset palette settings to system defaults.</p> <p>Do you wish to continue?</p>",
                  "Restore System Default Palette Settings",
                  StdIcon_Question,
                  StdButton_Yes, StdButton_No
            )).execute();

            if (result == StdButton_No)
            {
               return;
            }

            let filename_system = "DC_CombineImages_default_system_palettes_JSON.json";

            // First, reload the BACKUP SYSTEM palettes  from the relevant JSON file
            let input_files_array = FindFiles( folder_path, [ filename_system ], false );

            let num_files_found = input_files_array[0].length;

            if (num_files_found > 0)
            {
                let data_string = readTextFile(filename_system, folder_path);

                palettes.filter_palettes = JSON.parse(data_string).filter_palettes;

                // Then update the USER pallets to match in the relevant JSON file
                this.saveDefaultUserPalettes_JSON_implementation(palettes, folder_path);

            }

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: resetDefaultSystemPalettes_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: resetDefaultSystemPalettes_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
        }
    };

    //

    this.selectAndOpenImage_implementation = function(folder_path,b_multiple_images = false)
    {

        let function_name = "selectAndOpenImage_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>selectAndOpenImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            console.writeln("");
            console.writeln("--- OPEN AND SELECT IMAGES ---");
            console.writeln("");

            let image_file_name = "";

            let file_dialog = new OpenFileDialog;

            file_dialog.initialPath = this.input_images_folder;

            file_dialog.filters = [
                                     [ "All supported formats", ".xisf", ".fit", ".fits", ".fts", "tif", "jpg", "png" ],
                                     ["XISF files", "*.xisf"],
                                     ["FIT files",  "*.fit" ],
                                     ["FITS files", "*.fits"],
                                     ["FTS files",  "*.fts" ],
                                     ["TIF files",  "*.tif" ],
                                     ["JPG files",  "*.jpg" ],
                                     ["PNG files",  "*.png" ],
                                     ["All files",  "*.*"   ]
                                  ];

            file_dialog.multipleSelections = b_multiple_images;

            file_dialog.caption = "Select and open images";

            let opened_views_array = [];

            if (file_dialog.execute())
            {
                let image_file_names_array = file_dialog.fileNames;

                let num_images_files = image_file_names_array.length;

                for (let i = 0; i < num_images_files; i++)
                {
                    opened_views_array.push(openSelectedImage(image_file_names_array[i]) );
                }

                if (num_images_files >= 0)
                {
                    let first_openedfilename = image_file_names_array[0];
                    let first_opened_view_drive  =  File.extractDrive(first_openedfilename);
                    let first_opened_view_folder =  File.extractDirectory(first_openedfilename);
                    this.input_images_folder = first_opened_view_drive + first_opened_view_folder;
                }
            }

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: selectAndOpenImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            if (opened_views_array.length > 0)
            {
                return opened_views_array[0];
            }
            else
            {
                return null;
            }

        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: selectAndOpenImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);

            return null;
        }
    };

    //

    this.autoOpenInputImages_implementation = function()
    {

        let function_name = "autoOpenInputImages_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>autoOpenInputImages_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            let script_file_path    = #__FILE__;
            let drive  = File.extractDrive(script_file_path);
            let folder = File.extractDirectory(script_file_path);
            let root_folder  = drive + folder + "/../";
            let input_files_path = root_folder + "Input Images" ;

            let input_files_array = FindFiles( input_files_path, [ "*.xisf",
                                                               "*.fit",
                                                               "*.fits",
                                                               "*.tif",
                                                               "*.jpg",
                                                               "*.png"],
                                                              false );

             let num_input_files_subarrays = input_files_array.length;

             let num_images_found = 0;

             for (let i = 0; i < num_input_files_subarrays; i++)
             {
                 num_images_found += input_files_array[i].length;
             }

             if (num_images_found > 0)
             {
                 let word1 = (num_images_found ==  1 ? "image" : "images");
                 let word2 = (num_images_found ==  1 ? "it"    : "them");
                 let result = (new MessageBox(
                  "<p>" + num_images_found + " " + word1 + " found in the 'Input Images' folder.</p> <p>Do you wish to open " + word2 + "?</p>",
                  "AUTO-OPEN IMAGES",
                  StdIcon_Question,
                  StdButton_Yes, StdButton_No
                 )).execute();

                 if (result == StdButton_No)
                 {
                    return;
                 }

                 console.writeln("");
                 console.writeln("--- OPENING IMAGES FOUND IN INPUT FOLDER ---");
                 console.writeln("");
             }

             for (let i = 0; i < num_input_files_subarrays; i++)
             {
                 for (let j = 0; j < input_files_array[i].length; j++)
                 {
                     let file_path = input_files_array[i][j];

                     let new_view = openSelectedImage(file_path);

                     if (new_view)
                     {
                         this.auto_opened_views_array.push(new_view);
                     }
                 }
             }

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: autoOpenInputImages_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        }
        catch (exception)
        {
            errorProcessing_standard(function_name, exception.message);

           return;
        }

    };

    //

    this.getCombineTargetArray_implementation = function()
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>getCombineTargetArray_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let combine_target_array = null;

        if (this.working_images_shrink_mode == "FULL SIZE")
        {
            combine_target_array = this.view_array;
        }
        else
        {
            combine_target_array = this.working_view_array;

        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: getCombineTargetArray_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return combine_target_array;
    };

    //

    this.getWorkingImagesCenter_implementation = function(local_new_center_normalized_x,
                                                          local_new_center_normalized_y)
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>getWorkingImagesCenter_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let new_center_global_normalized     = [0.5, 0.5] ; // default

        let current_center_global_normalised = this.working_images_center_normalized;

        // Find size of main image
        let main_image_size_pixels = [ -1, -1];
        let index = -1;
        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (this.image_attributes_array[i].b_active)
            {
                index++;

                if (index == 0)
                {
                    main_image_size_pixels = [
                                                 this.view_array[i].image.width,
                                                 this.view_array[i].image.height
                                             ]
                    break;
                }
            }
        }
        if (index == -1) // ie no active (main) images
        {
            console.writeln( "no Active views" );
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: getWorkingImagesCenter_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return new_center_global_normalized ; // default;
        }

        // Find size of working images
        let working_images_size_pixels = [ -1, -1];
        let index2 = -1;
        for (let j = 0; j < NUM_VIEW_SLOTS; j++)
        {
            if (this.image_attributes_array[j].b_active && this.working_view_array.length > 0)
            {
                index2++;

                if (index2 == 0)
                {
                    working_images_size_pixels = [
                                                     this.working_view_array[j].image.width,
                                                     this.working_view_array[j].image.height
                                                 ]
                    break;
                }
            }
        }
        if (index2 == -1) // ie no active (working) images
        {

            working_images_size_pixels = main_image_size_pixels;

        }

        let current_center_global_pixels = [
                                               current_center_global_normalised[0] * main_image_size_pixels[0],
                                               current_center_global_normalised[1] * main_image_size_pixels[1]
                                           ]

        let new_center_local_pixels   = [
                                            local_new_center_normalized_x * working_images_size_pixels[0],
                                            local_new_center_normalized_y * working_images_size_pixels[1]
                                        ];

        let new_center_global_pixels = [
                                           current_center_global_pixels[0] + new_center_local_pixels[0],
                                           current_center_global_pixels[1] + new_center_local_pixels[1]
                                       ];

        // Enforce boundary limits
        if ((new_center_global_pixels[0] + working_images_size_pixels[0]/2) >  main_image_size_pixels[0])
        {
           new_center_global_pixels[0] = main_image_size_pixels[0] - working_images_size_pixels[0]/2;
           console.writeln("");
           console.writeln("Boundary limit enforced: x < main_image_size_pixels");
           console.writeln("");
         }
        else if ((new_center_global_pixels[0] - working_images_size_pixels[1]/2) <  0)
        {
           new_center_global_pixels[0] = working_images_size_pixels[1]/2;
           console.writeln("");
           console.writeln("Boundary limit enforced: x > 0");
           console.writeln("");
        }

        if ((new_center_global_pixels[1] + working_images_size_pixels[1]/2) >  main_image_size_pixels[1])
        {
           new_center_global_pixels[1] = main_image_size_pixels[1] - working_images_size_pixels[1]/2;
           console.writeln("");
           console.writeln("Boundary limit enforced: y < main_image_size_pixels");
           console.writeln("");
         }
        else if ((new_center_global_pixels[1] - working_images_size_pixels[1]/2) <  0)
        {
           new_center_global_pixels[1] = working_images_size_pixels[1]/2;
           console.writeln("");
           console.writeln("Boundary limit enforced: y > 0");
           console.writeln("");
        }

        new_center_global_normalized = [
                                           new_center_global_pixels[0] / main_image_size_pixels[0],   // Normalized
                                           new_center_global_pixels[1] / main_image_size_pixels[1]    // Normalized
                                       ];

        console.writeln("");


        console.writeln("main_image_size_pixels           : " + main_image_size_pixels);
        console.writeln("");
        console.writeln("current_center_global_normalised : " + current_center_global_normalised);
        console.writeln("");
        console.writeln("current_center_global_pixels     : " + current_center_global_pixels);
        console.writeln("");
        console.writeln("working_images_size_pixels       : " + working_images_size_pixels);
        console.writeln("");
        console.writeln("");
        console.writeln("local_new_center_normalized_x    : " + local_new_center_normalized_x);
        console.writeln("");
        console.writeln("local_new_center_normalized_y    : " + local_new_center_normalized_y);
        console.writeln("");
        console.writeln("new_center_local_pixels          : " + new_center_local_pixels);
        console.writeln("");
        console.writeln("new_center_global_pixels         : " + new_center_global_pixels);
        console.writeln("");
        console.writeln("new_center_global_normalized     : " + new_center_global_normalized);
        console.writeln("");

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: getWorkingImagesCenter_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return new_center_global_normalized;
    };

    //

    this.getWorkingView_implementation = function(view,
                                                  working_images_shrink_mode,
                                                  center_normalized = [0.5, 0.5],
                                                  name_prefix = "")
    {

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>getWorkingView_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        if (view == null || view.isNull)
        {
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: getWorkingView_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            return null;
        }

        if (name_prefix == "")
        {
           name_prefix = "WORKING_DUPLICATE"
        }

        let working_view = createDuplicate(view, name_prefix, false); // false == not visible

        let shrink_factor = 1.0;

        if (working_images_shrink_mode == "RESAMPLE_0.50" ||
            working_images_shrink_mode == "RESAMPLE_0.25" ||
            working_images_shrink_mode == "RESAMPLE_0.10")
        {
            switch (working_images_shrink_mode)
            {
                case "RESAMPLE_0.50": shrink_factor = 0.5;  working_view.id = name_prefix + "_" + "RESAMPLE_0_50_" ; break;
                case "RESAMPLE_0.25": shrink_factor = 0.25; working_view.id = name_prefix + "_" + "RESAMPLE_0_25_" ; break;
                case "RESAMPLE_0.10": shrink_factor = 0.1;  working_view.id = name_prefix + "_" + "RESAMPLE_0_10_" ; break;
            }
            let P = new Resample;
            P.xSize             = shrink_factor;
            P.ySize             = shrink_factor;
            P.mode              = Resample.prototype.RelativeDimensions;
            P.absoluteMode      = Resample.prototype.ForceWidthAndHeight;
            P.xResolution       = 72.000;
            P.yResolution       = 72.000;
            P.metric            = false;
            P.forceResolution   = false;
            P.interpolation     = Resample.prototype.Auto;
            P.clampingThreshold = 0.30;
            P.smoothness        = 1.50;
            P.noGUIMessages     = true;

            executeOn(working_view, P);
        }
        else if (working_images_shrink_mode == "CROP_0.50" ||
                 working_images_shrink_mode == "CROP_0.25" ||
                 working_images_shrink_mode == "CROP_0.10")
        {
            switch (working_images_shrink_mode)
            {
                case "CROP_0.50": shrink_factor = 0.5;  working_view.id = name_prefix + "_" + "CROP_0_50_" ; break;
                case "CROP_0.25": shrink_factor = 0.25; working_view.id = name_prefix + "_" + "CROP_0_25_" ; break;
                case "CROP_0.10": shrink_factor = 0.1;  working_view.id = name_prefix + "_" + "CROP_0_10_" ; break;
            }
            let P = new DynamicCrop;
            P.centerX           = center_normalized[0];
            P.centerY           = center_normalized[1];
            P.width             = shrink_factor;
            P.height            = shrink_factor;
            P.angle             = 0.0000;
            P.scaleX            = 1.00000;
            P.scaleY            = 1.00000;
            P.optimizeFast      = true;
            P.interpolation     = DynamicCrop.prototype.Auto;
            P.clampingThreshold = 0.30;
            P.smoothness        = 1.50;
            P.xResolution       = 72.000;
            P.yResolution       = 72.000;
            P.metric            = false;
            P.forceResolution   = false;
            P.red               = 0.000000;
            P.green             = 0.000000;
            P.blue              = 0.000000;
            P.alpha             = 1.000000;
            P.noGUIMessages     = true;

            executeOn(working_view, P);
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: getWorkingView_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return working_view;
    }

    //

    this.setWorkingImagesShrinkMode_implementation = function(working_images_shrink_mode,
                                                              working_images_center_normalized = [0.5, 0.5])
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>setWorkingImagesShrinkMode_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        // Only do something if there is a change
        if (this.working_images_shrink_mode != working_images_shrink_mode
           || this.working_images_center_normalized != working_images_center_normalized)
        {

            this.working_images_shrink_mode       = working_images_shrink_mode;
            this.working_images_center_normalized = working_images_center_normalized;

            if (this.working_images_shrink_mode != "FULL SIZE" )
            {
                this.createWorkingImages_implementation(working_images_shrink_mode, working_images_center_normalized);
            }
            else
            {
                // Close the all previous working_view_array[] entries
                for (let i = 0; i < NUM_VIEW_SLOTS; i++)
                {
                    if (this.working_view_array[i] != null)
                    {
                        closeWindow(this.working_view_array[i]);
                        this.working_view_array[i] = null;
                    }
                }
            }
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: setWorkingImagesShrinkMode_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    }

    //

    this.createWorkingImages_implementation = function(working_images_shrink_mode,
                                                       working_images_center_normalized = [0.5, 0.5])
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>createWorkingImages_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let index = -1;
        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (this.image_attributes_array[i].b_active)
            {
                index++;
                if (this.view_array[i] != null && ! this.view_array[i].isNull)
                {
                    this.updateStatus_implementation(this.view_array[i].id);
                }

                // Close the any previous view in that slot (only do this for working_view_array[] entries)
                if (this.working_view_array[i] != null)
                {
                    closeWindow(this.working_view_array[i]);
                    this.working_view_array[i] = null;
                }

                // Create a new working image
                let new_working_view = this.getWorkingView_implementation(this.view_array[i],
                                                                          working_images_shrink_mode,
                                                                          working_images_center_normalized,
                                                                          "WORKING_DUPLICATE");

                // Add the new working view to working_view_array[]
                this.working_view_array[i] = new_working_view;

                if (index == 0)
                {
                    this.working_images_size = [
                                                  new_working_view.image.width,
                                                  new_working_view.image.height
                                               ]
                }

            }
        }

        this.clearStatus_implementation();

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: createWorkingImages_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    }

    //

    this.createWorkingImagesForColorizedImagePreview_implementation = function()
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>createWorkingImagesForColorizedImagePreview_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let b_first_active_image_found = false;

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {

            if (this.view_array[i] != null && ! this.view_array[i].isNull)
            {
                this.updateStatus_implementation(this.view_array[i].id);
            }

            if (this.working_view_for_colorized_images_array[i] != null)
            {
                closeWindow(this.working_view_for_colorized_images_array[i]);
                this.working_view_for_colorized_images_array[i] = null;
            }

            // Create a new working_image_for_colorized_image_preview
            let new_working_image_for_colorized_image_preview = this.getWorkingView_implementation(this.view_array[i],
                                                                                  "RESAMPLE_0.10",
                                                                                  [0.5, 0.5],
                                                                                  "WORKING_DUPLICATE_FOR_COLORIZED_IMAGE");
            // Add the new working view to working_view_array[] (can be null !)
            this.working_view_for_colorized_images_array[i] = new_working_image_for_colorized_image_preview;

            // Update the  colorized_image_preview_size variable if an active image is found
            if (new_working_image_for_colorized_image_preview && ! b_first_active_image_found)
            {
                 b_first_active_image_found = true;

                 this.colorized_image_preview_size = [
                                                 new_working_image_for_colorized_image_preview.image.width,
                                                 new_working_image_for_colorized_image_preview.image.height
                                                     ]
            }
        }

        if (b_first_active_image_found)
        {
            this.colorized_image_preview.reset(this.colorized_image_preview_size[0],
                                               this.colorized_image_preview_size[1]);
        }
        else
        {
            this.colorized_image_preview_size = [-1, -1] ;
        }

        this.clearStatus_implementation();

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: createWorkingImagesForColorizedImagePreview_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    }

    //

    this.setColorSpread_implementation = function()
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>setColorSpread_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let num_active_views = 0;
        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (this.image_attributes_array[i].b_active)
            {
                num_active_views++;
            }
        }

        let color_spacing = 1 / (num_active_views);

        let saturation = 1; let value = 1;

        let hue = 0;

        let active_row_index = -1;

        for (let row = 0; row < NUM_VIEW_SLOTS; row++)
        {
            if (this.image_attributes_array[row].b_active)
            {
                active_row_index++;

                hue = color_spacing * active_row_index;

                let new_RGB = [ 0, 0 , 0 ];

                new_RGB = convertHSVtoRGB(hue, saturation, value);

                this.image_attributes_array[row].red   = new_RGB[0];
                this.image_attributes_array[row].green = new_RGB[1];
                this.image_attributes_array[row].blue  = new_RGB[2];
            }
        };

        this.setSettingsChangedFlag_implementation(true);

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: setColorSpread_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    };

    //

    this.updateColorizedImagePreview_implementation = function(row)
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>updateColorizedImagePreview_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        this.updateStatus_implementation("Colorized Image Preview: Colorizing");

        // For colorized_image_preview, only support "MULTIPLY" colorize method
        // and linked stretch.  "MULTIPLY" is faster and allows mismatched
        // geometries between grayscale images and color swatches
        this.colorized_image_preview.colorize(this.working_view_for_colorized_images_array[row],
                                     this.image_attributes_array[row].red,
                                     this.image_attributes_array[row].green,
                                     this.image_attributes_array[row].blue,
                                     "MULTIPLY" );  // "MULTIPLY" == this.colorize_method

        let current_stretch_settings = g_helper_data.H;

        this.updateStatus_implementation("Colorized Image Preview: Autostretch");

        // function setAutoSTFparameters(view,
        //                               shadows_clipping  = SHADOWS_CLIP ,
        //                               target_background = TARGET_BKG,
        //                               b_rgb_linked      = RGB_LINK)

        setAutoSTFparameters(this.colorized_image_preview.view,
                             SHADOWS_CLIP, TARGET_BKG,
                             true );  // true == linked RGB channels

        applyHistogramTransformation(this.colorized_image_preview.view,
                                         g_helper_data.H);

        g_helper_data.H = current_stretch_settings;

        let H = [ // c0, m, c1, r0, r1
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, this.colorized_image_preview_midtone_adjustment, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000]
                ];

        applyHistogramTransformation(this.colorized_image_preview.view, H);

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: updateColorizedImagePreview_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    }

    //

    this.selectRow_implementation = function(row, b_clear_selection = false)
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>selectRow_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        if ((this.image_attributes_array[row].b_active) &&
            (this.selected_view_row != row)             &&
            (! b_clear_selection))
        {
            this.selected_view_row = row;

            this.updateColorizedImagePreview_implementation(row);

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: selectRow_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return true;
        }
        else
        {
            this.selected_view_row = -1;

            this.colorized_image_preview.clear();

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: selectRow_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }
    };

    //

    this.setPalette_implementation = function(palette,
                                              b_reset_image_attributes_array = false)
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>setPalette_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        this.palette = palette;

        if (b_reset_image_attributes_array)
        {
            this.resetImageAttributeArray_implementation();
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: setPalette_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    };

    //

    this.getFilterName_implementation = function(view, filter_names, filter_names_JWST, priority_JWST_filter_array)
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>getFilterName_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        // String case sensitivity approach:
        // Strings are used as keys in the filter-palette database, so case sensitivity
        // is important. We need to ensure to tests are not unduely restrictive for
        // images imported from the external desktop or filesystem
        // (as would be the case for case sensitive testing), but also that case
        // insensitve tests do not result in falures later due to invalid keys.
        // Therefore, string comparison tests will be case-insensitive;
        // and after any testing, strings names of input filters (from metadata
        // or file names) will remain or be converted (if necessary ) to the case
        // used by the system.

        // Return the filter type as recorded in the ImageWindow metadata

        let filter_name = "NONE";

        let keywords = [];

        if (view == null || view.isNull)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: getFilterName_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return filter_name;
        }

        keywords = view.window.keywords;

        for ( let i = 0; i < keywords.length; ++i )
        {
            if (   keywords[i].name.toUpperCase().trim() == "FILTER"
               ||  keywords[i].name.toUpperCase().trim() == "INSFLNAM"
               ||  keywords[i].name.toUpperCase().trim() == "SURVEY")
            {
                filter_name = keywords[i].strippedValue.trim();
                break;
            }
        }

        // If no metadata solution, examine the view/file name

        if (filter_name == "NONE")
        {
            let match1_position  = -1;
            let match1_string    = "";
            let match2_position  = -1;
            let match2_string    = "";
            let match3_position  = -1;
            let match3_string    = "";
            let match4_position  = -1;
            let match4_string    = "";
            let match5_position  = -1;
            let match5_string    = "";

            let exclude_list     = [ "IR", "I", "R", "G", "U", "Z"];  // Upper case for testing

            let num_matches      = 0;

            let view_name_uppercase = view.id.toUpperCase().trim();   // Upper case for testing

            let filterNamesArrayLength = filter_names.length

            for (let i = 0; i < filter_names.length; i++)
            {

                let search_string = filter_names[i].toUpperCase();    // Upper case for testing

                let  exclude_list_index = exclude_list.indexOf(search_string);

                if (exclude_list_index >= 0 )
                {
                    continue;
                }

                let string_position = view_name_uppercase.search(search_string);

                if (string_position >= 0)
                {

                    num_matches++;

                    if (num_matches == 1)
                    {
                        match1_position = string_position;
                        match1_string   = filter_names[i];  // Case remains as input
                    }
                    else if (num_matches == 2)
                   {
                        match2_position = string_position;
                        match2_string   = filter_names[i];  // Case remains as input
                    }
                    else if (num_matches == 3)
                    {
                        match3_position = string_position;
                        match3_string   = filter_names[i];  // Case remains as input
                    }
                    else if (num_matches == 4)
                    {
                        match4_position = string_position;
                        match4_string   = filter_names[i];  // Case remains as input
                    }
                    else if (num_matches == 5)
                    {
                        match5_position = string_position;
                        match5_string   = filter_names[i];  // Case remains as input
                        break;
                    }

                }
                exclude_list_index = -1;
            }

            // Unable to 'see' FILTER keyword in JWST  metadata, so using filenames instead
            // Very imperfect matching system ... Could do with work
            if (num_matches == 1)
            {
                filter_name = match1_string;
            }
            else if ( priority_JWST_filter_array.indexOf(match1_string) >= 0)
            {
               filter_name = match1_string;
            }
            else if ( priority_JWST_filter_array.indexOf(match2_string) >= 0)
            {
               filter_name = match2_string;
            }
            else if ( priority_JWST_filter_array.indexOf(match3_string) >= 0)
            {
               filter_name = match3_string;
            }
            else if ( priority_JWST_filter_array.indexOf(match4_string) >= 0)
            {
               filter_name = match4_string;
            }
            else if ( priority_JWST_filter_array.indexOf(match5_string) >= 0)
            {
               filter_name = match5_string;
            }
            else if ( filter_names_JWST.indexOf(match1_string) >= 0)
            {
               filter_name = match1_string;
            }
            else if ( filter_names_JWST.indexOf(match2_string) >= 0)
            {
               filter_name = match2_string;
            }
            else if ( filter_names_JWST.indexOf(match3_string) >= 0)
            {
               filter_name = match3_string;
            }
            else if ( filter_names_JWST.indexOf(match4_string) >= 0)
            {
               filter_name = match4_string;
            }
            else if ( filter_names_JWST.indexOf(match5_string) >= 0)
            {
               filter_name = match5_string;
            }
            else
            {
                filter_name = "NONE";
            }
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: getFilterName_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return filter_name.toUpperCase();  // Uppercase used by the system
    };

    //

    this.resetImageAttributeArrayRow_implementation = function(row)
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>resetImageAttributeArrayRow_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        // This script maintains the array 'image_attributes_array' of size NUM_VIEW_SLOTS.
        // An 'ImageAttributesSet' is recorded in each row of the array,
        // detailing the following elements: b_active, bSelected, filter, amount, and RGB values

        // This function reset attributes in a specified row (for a particular image),
        // based on the filter type, and if relevant, the currently selected palette

        // String case sensitivity approach.
        // Strings are used as keys in filter-palette database, so strings cases
        // are important. In this function, all tests are case sensitive.

        function array_includes(arr, str)
        {
            for (let i = 0; i < arr.length; i++)
            {
               if (arr[i] === str)
                {
                    return true;
                }
            }
            return false;
        };

        let filter_name  = this.image_attributes_array[row].filter;
        let palette_name = this.palette;

        let palette_settings = [100, 255, 255, 255];

        if (array_includes(this.filter_names, filter_name))
        {
             let valid_filter_palettes = Object.getOwnPropertyNames(this.palettes.filter_palettes[filter_name]);

             if (array_includes(valid_filter_palettes, palette_name))
             {
                 let filter_settings = this.palettes.filter_palettes[filter_name];
                 palette_settings = filter_settings[this.palette];
             }
        }

        this.image_attributes_array[row].amount = palette_settings[0] / 100.0;
        this.image_attributes_array[row].red    = palette_settings[1] / 255.0;
        this.image_attributes_array[row].green  = palette_settings[2] / 255.0;
        this.image_attributes_array[row].blue   = palette_settings[3] / 255.0;

        this.setSettingsChangedFlag_implementation(true);

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: resetImageAttributeArrayRow_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    };

    //

    this.resetImageAttributeArray_implementation = function()
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>resetImageAttributeArray_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        // This script maintains the array 'image_attributes_array' of size NUM_VIEW_SLOTS.
        // An 'ImageAttributesSet' is recorded in each row of the array,
        // detailing the following elements: b_active, filter, amount, and RGB values

        // This function iterates through the array and resets each row,
        // based on the filter type, and if relevant, the currently selected Palette

        // Iterate through the image attribute array
        for (let row = 0; row < NUM_VIEW_SLOTS; row++)
        {

            this.resetImageAttributeArrayRow_implementation(row);
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: resetImageAttributeArray_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    };

    //

    this.checkNewViewIsGrayscale_implementation = function(view)
    {
        // Check to see whether the specified new view is grayscale

        let function_name = "checkNewViewIsGrayscale_implementation()";

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>checkNewViewIsGrayscale_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            if (view == null || view.isNull)
            {
                throw new Error(function_name + " : NULL VIEW");
            }

            console.writeln( "" );
            console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

            let numChannels = view.image.numberOfChannels;

            if (numChannels != 1)
            {
               console.writeln( "<end><cbr/><br/><b>","Error : </b>Image is not grayscale" );
               return false;
            }
            console.writeln( "<end><cbr/><br/><b>","OK", "</b>" );

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: checkNewViewIsGrayscale_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return true;
        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: checkNewViewIsGrayscale_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);

            return false;
        }
    };

    //

    this.checkNewViewSize_implementation = function(view, row)
    {
        // Check to see whether the specified new view is the same size as other active views

        let function_name = "checkNewViewSize_implementation()";

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>checkNewViewSize_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            if (view == null || view.isNull)
            {
                throw new Error(function_name + " : NULL VIEW");
            }

            console.writeln( "" );
            console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

            for (let testRow = 0; testRow < NUM_VIEW_SLOTS; testRow++)
            {
                if (testRow != row   && this.image_attributes_array[testRow].b_active )
                {
                    let testView   = this.view_array[testRow];
                    let testWidth  = testView.image.width;
                    let testHeight = testView.image.height;

                    if ( view.image.width !=  testWidth || view.image.height !=  testHeight)
                    {
                        console.writeln( "<end><cbr/><br/><b>","WARNING : </b>View is different size to other active views - unpredictable results possible; linear fit not possible");
                        return false;
                    }
                }
            }
            console.writeln( "<end><cbr/><br/><b>","OK", "</b>" );

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: checkNewViewSize_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return true;
        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: checkNewViewSize_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);

            return false;
        }
    };

    //

    this.closeCombinedImage_implementation = function(optionalText = "" )
    {

        // This script creates a single hidden 'COMBINED_IMAGE',
        // which is used with 'drawScaledBitmap()' to draw the image on to the
        // Dialog's 'review_image_control'.

        // This function closes COMBINED_IMAGE when called.

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>closeCombinedImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        console.writeln("");
        console.writeln("Closing CombineImage");
        console.writeln("");

        if (this.combined_image_view && ! this.combined_image_view.isNull)
        {
            closeWindow(this.combined_image_view);
        }

        this.combined_image_view = null;

        console.writeln("");
        console.writeln("Also closing any leftover CombineImage windows that may persist if previous script instance closed prematurely");
        console.writeln("");

        let previous_combined_image_window = ImageWindow.windowById("COMBINED_IMAGE");

        if (previous_combined_image_window && ! previous_combined_image_window.isNull && ! previous_combined_image_window.mainView.isNull)
        {
            closeWindow(previous_combined_image_window.mainView); // Helper functions expose main views, not windows
        }
        previous_combined_image_window = null;

        console.writeln("");
        console.writeln("Also closing any leftover Splash Image windows that may persist if previous script instance closed prematurely");
        console.writeln("");

        let previous_splash_image_window = ImageWindow.windowById("SPLASH_IMAGE");

        if (previous_splash_image_window && ! previous_splash_image_window.isNull && ! previous_splash_image_window.mainView.isNull)
        {
            closeWindow(previous_splash_image_window.mainView); // Helper functions expose main views, not windows
        }
        this.previous_splash_image_window = null;

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: closeCombinedImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    };

    //

    this.conditionalRescale_implementation = function(view, b_individual_channels = false)
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>conditionalRescale_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        let bRescaleRequired = false;
        view.image.pushSelections();
        let max = 0.0; let min = 1.0;
        for (let channel = 0; channel < 3; channel++)
        {
            view.image.selectedChannel = channel;
            max = view.image.maximum();
            min = view.image.minimum();
            console.writeln( "<end><cbr/><br/>" );
            console.writeln( "<end><cbr/><br/><b>","Max : " + max, "</b>" );
            console.writeln( "<end><cbr/><br/><b>","Min : " + min, "</b>" );
            if (max  > 1 || min  < 0)
            {
                 bRescaleRequired = true;
                // break;
            }
        }
        if (bRescaleRequired)
        {
            console.writeln( "<end><cbr/><br/>" );
            console.writeln( "<end><cbr/><br/><b>","Rescale required.", "</b>" );
        }
        else
        {
            console.writeln( "<end><cbr/><br/>" );
            console.writeln( "<end><cbr/><br/><b>","Rescale not required.", "</b>" );
        }

        view.image.popSelections();

        if (bRescaleRequired)
        {
            let P = new Rescale;
            if (b_individual_channels)
            {
                P.mode = Rescale.prototype.RGBK_Individual;
            }
            else
            {
                P.mode = Rescale.prototype.RGBK;
            }
            view.beginProcess(UndoFlag_NoSwapFile);
            P.executeOn( view );
            view.endProcess();
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: conditionalRescale_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return view;
    };


    //

    this.appendView_implementation = function(view)
    {
        // Add a new view/image
        // to the first available row in view_array and image_attributes_array,

        let function_name = "appendView_implementation()";

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>appendView_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            if (view == null || view.isNull)
            {
                throw new Error(function_name + " : NULL VIEW");
            }

            console.writeln( "" );
            console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

            // Iterate through the image attribute array
            for (let row = 0; row < NUM_VIEW_SLOTS; row++)
            {
                if ( this.image_attributes_array[row].b_active == false)
                {
                    this.addViewToRow_implementation(view, row)
                    return true;
                }
            }

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: appendView_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: appendView_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);

            return false;
        }
    };

    //

    this.addViewToRow_implementation = function(view, row = -1)
    {
        // Add a new view/image
        // to a specific row in this.view_array and this.image_attributes_array,
        // then update all attributes in this.image_attributes_array

        let function_name = "addViewToRow_implementation()";

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>addViewToRow_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        // Allow null view to enter this function - with restrictions

        try
        {

            this.updateStatus_implementation("Add view to row: checking for valid image");

            if (view != null && ! view.isNull)
            {

                console.writeln( "" );
                console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

                if (! this.checkNewViewIsGrayscale_implementation(view))
                {
                    (new MessageBox(
                          "<p>Image is not grayscale.</p>",
                          "ERROR",
                          StdIcon_Error,
                          StdButton_Cancel)).execute();
                    this.view_array[row]                      = null;
                    this.image_attributes_array[row].b_active = false;
                    this.image_attributes_array[row].filter   = "NONE";

                   console.writeln( "" );
                   console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                   console.writeln( "<b>END: addViewToRow_implementation()</b>" );
                   console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

                    return false;
                }

                if (view != null && !view.isNull && ! this.checkNewViewSize_implementation(view, row))
                {
                    (new MessageBox(
                          "<p>New view not same size as other views.</p>",
                          "WARNING",
                          StdIcon_Warning,
                          StdButton_Ok
                      )).execute();
                }
            }

            if (row > -1 && row < NUM_VIEW_SLOTS)
            {
                // For view_array[]
                // Add the new view - Do not close previous view/window
                if (view != null && ! view.isNull)
                {
                    this.view_array[row] = view;
                }
                else
                {
                    this.view_array[row] = null; // Ok if view is null
                }

                // For working_view_array[]
                // Close the any previous entry
                if (this.working_view_array[row] != null)
                {
                    closeWindow(this.working_view_array[row]);
                    this.working_view_array[row] = null;
                }

                if (this.working_images_shrink_mode != "FULL SIZE" && view != null && ! view.isNull)
                {
                    // Create a new working image (might be null)
                    let new_working_view = this.getWorkingView_implementation(view,
                                                                              this.working_images_shrink_mode,
                                                                              this.working_images_center_normalized,
                                                                              "WORKING_DUPLICATE");
                    // Add the new working view to working_view_array[]
                    if (new_working_view)
                    {
                        this.working_view_array[row]    = new_working_view;

                        this.working_images_size = [
                                                      new_working_view.image.width,
                                                      new_working_view.image.height
                                                   ]
                    }
                }

                // Image attributes
                this.image_attributes_array[row].b_active = !view.isNull;  // Active or Not Active

                this.image_attributes_array[row].filter  = this.getFilterName_implementation(view ,
                                                                                             this.filter_names,
                                                                                             this.filter_names_JWST,
                                                                                             this.priority_JWST_filter_array );
                this.resetImageAttributeArrayRow_implementation(row);

            }

            // Now the same(ish) for the working_view_for_colorized_images_array[]
            if (row > -1 && row < NUM_VIEW_SLOTS)
            {
                // Close the any previous view in that slot
                if (this.working_view_for_colorized_images_array[row] != null)
                {
                    closeWindow(this.working_view_for_colorized_images_array[row]);
                    this.working_view_for_colorized_images_array[row] = null;
                }
            }

            let new_working_image_for_colorized_image_preview = this.getWorkingView_implementation(view,
                                                                                                   "RESAMPLE_0.10",
                                                                                                   [ 0.5, 0.5 ],
                                                                                                   "WORKING_DUPLICATE_FOR_COLORIZED_IMAGE");

            // Add the new working view to working_view_for_colorized_images_array[] (can  be null !)
            this.working_view_for_colorized_images_array[row] = new_working_image_for_colorized_image_preview;

            // Update the colorized_image_preview if new_working_image_for_colorized_image_preview is not null and dimensions have changed
            if (new_working_image_for_colorized_image_preview)
            {

                 this.colorized_image_preview_size = [
                                                new_working_image_for_colorized_image_preview.image.width,
                                                new_working_image_for_colorized_image_preview.image.height
                                                     ];

                 this.colorized_image_preview.reset(this.colorized_image_preview_size[0],
                                                    this.colorized_image_preview_size[1]);
            }

            this.setSettingsChangedFlag_implementation(true);

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: addViewToRow_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return true;
        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: addViewToRow_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);

            return false;
        }
    };

    //

    this.blackpointClipImage_implementation = function(view)
    {

        let function_name = "blackpointClipImage_implementation()";

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>blackpointClipImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {

            if (view == null || view.isNull)
            {
                let message = "Null target view. Aborting.";

                throw new Error(message);

            }

            console.writeln( "" );
            console.writeln( "<end><cbr/><br/><b> Target view : </b>" + view.fullId );

            if (view != null && ! view.isNull)
            {
                this.updateStatus_implementation(view.id);
            }

            // function blackpointClip(view, bLinked = true, bRescale = false)
            if (!blackpointClip(view, true))
            {
                let message = "Blackpoint Clip failed. Aborting. ";

                throw new Error(message);
            }

            this.setSettingsChangedFlag_implementation(true);

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: blackpointClipImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return true;

        }
        catch (exception)
        {
            (new MessageBox(function_name + "<p>" + exception.message + "<p>", "ERROR", StdIcon_Error, StdButton_Ok)).execute();

            this.setSettingsChangedFlag_implementation(true);

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: blackpointClipImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }
    };

    //

    this.linearFitImage_implementation = function(view, reference_view,
                                                  b_protect_null_RGB_channels = false,
                                                  b_continue_on_error     = false,
                                                  b_silent_on_error       = false)
    {

        let function_name = "linearFitImage_implementation()";

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>linearFitImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {

            if (view == null || view.isNull)
            {
                let message = "Null target view. Aborting.";

                throw new Error(message);

            }

            if (reference_view == null || reference_view.isNull)
            {
                let message = "Null reference view. Aborting.";

                throw new Error(message);
            }

            if (view.fullId != reference_view.fullId ) // ie don't linear fit image with itself
            {

                console.writeln("");
                console.writeln("Linear Fit Reference : " + reference_view.fullId);
                console.writeln("Linear Fit Target    : " + view.fullId );
                console.writeln("");

                if ((view.image.width == reference_view.image.width)
                    && (view.image.height == reference_view.image.height) )
                {

                    // function applyLinearFit(sourceView, targetView)

                    if (view != null && ! view.isNull)
                    {
                        this.updateStatus_implementation("LF: " + view.id);
                    }

                    if (!applyLinearFit(reference_view, view,
                                        b_protect_null_RGB_channels,
                                        b_continue_on_error,
                                        b_silent_on_error))
                    {
                         throw new Error("Linear Fit failed.");
                    }

                    this.setSettingsChangedFlag_implementation(true);

                    console.writeln( "" );
                    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
                    console.writeln( "<b>END: linearFitImage_implementation()</b>" );
                    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

                    return true;
                }
                else
                {
                    let message = "Target image is not same size as reference image. Aborting.";

                    throw new Error(message);
                }
            }
            else
            {

             console.writeln( "" );
             console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
             console.writeln( "<b>END: linearFitImage_implementation()</b>" );
             console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

             return true;  // No need to throw an Error, just skip
            }
        }
        catch (exception)
        {
            console.writeln("");
            console.writeln("------------------------------------------------------------------------------");
            console.writeln("");
            console.writeln(function_name + " : " + exception.message);

            if (!b_silent_on_error)
            {
                console.writeln("");
                console.writeln("------------------------------------------------------------------------------");
                console.writeln("");

                (new MessageBox(function_name + "<p>" + exception.message + "</p>", "ERROR", StdIcon_Error, StdButton_Ok)).execute();
            }
            else
            {
                console.writeln("b_silent_on_error is set to true; returning without MessageBox");
                console.writeln("");
                console.writeln("------------------------------------------------------------------------------");
                console.writeln("");
            }

            this.setSettingsChangedFlag_implementation(true);

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: linearFitImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return false;
        }
    };

    //

    this.saveColorizedImage_implementation = function(color_image, name,
                                                      b_with_auto_stretch,
                                                      folder_path,
                                                      file_extension = ".xisf")
    {
        let function_name = "saveColorizedImage_implementation()"

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>saveColorizedImage_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        try
        {
            let duplicate = createDuplicate(color_image);

            let colorized_image_file_name  = name + file_extension;

            if (!File.directoryExists(folder_path))
            {
                File.createDirectory(folder_path);
            }

            if (b_with_auto_stretch)
            {
                applyResetScreenTransferFunction(duplicate); // ie no STF as well as permanent stretch

                applyAutoHistogramTransformation(duplicate, true);  // true == linked channels
            }

            duplicate.window.saveAs(folder_path + "/" + colorized_image_file_name, false, false, false, false);

            closeWindow(duplicate);
            duplicate = null;

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: saveColorizedImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        }
        catch (exception)
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: saveColorizedImage_implementation()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            errorProcessing_standard(function_name, exception.message);
        }
    };

    //

    this.closeWorkingImages_implementation = function()
    {

        // This function closes hidden working images stored in working_view_array[],

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>closeWorkingImages_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        for (let i = 0; i < NUM_VIEW_SLOTS; i++)
        {
            if (this.working_view_array[i] && ! this.working_view_array[i].isNull)
            {
                closeWindow(this.working_view_array[i] );
                this.working_view_array[i]  = null;
            }
            if (this.working_view_for_colorized_images_array[i] && ! this.working_view_for_colorized_images_array[i].isNull)
            {
                closeWindow(this.working_view_for_colorized_images_array[i] );
                this.working_view_for_colorized_images_array[i]  = null;
            }
            this.image_attributes_array[i].b_active = false;
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: closeWorkingImages_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    };

    //

    this.close_implementation = function()
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>close_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        if (this.review_image_view && ! this.review_image_view.isNull)
        {
             closeWindow(this.review_image_view);
             this.review_image_view = null;
        }

        this.closeCombinedImage_implementation(": using Window Close Button");

        this.closeWorkingImages_implementation();

        this.colorized_image_preview.close();

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: close_implementation()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    };

};

TheApplication.prototype = new Object;

///////////////////////////////////////////////////////////////////////////////
// Class: ColorizedImage()
///////////////////////////////////////////////////////////////////////////////

function ColorizedImage(parent, name = "COLORIZED_IMAGE", width = -1, height = -1)
{
    this.__base__ = Object;
    this.__base__();

    // MEMBER PROPERTIES -------------------------------------------------------

    this.view = null;

    this.parent = parent;

    this.width  = width;
    this.height = height;

    // MEMBER METHODS ----------------------------------------------------------

    this.reset = function(width = -1, height = -1)
    {
        if (width < 1 || height < 1)
        {
           return false;
        }

        // Only recreate ImageWindow if width or height is changed
        if (this.width != width || !this.height != height || !this.view)
        {
            this.width  = width;
            this.height = height;

            // close any existing window
            if (this.view)
            {
                closeWindow(this.view);
                this.view = null;
            }

            let colorized_image_window = new ImageWindow(this.width, this.height,
                                                         3,    // 3 channels only
                                                         32,   // Min 32 bitsPerSample,
                                                         true, // floatSample
                                                         true, // Colour image
                                                         name);

            this.view = colorized_image_window.mainView;
        }

        gc(); // Garbage collection

        return true;
    };

    //

    this.colorize = function(input_view_grayscale,
                             base_color_red_channel, base_color_green_channel, base_color_blue_channel,
                             colorize_method)
    {
        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>ColorizedImage.colorize()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln("Red      : " + base_color_red_channel   * 255) ;
        console.writeln("Green    : " + base_color_green_channel * 255) ;
        console.writeln("Blue     : " + base_color_blue_channel  * 255) ;
        console.writeln("Colorize : " + colorize_method) ;
        console.writeln( "<b>", "----------------------------------------------------------------------", "</b>" );
        console.writeln("");

        let function_name = "ColorizedImage.colorize()"

        if (! input_view_grayscale || input_view_grayscale.isNull)
        {
            throw new Error(function_name + " - input_view_grayscale is Null ");
        }

        if (! this.reset(input_view_grayscale.image.width,
                        input_view_grayscale.image.height))
        {
            throw new Error(function_name + " - ColorizedImage.reset failed");
        }

        if (! input_view_grayscale || input_view_grayscale.isNull)
        {
            throw new Error(function_name + " - input_view_grayscale is Null ");
        }

        let b_use_single_expression = false;
        let b_rescale               = false;
        let b_create_new_image      = false;
        let colorSpace              = PixelMath.prototype.RGB;
        let b_show_new_image        = false;
        let new_image_id            = "updated_image";

        if (colorize_method.toUpperCase() == "MULTIPLY")
        {

                // Simple multiplication to 'modulate' the color

                let expression  = input_view_grayscale.id + " * "  + base_color_red_channel.toString();
                let expression1 = input_view_grayscale.id + " * "  + base_color_green_channel.toString();
                let expression2 = input_view_grayscale.id + " * "  + base_color_blue_channel.toString();
                let expression3 = "";
                let symbols     = "";

                applyPixelMath(this.view,
                               expression, expression1, expression2, expression3,
                               b_use_single_expression,
                               symbols, b_rescale, b_create_new_image,
                               new_image_id, colorSpace, b_show_new_image);

        }
        else
        {
            // First prepare a 'flat' color swatch with no input from the grayscale image

            let expression  = base_color_red_channel.toString();
            let expression1 = base_color_green_channel.toString();
            let expression2 = base_color_blue_channel.toString();
            let expression3 = "";
            let symbols     = "";

            applyPixelMath(this.view,
                           expression, expression1, expression2, expression3,
                           b_use_single_expression,
                           symbols, b_rescale, b_create_new_image,
                           new_image_id, colorSpace, b_show_new_image);

            // Now use the grayscale image to replace the HSI Intensity or HSV Value in the color swatch

            if (! input_view_grayscale || input_view_grayscale.isNull)
            {
                throw new Error(function_name + " - input_view_grayscale is Null ");
            }

            if (colorize_method.toUpperCase() == "HSI INTENSITY REPLACE")
            {
                // Replace the I (Intensity) Channel using the HSI model
                let HSI_channels = extractHSIChannels(this.view);

                if (input_view_grayscale  && ! input_view_grayscale.isNull)
                {
                    let P = new ChannelCombination;
                    P.colorSpace = ChannelCombination.prototype.HSI;
                    P.channels = [ // enabled, id
                                     [true, HSI_channels.H.id],
                                     [true, HSI_channels.S.id],
                                     [true, input_view_grayscale.id]
                                 ];
                    executeOn(this.view, P);
                }
                else
                {
                   throw new Error(function_name + " - input_view_grayscale is Null ");
                }

                closeWindow(HSI_channels.H);
                HSI_channels.H = null;
                closeWindow(HSI_channels.S);
                HSI_channels.S = null;
                closeWindow(HSI_channels.I);
                HSI_channels.I = null;
            }
            else if (colorize_method.toUpperCase() == "HSV VALUE REPLACE")
            {
                // Replace the V (Value) Channel using the HSV model
                let HSV_channels = extractHSVChannels(this.view);

                if (input_view_grayscale  && ! input_view_grayscale.isNull)
                {
                    let P = new ChannelCombination;
                    P.colorSpace = ChannelCombination.prototype.HSV;
                    P.channels = [ // enabled, id
                                     [true, HSV_channels.H.id],
                                     [true, HSV_channels.S.id],
                                     [true, input_view_grayscale.id]
                                 ];

                    executeOn(this.view, P);
                }
                else
                {
                    throw new Error(function_name + " - input_view_grayscale is Null ");
                }

                closeWindow(HSV_channels.H);
                HSV_channels.H = null;
                closeWindow(HSV_channels.S);
                HSV_channels.S = null;
                closeWindow(HSV_channels.V);
                HSV_channels.V = null;
            }
            else if (colorize_method.toUpperCase() == "LRGBCOMBINATION")
            {

                if (input_view_grayscale  && ! input_view_grayscale.isNull)
                {
                    let P = new RGBWorkingSpace;
                    P.channels = [ // Y, x, y
                                     [1.000000, 0.648431, 0.330856],
                                     [1.000000, 0.321152, 0.597871],
                                     [1.000000, 0.155886, 0.066044]
                                 ];

                    P.gamma = 1.00;
                    P.sRGBGamma = false;
                    P.applyGlobalRGBWS = false;
                    executeOn(this.view, P);
                    executeOn(input_view_grayscale, P);

                    let P2 = new LRGBCombination;
                    P2.channels = [ // enabled, id, k
                                      [false, "", 1.00000],
                                      [false, "", 1.00000],
                                      [false, "", 1.00000],
                                      [true, input_view_grayscale.id, 1.00000]
                                  ];
                    P2.mL = 0.500;
                    P2.mc = 0.500;
                    P2.clipHighlights = true;
                    P2.noiseReduction = false;
                    P2.layersRemoved = 4;
                    P2.layersProtected = 2;

                    executeOn(this.view, P2);
                }
                else
                {
                    throw new Error(function_name + " - input_view_grayscale is Null ");
                }
            }
        }

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: ColorizedImage.colorize()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    };

    //

    this.clear = function()
    {
       if (this.view)
       {
            let b_use_single_expression = false;
            let b_rescale               = false;
            let b_create_new_image      = false;
            let colorSpace              = PixelMath.prototype.RGB;
            let b_show_new_image        = true;
            let new_image_id            = "updated_image";

            let expression  = "0.5";
            let expression1 = "0.5";
            let expression2 = "0.5";
            let expression3 = "";
            let symbols     = "";

            applyPixelMath(this.view,
                           expression, expression1, expression2, expression3,
                           b_use_single_expression,
                           symbols, b_rescale, b_create_new_image,
                           new_image_id, colorSpace, b_show_new_image);
       }
    };

    //

    this.save = function(name, b_with_auto_stretch, folder_path, file_extension = ".xisf")
    {
       this.parent.saveColorizedImage(this.view,
                                      name + "_colorized",
                                      b_with_auto_stretch,
                                      folder_path,
                                      file_extension);
    };

    //

    this.close = function()
    {
        if (this.view)
        {
            closeWindow(this.view);
            this.view = null;
            delete this.view;
        }
    };

};

ColorizedImage.prototype = new Object;

//------------------------------------------------------------------------------

#endif   // __DC_CombineImages_the_app_js

///////////////////////////////////////////////////////////////////////////////
// EOF DC_CombineImages_the_app.js
///////////////////////////////////////////////////////////////////////////////
