//////////////////////////////////////////////////////////////////////////////
// DC_ProcessHelpers_v3.00dx_extract_mod.js
///////////////////////////////////////////////////////////////////////////////
//
// A wrappper script to provide basic helper functions;
//
// Copyright (c) 2018-2022 Dean Carr
//
///////////////////////////////////////////////////////////////////////////////

#ifndef __DC_ProcessHelpers_v3_00dx_extract_mod_js
#define __DC_ProcessHelpers_v3_00dx_extract_mod_js

//------------------------------------------------------------------------------

#include <pjsr/ColorSpace.jsh>
#include <pjsr/UndoFlag.jsh>
#include <pjsr/StdIcon.jsh>
#include <pjsr/StdButton.jsh>

//------------------------------------------------------------------------------

// Default STF Parameters

// Shadows clipping point measured in sigma units from the main histogram peak.
#define SHADOWS_CLIP -1.25

// Target background in the [0,1] range.
#define TARGET_BKG    0.25

// Whether to apply the same STF to all channels, or treat each channel separately.
#define RGB_LINK      true

#define DEFAULT_FILE_EXT ".xisf"
#define DEFAULT_PATH     "E:/Temp"

///////////////////////////////////////////////////////////////////////////////
// Global Variables
///////////////////////////////////////////////////////////////////////////////

function GlobalHelperData()
{
   this.script_title     = "Helper Script"

   this.script_version   = "v3.00dx - Extract";

   this.script_file_path = #__FILE__;

   this.STF  = [ // c0, c1, m, r0, r1
                   [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                   [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                   [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                   [0.00000, 1.00000, 0.50000, 0.00000, 1.00000]
               ];

   this.H    = [ // c0, m, c1, r0, r1
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                  [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000]
              ];

   this.default_file_extension    = DEFAULT_FILE_EXT;

   this.channel_extraction_index = 0;

}

var g_helper_data = new GlobalHelperData();

///////////////////////////////////////////////////////////////////////////////
// Standard Error Processing Functions
///////////////////////////////////////////////////////////////////////////////

function errorProcessing_standard(function_name, message)
{

    let script_string = g_helper_data.script_title
                 + " "   + g_helper_data.script_version;

    let long_msg = "<p><b>Error</b>: " + message + "</p>";

    console.criticalln();
    console.criticalln("<p>" + script_string + " " + function_name + "</p>"
                      + long_msg);
    console.flush();

    if ( message.contains("abort"))
    {
        throw new Error(message);
    }
    else
    {
        let messageBox = new MessageBox("<p>" + function_name + "</p>" + long_msg,
                                        script_string,
                                        StdIcon_Error,
                                        StdButton_Abort,
                                        StdButton_Ignore);

        let result = messageBox.execute();

        if (result == StdButton_Abort )
        {
            throw new Error("abort");
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// Window Functions
///////////////////////////////////////////////////////////////////////////////

function closeWindow(view, bForced = true)
{
    let function_name = "closeWindow()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        if (bForced)
        {
            view.window.forceClose();
            view = null;
        }
        else
        {
            view.window.close();
            view = null;
        }

    }
    catch (exception)
    {
       errorProcessing_standard(function_name, exception.message);
    }
}

///////////////////////////////////////////////////////////////////////////////
// Process Application Functions
///////////////////////////////////////////////////////////////////////////////

function applyScreenTransferFunction(view, STF)
{
    let function_name = "applyScreenTransferFunction()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let P = new ScreenTransferFunction;
        P.STF = STF;
        P.interaction = ScreenTransferFunction.prototype.SeparateChannels;

        P.executeOn( view );
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);
    }
}

//------------------------------------------------------------------------------

function setAutoSTFparameters(view,
                              shadows_clipping  = SHADOWS_CLIP ,
                              target_background = TARGET_BKG,
                              b_rgb_linked      = RGB_LINK)
{
    let function_name = "setAutoSTFparameters()";

    g_helper_data.STF = [ // c0, c1, m, r0, r1
                            [0, 1, 0.5, 0, 1],
                            [0, 1, 0.5, 0, 1],
                            [0, 1, 0.5, 0, 1],
                            [0, 1, 0.5, 0, 1]
                        ];

    g_helper_data.H    = [ // c0, m, c1, r0, r1
                             [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                             [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                             [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                             [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000],
                             [0.00000000, 0.50000000, 1.00000000, 0.00000000, 1.00000000]
                         ];

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let stf = new ScreenTransferFunction;

        let n = view.image.isColor ? 3 : 1;

        let median = view.computeOrFetchProperty( "Median" );

        let mad = view.computeOrFetchProperty( "MAD" );
        mad.mul( 1.4826 ); // coherent with a normal distribution

        if ( b_rgb_linked )
        {
            /*
             * Try to find how many channels look as channels of an inverted image.
             * We know a channel has been inverted because the main histogram peak is
             * located over the right-hand half of the histogram. Seems simplistic
             * but this is consistent with astronomical images.
             */
            let invertedChannels = 0;
            for ( let c = 0; c < n; ++c )
                if ( median.at( c ) > 0.5 )
                    ++invertedChannels;

            if ( invertedChannels < n )
            {
                /*
                 * Noninverted image
                 */
                 let c0 = 0, m = 0;
                 for ( let c = 0; c < n; ++c )
                 {
                     if ( 1 + mad.at( c ) != 1 )
                     c0 += median.at( c ) + shadows_clipping * mad.at( c );
                     m  += median.at( c );
                 }
                 c0 = Math.range( c0/n, 0.0, 1.0 );
                 m = Math.mtf( target_background, m/n - c0 );

                 g_helper_data.STF = [ // c0, c1, m, r0, r1
                              [c0, 1, m, 0, 1],
                              [c0, 1, m, 0, 1],
                              [c0, 1, m, 0, 1],
                              [0, 1, 0.5, 0, 1]
                          ];
                 g_helper_data.H  = [ // c0,  m, c1, r0, r1
                             [0, 0.5, 1, 0, 1],
                             [0, 0.5, 1, 0, 1],
                             [0, 0.5, 1, 0, 1],
                             [c0, m,  1, 0, 1],
                             [0, 0.5, 1, 0, 1]
                         ];
            }
            else
            {
                 /*
                  * Inverted image
                  */
                c1 = 0, m = 0;
                for ( let c = 0; c < n; ++c )
                {
                    m  += median.at( c );
                    if ( 1 + mad.at( c ) != 1 )
                        c1 += median.at( c ) - shadows_clipping * mad.at( c );
                    else
                        c1 += 1;
                }
                c1 = Math.range( c1/n, 0.0, 1.0 );
                m = Math.mtf( c1 - m/n, target_background );

                g_helper_data.STF =
                [ // c0, c1, m, r0, r1
                              [0, c1, m, 0, 1],
                              [0, c1, m, 0, 1],
                              [0, c1, m, 0, 1],
                              [0, 1, 0.5, 0, 1]
                ];
                g_helper_data.H  =
                [ // c0,  m, c1, r0, r1
                             [0, 0.5, 1, 0, 1],
                             [0, 0.5, 1, 0, 1],
                             [0, 0.5, 1, 0, 1],
                             [0, m,  c1, 0, 1],
                             [0, 0.5, 1, 0, 1]
                ];
            }
        }
        else
        {
            /*
            * Unlinked RGB channnels: Compute automatic stretch functions for
            * individual RGB channels separately.
            */
            let A =
            [ // c0, c1, m, r0, r1
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1],
               [0, 1, 0.5, 0, 1]
            ];

            for ( let c = 0; c < n; ++c )
            {
                if ( median.at( c ) < 0.5 )
                {
                    /*
                    * Noninverted channel
                    */
                    let c0 = (1 + mad.at( c ) != 1) ? Math.range( median.at( c ) + shadows_clipping * mad.at( c ), 0.0, 1.0 ) : 0.0;
                    let m  = Math.mtf( target_background, median.at( c ) - c0 );
                    A[c] = [c0, 1, m, 0, 1];
                }
                else
                {
                   /*
                   * Inverted channel
                   */
                   let c1 = (1 + mad.at( c ) != 1) ? Math.range( median.at( c ) - shadows_clipping * mad.at( c ), 0.0, 1.0 ) : 1.0;
                   let m  = Math.mtf( c1 - median.at( c ), target_background );
                   A[c] = [0, c1, m, 0, 1];
                }
            }

            g_helper_data.STF = A;

            g_helper_data.H =
            [ // reverse STF order for 2nd and 3rd elements in each row
                [g_helper_data.STF[0][0],
                             g_helper_data.STF[0][2],
                             g_helper_data.STF[0][1],
                             g_helper_data.STF[0][3],
                             g_helper_data.STF[0][4]],
                [g_helper_data.STF[1][0],
                             g_helper_data.STF[1][2],
                             g_helper_data.STF[1][1],
                             g_helper_data.STF[1][3],
                             g_helper_data.STF[1][4]],
                [g_helper_data.STF[2][0],
                             g_helper_data.STF[2][2],
                             g_helper_data.STF[2][1],
                             g_helper_data.STF[2][3],
                             g_helper_data.STF[2][4]],
                [0, 0.5, 1, 0, 1],
                [0, 0.5, 1, 0, 1]
            ] ;
        }


        // NEW

        stf.STF = g_helper_data.STF;

        for ( let c = 0; c < n; ++c )
        {
            console.writeln( "channel #", c );
            console.writeln( format( "c0 = %.6f", stf.STF[c][0] ) );
            console.writeln( format( "m  = %.6f", stf.STF[c][2] ) );
            console.writeln( format( "c1 = %.6f", stf.STF[c][1] ) );
        }

        stf.executeOn( view );
    }
    catch (exception)
    {
       errorProcessing_standard(function_name, exception.message);
    }
}

//------------------------------------------------------------------------------

function applyBringToFront(view)
{
    let function_name = "applyBringToFront()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );


        view.window.bringToFront();

    }
    catch (exception)
    {
       errorProcessing_standard(function_name, exception.message);
    }
}

//------------------------------------------------------------------------------

function applyResetScreenTransferFunction(view)
{
    let function_name = "applyResetScreenTransferFunction()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let STF = [ // c0, c1, m, r0, r1
                     [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                     [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                     [0.00000, 1.00000, 0.50000, 0.00000, 1.00000],
                     [0.00000, 1.00000, 0.50000, 0.00000, 1.00000]
                 ]

        applyScreenTransferFunction(view, STF, true);
    }
    catch (exception)
    {
       errorProcessing_standard(function_name, exception.message);
    }
}

//------------------------------------------------------------------------------

function applyAutoScreenTransferFunction(view, b_rgb_linked = RGB_LINK)
{
    let function_name = "applyAutoScreenTransferFunction()";
    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        setAutoSTFparameters(view, SHADOWS_CLIP, TARGET_BKG, b_rgb_linked);

        applyScreenTransferFunction(view, g_helper_data.STF);
    }
    catch (exception)
    {
       errorProcessing_standard(function_name, exception.message);
    }
}

//------------------------------------------------------------------------------

function applyHistogramTransformation(view, H)
{
    let function_name = "applyHistogramTransformation()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let P = new HistogramTransformation();
        P.H   = H;

        executeOn(view, P);
    }
    catch (exception)
    {
       errorProcessing_standard(function_name, exception.message);
    }
}

//------------------------------------------------------------------------------

function applyAutoHistogramTransformation(view, b_rgb_linked = RGB_LINK)
{
    let function_name = "applyAutoHistogramTransformation()";
    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        setAutoSTFparameters(view, SHADOWS_CLIP, TARGET_BKG, b_rgb_linked);

        applyHistogramTransformation(view, g_helper_data.H);

        applyResetScreenTransferFunction(view);

    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);
    }
}

//------------------------------------------------------------------------------

function applyLinearFit(source_view, target_view,
                        b_protect_null_channels = false,
                        b_continue_on_error     = false,
                        b_silent_on_error       = false)
{
    let function_name = "applyLinearFit()";

    try
    {
        if (source_view == null || source_view.isNull)
        {
            throw new Error("Null source view");
        }

        if (target_view == null || target_view.isNull)
        {
            throw new Error("Null target view");
        }

        console.writeln( "" );
        console.writeln( function_name);
        console.writeln( "<b>Source View           : </b>" + source_view.fullId      );
        console.writeln( "<b>Target View           : </b>" + target_view.fullId      );
        console.writeln( "<b>Protect null channels : </b>" + b_protect_null_channels );

        let numChannels_source = source_view.image.numberOfChannels;
        let numChannels_target = target_view.image.numberOfChannels;
        let b_grayscale = false;

        if (numChannels_source == 1 || numChannels_target == 1  )
        {
            b_grayscale = true;
        }

        if ((!b_grayscale && ! b_protect_null_channels) || b_grayscale)
        {
            console.writeln( "Linear Fit - Normal mode (either grayscale images or RGB channels not protected)");

            let P = new LinearFit;
            P.referenceViewId = source_view.id;
            P.rejectLow  = 0.000000;
            P.rejectHigh = 0.920000;

            if (!executeOn(target_view, P, true, true)) // true, true == bWriteSource, b_silent_on_error
            {
                throw new Error("Linear Fit failed");
            }
        }
        else
        {
            console.writeln( "" );
            console.writeln( "Linear Fit - Protected mode (zero-value channels will not be fitted)");
            console.writeln( "" );

            let protect_channel = [false, false, false];

            console.writeln("");
            console.writeln("Check the status of each channels of each image");
            console.writeln("");

            source_view.image.pushSelections();
            let max = 0.0; let min = 1.0;
            for (let channel = 0; channel < source_view.image.numberOfChannels; channel++)
            {
                source_view.image.selectedChannel = channel;
                max = source_view.image.maximum();
                min = source_view.image.minimum();
                console.writeln( "<end><cbr/><br/>" );
                console.writeln( "<end><cbr/><br/><b>","Source View - Channel " + channel + " Min : " + min + " Max: " + max, "</b>" );
                if (max  <= 0 || min  >= 1)
                {
                    protect_channel[channel] = true;
                }
            }
            source_view.image.popSelections();

            target_view.image.pushSelections();
            max = 0.0; min = 1.0;
            for (let channel = 0; channel < target_view.image.numberOfChannels; channel++)
            {
                target_view.image.selectedChannel = channel;
                max = target_view.image.maximum();
                min = target_view.image.minimum();
                console.writeln( "<end><cbr/><br/>" );
                console.writeln( "<end><cbr/><br/><b>","Target View - Channel " + channel + " Min : " + min + " Max: " + max, "</b>" );
                if (max  <= 0 || min  >= 1)
                {
                    protect_channel[channel] = true;
                }
            }
            target_view.image.popSelections();

            // If no valid channels, quit

            if (protect_channel[0] && protect_channel[1] && protect_channel[2])
            {
                throw new Error("Linear Fit failed - All channels have a zero value in either source or target images");  // Cannnot continue
            }

            // If at least one channel is ok, continue

            // split channels
            let source_RGBChannels = extractRGBChannels(source_view);
            let target_RGBChannels = extractRGBChannels(target_view);

            // Linear fit corresponding channels if valid
            let min_num_channels = Math.min(source_view.image.numberOfChannels,
                                            target_view.image.numberOfChannels);

            for (let channel = 0; channel < min_num_channels; channel++)
            {
                if (!protect_channel[channel])
                {
                    let P = new LinearFit;
                    P.rejectLow  = 0.000000;
                    P.rejectHigh = 0.920000;

                    if (channel == 0)
                    {
                        console.writeln( "Fitting Red channel");
                        P.referenceViewId = source_RGBChannels.red.id;
                        if (!executeOn(target_RGBChannels.red, P, true, true ) && ! b_continue_on_error) // true == b_silent_on_error) && ! b_continue_on_error)
                        {
                            throw new Error("Linear Fit has failed for channel 0 (Red)");
                        }
                    }
                    else if (channel == 1)
                    {
                        console.writeln( "Fitting Green channel");
                        P.referenceViewId = source_RGBChannels.green.id;
                        if (!executeOn(target_RGBChannels.green, P, true, true ) && ! b_continue_on_error) // true == b_silent_on_error) && ! b_continue_on_error)
                        {
                            throw new Error("Linear Fit has failed for channel 1 (Green)");
                        }
                    }
                    else if (channel == 2)
                    {
                        console.writeln( "Fitting Blue channel");
                        P.referenceViewId = source_RGBChannels.blue.id;
                        if (!executeOn(target_RGBChannels.blue, P, true, true ) && ! b_continue_on_error) // true == b_silent_on_error) && ! b_continue_on_error)
                        {
                            throw new Error("Linear Fit has failed for channel 2 (Blue)");
                        }
                    }
                }
                else
                {
                    console.writeln("");
                    console.writeln("Protected Mode - Rejecting channel : " + channel);
                    console.writeln("");
                }
            }

            // Recombine channels

            let P = new ChannelCombination;
            P.colorSpace = ChannelCombination.prototype.RGB;
            P.channels = [ // enabled, id
                             [true, target_RGBChannels.red.id],
                             [true, target_RGBChannels.green.id],
                             [true, target_RGBChannels.blue.id ]
                         ];

            executeOn(target_view, P);

            closeWindow(source_RGBChannels.red);
            source_RGBChannels.red = null;
            closeWindow(source_RGBChannels.green);
            source_RGBChannels.green = null;
            closeWindow(source_RGBChannels.blue);
            source_RGBChannels.blue = null;

            source_RGBChannels = null;

            closeWindow(target_RGBChannels.red);
            target_RGBChannels.red = null;
            closeWindow(target_RGBChannels.green);
            target_RGBChannels.green = null;
            closeWindow(target_RGBChannels.blue);
            target_RGBChannels.blue = null;

            target_RGBChannels = null;
        }

        target_view.id = target_view.fullId + "_f";

        return true;

    }
    catch (exception)
    {
        console.writeln("");
        console.writeln("------------------------------------------------------------------------------");
        console.writeln("");
        console.writeln(function_name + " : " + exception.message);

        if (!b_silent_on_error)
        {
            console.writeln("");
            console.writeln("------------------------------------------------------------------------------");
            console.writeln("");

            (new MessageBox(function_name + "<p>" + exception.message + "</p>", "ERROR", StdIcon_Error, StdButton_Ok)).execute();
        }
        else
        {
           console.writeln("b_silent_on_error is set to true; returning without MessageBox");
           console.writeln("");
           console.writeln("------------------------------------------------------------------------------");
           console.writeln("");
        }

        return false;
    }
}

//------------------------------------------------------------------------------

function applyPixelMath_scary(view,
                        expression, expression1,  expression2, expression3,
                        useSingleExpression,
                        symbols, rescale, createNewImage, newImageId,
                        colorSpace = PixelMath.prototype.SameAsTarget,
                        bShowNewImage = true, bTruncate = true)
{
    let function_name = "applyPixelMath_scary()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let P = new PixelMath;
        P.expression           = expression;
        P.expression1          = expression1;
        P.expression2          = expression2;
        P.expression3          = expression3;
        P.useSingleExpression  = useSingleExpression;
        P.symbols              = symbols;
        P.generateOutput       = true;
        P.singleThreaded       = false;
        P.use64BitWorkingImage = false;
        P.rescale              = rescale;
        P.rescaleLower         = 0;
        P.rescaleUpper         = 1;
        P.truncate             = bTruncate;
        P.truncateLower        = 0;
        P.truncateUpper        = 1;
        P.createNewImage       = createNewImage;
        P.showNewImage         = bShowNewImage;
        P.newImageId           = newImageId;
        P.newImageWidth        = 0;
        P.newImageHeight       = 0;
        P.newImageAlpha        = false;
        P.newImageColorSpace   = colorSpace;
        P.newImageSampleFormat = PixelMath.prototype.SameAsTarget;

        executeOn(view, P);
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);
    }
}

//------------------------------------------------------------------------------

function applyPixelMath(view,
                        expression, expression1,  expression2, expression3,
                        useSingleExpression,
                        symbols, rescale, createNewImage, newImageId,
                        colorSpace = PixelMath.prototype.SameAsTarget,
                        bShowNewImage = true)
{
    let function_name = "applyPixelMath()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let P = new PixelMath;
        P.expression           = expression;
        P.expression1          = expression1;
        P.expression2          = expression2;
        P.expression3          = expression3;
        P.useSingleExpression  = useSingleExpression;
        P.symbols              = symbols;
        P.generateOutput       = true;
        P.singleThreaded       = false;
        P.use64BitWorkingImage = false;
        P.rescale              = rescale;
        P.rescaleLower         = 0;
        P.rescaleUpper         = 1;
        P.truncate             = true;
        P.truncateLower        = 0;
        P.truncateUpper        = 1;
        P.createNewImage       = createNewImage;
        P.showNewImage         = bShowNewImage;
        P.newImageId           = newImageId;
        P.newImageWidth        = 0;
        P.newImageHeight       = 0;
        P.newImageAlpha        = false;
        P.newImageColorSpace   = colorSpace;
        P.newImageSampleFormat = PixelMath.prototype.SameAsTarget;

        executeOn(view, P);
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);
    }
}

///////////////////////////////////////////////////////////////////////////////
// Create Functions
///////////////////////////////////////////////////////////////////////////////

function createDuplicate(view, duplicateViewID = "DuplicateView",
                         bVisible = true)
{
    let function_name = "createDuplicate()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let duplicateWindow = new ImageWindow(view.image.width,
                                              view.image.height,
                                              view.image.numberOfChannels,
                                              view.window.bitsPerSample,
                                              view.window.isFloatSample,
                                              view.image.colorSpace != ColorSpace_Gray,
                                              duplicateViewID);
        let duplicateView = duplicateWindow.mainView;

        duplicateView.id = duplicateViewID;

        console.writeln( "" );
        console.writeln( "<b>Source View    : ", view.fullId, "</b>:" );
        console.writeln( "<b>Duplicate View : ", duplicateView.fullId, "</b>:" );

        duplicateView.beginProcess(UndoFlag_NoSwapFile);
        duplicateView.image.assign(view.image );
        duplicateView.endProcess();

        duplicateWindow.visible = bVisible;

        return duplicateView;
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);
        return null;
    }
}

//------------------------------------------------------------------------------

function createDuplicateWithHisto(view, duplicateViewID = "DuplicateView", bVisible = true)
{
    let function_name = "createDuplicateWithHisto()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let duplicateView = createDuplicate(view, duplicateViewID, bVisible);

        applyAutoHistogramTransformation(duplicateView);

        return duplicateView;
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);
        return null;
    }
}

///////////////////////////////////////////////////////////////////////////////
// Utility Functions
///////////////////////////////////////////////////////////////////////////////

function executeOn(view, process, bWriteSource = true, b_silent_on_error = false )
{
    let function_name = "executeOn()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        if (bWriteSource)
        {
            let src = process.toSource();
            console.writeln( "<end><cbr/><br/>" );
            console.writeln( "<end><cbr/><br/><b>","Process Source:", "</b>" );
            console.writeln( "<end><cbr/><br/>" );
            console.writeln( "<end><cbr/><br/><b>", src, "</b>" );
        }

        view.beginProcess(UndoFlag_NoSwapFile);
        let result = process.executeOn(view);
        view.endProcess();

        if (!result)
        {
            throw new Error("["+ process.processId() + "] : failed");
        }

        return true;
    }
    catch (exception)
    {
        console.writeln("");
        console.writeln("------------------------------------------------------------------------------");
        console.writeln("");
        console.writeln(function_name + " : " + exception.message);

        if (!b_silent_on_error)
        {
            console.writeln("");
            console.writeln("------------------------------------------------------------------------------");
            console.writeln("");

            (new MessageBox(function_name + "<p>" + exception.message + "</p>", "ERROR", StdIcon_Error, StdButton_Ok)).execute();
        }
        else
        {
           console.writeln("b_silent_on_error is set to true; returning without MessageBox");
           console.writeln("");
           console.writeln("------------------------------------------------------------------------------");
           console.writeln("");
        }

        return false;
    }
};

//------------------------------------------------------------------------------

function blackpointClip(view, bLinked = true, bRescale = false)
{
    let function_name = "blackpointClip()";

    try
    {
        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        let numChannels = view.image.numberOfChannels;
        let channelMinimum = 1;
        let channelMinimumArray = [];
        let minimumMinimum = 1;

        let bUseSingleExpression  = false;
        let bCreateNewImage       = false;
        let colorSpace            = PixelMath.prototype.RGB;
        let bShowNewImage         = true;
        let newImageId            = view.fullId + "_BlackpointClipped";

        let expression  = "";
        let expression1 = "";
        let expression2 = "";
        let expression3 = "";
        let symbols     = "";

        view.image.pushSelections();
        console.writeln( "<end><cbr/><br/>" );
        for (let channel = 0; channel < numChannels; channel++)
        {
            view.image.selectedChannel = channel;
            channelMinimum = view.image.minimum();
            channelMinimumArray.push(channelMinimum);
            minimumMinimum = Math.min(channelMinimum, minimumMinimum );
            console.writeln( "<end><cbr/><br/><b>","Channel " + channel + " Minumum: " + channelMinimum, "</b>" );
        }
        view.image.popSelections();
        console.writeln( "<end><cbr/><br/><b>","Minimum for all channels: " + minimumMinimum, "</b>" );

        if (bLinked || numChannels == 1)
        {
            bUseSingleExpression = true;
            expression = "$T" + "-" + minimumMinimum;
        }
        else
        {
            bUseSingleExpression = false;
            if (numChannels > 0)
            {
                expression = "$T" + "-" + channelMinimumArray[0];
            }
            if (numChannels > 1)
            {
                expression1 = "$T[1]" + "-" + channelMinimumArray[1];
            }
            if (numChannels > 2)
            {
                expression2 = "$T[2]" + "-" + channelMinimumArray[2];
            }
            if (numChannels > 3)
            {
                expression3 = "$T[3]" + "-" + channelMinimumArray[3];
            }
        }

        applyPixelMath(view,
                   expression, expression1, expression2, expression3,
                   bUseSingleExpression,
                   symbols, bRescale, bCreateNewImage,
                   newImageId, colorSpace, bShowNewImage);

        view.id = view.fullId + "_BPC";

        return true;

    }
    catch (exception)
    {

        (new MessageBox(function_name + "<p>" + exception.message + "<p>", "ERROR", StdIcon_Error, StdButton_Ok)).execute();

        return false;
    }
};

//------------------------------------------------------------------------------

function FindFiles( baseDirectory, templates, recursive )
{
   if ( !baseDirectory.endsWith( '/' ) )
      baseDirectory += '/';

   if ( templates === undefined )
      templates = ['*'];
   else if ( !(templates instanceof Array) )
      templates = [templates];

   if ( recursive === undefined )
      recursive = true;

   let files = [];
   for ( let i = 0; i < templates.length; ++i )
      files.push( searchDirectory( baseDirectory + templates[i], recursive ) );

   return files;
};

///////////////////////////////////////////////////////////////////////////////
// Utility Functions - MOD (Pending inclusion in new primary Process Helpers Script)
///////////////////////////////////////////////////////////////////////////////

function convertHSVtoRGB(hue, saturation, value)
{

    let red = 0.0, green = 0.0, blue = 0.0;

    let i = Math.floor(hue * 6.0);
    let f = hue * 6 - i;
    let p = value * (1 - saturation);
    let q = value * (1 - f * saturation);
    let t = value * (1 - (1 - f) * saturation);

    switch (i % 6)
    {
        case 0: red = value, green = t,     blue = p;
                break;
        case 1: red = q,     green = value, blue = p;
                break;
        case 2: red = p,     green = value, blue = t;
                break;
        case 3: red = p,     green = q,     blue = value;
                break;
        case 4: red = t,     green = p,     blue = value;
                break;
        case 5: red = value, green = p,     blue = q;
                break;
    }

    return [ red, green, blue ];

}

//------------------------------------------------------------------------------

function convertRGBtoHSV(red, green, blue)
{

   let max = Math.max(red, green, blue), min = Math.min(red, green, blue);
   let hue, saturation, value = max;

   let d = max - min;
   saturation = max == 0 ? 0 : d / max;

   if (max == min)
   {
     hue = 0; // achromatic
   }
   else
   {
       switch (max)
       {
           case red: hue = (green - blue) / d + (green < blue ? 6 : 0);
                break;

           case green: hue = (blue - red) / d + 2;
                break;

           case blue: hue = (red - green) / d + 4;
                break;
      }

      hue /= 6;
  }

  return [ hue, saturation, value ];
}

//------------------------------------------------------------------------------

function RGBCOLOR()
{
    this.__base__ = Object;
    this.__base__();

    this.red   = 0;
    this.green = 0;
    this.blue  = 0;
}
RGBCOLOR.prototype = new Object;

//------------------------------------------------------------------------------

function HSICOLOR()
{
    this.__base__ = Object;
    this.__base__();

    this.H = null;
    this.S = null;
    this.I = null;
}
HSICOLOR.prototype = new Object;

//------------------------------------------------------------------------------

function HSVCOLOR()
{
    this.__base__ = Object;
    this.__base__();

    this.H = null;
    this.S = null;
    this.V = null;
}
HSVCOLOR.prototype = new Object;

//------------------------------------------------------------------------------

function extractHSIChannels(view)
{
    let function_name = "extractHSIChannels()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let H_id   = "_H_"  + g_helper_data.channel_extraction_index++;
        let Si_id  = "_Si_" + g_helper_data.channel_extraction_index++;
        let I_id   = "_I_"  + g_helper_data.channel_extraction_index++;

        let P = new ChannelExtraction;
        P.colorSpace = ChannelExtraction.prototype.HSI;
        P.channels =
        [ // enabled, id
            [true, H_id],
            [true, Si_id],
            [true, I_id]
        ];
        P.sampleFormat = ChannelExtraction.prototype.SameAsSource;

        executeOn(view, P);

        let HSI_channels = new HSICOLOR();

        // Provide access to the H Channel
        HSI_channels.H = ImageWindow.windowById(H_id).mainView;
        // Handle error if mask cannot be found
        if (HSI_channels.H.isNull)
        {
            throw Error( "Internal error: 'H' image not found" );
        }
        // Provide access to the S Channel
        HSI_channels.S = ImageWindow.windowById(Si_id).mainView;
        // Handle error if mask cannot be found
        if (HSI_channels.S.isNull)
        {
            throw Error( "Internal error: 'Si' image not found" );
        }
        // Provide access to the I Channel
        HSI_channels.I = ImageWindow.windowById(I_id).mainView;
        // Handle error if mask cannot be found
        if (HSI_channels.I.isNull)
        {
            throw Error( "Internal error: 'I' image not found" );
        }

        console.writeln("");
        console.writeln("returning references: " + HSI_channels.H.id + " " + HSI_channels.S.id + " " + HSI_channels.I.id);
        console.writeln("");

        return HSI_channels;
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);

        return HSI_channels;

    }
}

//------------------------------------------------------------------------------

function extractHSVChannels(view)
{
    let function_name = "extractHSVChannels()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let H_id   = "_H_"  + g_helper_data.channel_extraction_index++;
        let Sv_id  = "_Sv_" + g_helper_data.channel_extraction_index++;
        let V_id   = "_V_"  + g_helper_data.channel_extraction_index++;

        let P = new ChannelExtraction;
        P.colorSpace = ChannelExtraction.prototype.HSV;
        P.channels =
        [ // enabled, id
            [true, H_id],
            [true, Sv_id],
            [true, V_id]
        ];
        P.sampleFormat = ChannelExtraction.prototype.SameAsSource;

        executeOn(view, P);

        let HSV_channels = new HSVCOLOR();

        // Provide access to the H Channel
        HSV_channels.H = ImageWindow.windowById(H_id).mainView;
        // Handle error if mask cannot be found
        if (HSV_channels.H.isNull)
        {
            throw Error( "Internal error: 'H' image not found" );
        }
        // Provide access to the S Channel
        HSV_channels.S = ImageWindow.windowById(Sv_id).mainView;
        // Handle error if mask cannot be found
        if (HSV_channels.S.isNull)
        {
            throw Error( "Internal error: 'Sv' image not found" );
        }
        // Provide access to the V Channel
        HSV_channels.V = ImageWindow.windowById(V_id).mainView;
        // Handle error if mask cannot be found
        if (HSV_channels.V.isNull)
        {
            throw Error( "Internal error: 'V' image not found" );
        }

        console.writeln("");
        console.writeln("returning references: " + HSV_channels.H.id + " " + HSV_channels.S.id + " " + HSV_channels.V.id);
        console.writeln("");

        return HSV_channels;
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);

        return HSV_channels;
    }
}

//------------------------------------------------------------------------------

function extractRGBChannels(view, outRGBChannels)
{
    let function_name = "extractRGBChannels()";

    try
    {
        if (view == null || view.isNull)
        {
            throw new Error("Null view");
        }

        console.writeln( "" );
        console.writeln( "<end><cbr/><br/><b>", function_name + " : </b>" + view.fullId );

        let red_id   = "_RED_"   + g_helper_data.channel_extraction_index++;
        let green_id = "_GREEN_" + g_helper_data.channel_extraction_index++;
        let blue_id  = "_BLUE_"  + g_helper_data.channel_extraction_index++;

        let P = new ChannelExtraction;
        P.colorSpace = ChannelExtraction.prototype.RGB;
        P.channels =
        [ // enabled, id
            [true, red_id],
            [true, green_id],
            [true, blue_id]
        ];
        P.sampleFormat = ChannelExtraction.prototype.SameAsSource;

        executeOn(view, P);

        // Provide access to the Red Channel
        let redChannel = ImageWindow.windowById(red_id).mainView;
        // Handle error if mask cannot be found
        if (redChannel.isNull)
        {
            throw Error( "Internal error: RED image found" );
        }
        // Provide access to the Green Channel
        let greenChannel = ImageWindow.windowById(green_id).mainView;
        // Handle error if mask cannot be found
        if (greenChannel.isNull)
        {
            throw Error( "Internal error: GREEN image found" );
        }
        // Provide access to the Blue Channel
        let blueChannel = ImageWindow.windowById(blue_id).mainView;
        // Handle error if mask cannot be found
        if (blueChannel.isNull)
        {
            throw Error( "Internal error: BLUE image found" );
        }

        let RGBChannels = new RGBCOLOR();
        RGBChannels.red   = redChannel;
        RGBChannels.green = greenChannel;
        RGBChannels.blue  = blueChannel;

        console.writeln("");
        console.writeln("returning references: " + RGBChannels.red.id + " " + RGBChannels.green.id + " " + RGBChannels.blue.id);
        console.writeln("");

        return RGBChannels;
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);

        return RGBChannels;
    }
}

//------------------------------------------------------------------------------

function readTextFile(filename, folder_path)
{

    let function_name = "readTextFile()"

    console.writeln( "" );
    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    console.writeln( "<b>readTextFile()</b>" );
    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    try
    {
        let buffer = readFile(filename, folder_path);

        if (buffer)
        {

            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: readTextFile()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return buffer.utf8ToString();
        }
        else
        {
            console.writeln( "" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
            console.writeln( "<b>END: readTextFile()</b>" );
            console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

            return "";

        }
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);

        return "";

    }

}

//------------------------------------------------------------------------------

function readFile(filename, folder_path)
{

    let function_name = "readFile()"

    console.writeln( "" );
    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    console.writeln( "<b>readFile()</b>" );
    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    try
    {
        console.writeln("");
        console.writeln("--- READING " + filename + " ---");
        console.writeln("");

        let file_path = folder_path +  "/" + filename;

        let file = new File;

        file.openForReading( file_path );

        let buffer = file.read( DataType_ByteArray, file.size );

        file.close();

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: readFile()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

        return buffer;

       }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);

        return null;

    }
}

//------------------------------------------------------------------------------

function saveTextFile(text, filename, folder_path)
{

    let function_name = "saveTextFile()"

    console.writeln( "" );
    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    console.writeln( "<b>saveTextFile()</b>" );
    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    try
    {

        console.writeln("");
        console.writeln("--- SAVING " + filename + " ---");
        console.writeln("");

        if (!File.directoryExists(folder_path))
        {
            File.createDirectory(folder_path);
        }

        let file_path = folder_path + "/" + filename;

        let file = File.createFileForWriting( file_path );

        file.write( text, DataType_UCString );

        file.close();

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: saveTextFile()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    }

    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);
    }

}

//------------------------------------------------------------------------------

function saveFile(data, filename, folder_path)
{

    let function_name = "saveFile()"

    console.writeln( "" );
    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
    console.writeln( "<b>saveFile()</b>" );
    console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    try
    {

        console.writeln("");
        console.writeln("--- SAVING " + filename + " ---");
        console.writeln("");

        if (!File.directoryExists(folder_path))
        {
            File.createDirectory(folder_path);
        }

        let file_path = folder_path + "/" + filename;

        let file = File.createFileForWriting( file_path );

        file.write( data );

        file.close();

        console.writeln( "" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );
        console.writeln( "<b>END: saveFile()</b>" );
        console.writeln( "<b>","----------------------------------------------------------------------", "</b>" );

    }

    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);
    }

}

//------------------------------------------------------------------------------

function openSelectedImage(file_path)
{

    let function_name = "openSelectedImage()"

    try
    {
        if (!File.exists(file_path ))
        {
            throw new Error("File not found: " + file_path);
        }

        let imageWin = ImageWindow.open(file_path);

        if(imageWin.isNull)
        {
            throw Error( "File not found: " + file_path);
        }
        else
        {
            imageWin[0].show();
        }

        return imageWin[0].mainView;
    }
    catch (exception)
    {
        errorProcessing_standard(function_name, exception.message);

        return null;
    }

}

//------------------------------------------------------------------------------

#endif   // __DC_ProcessHelpers_v3_00dx_extract_mod_js

// ****************************************************************************
// EOF DC_ProcessHelpers_v3.00dx_extract_mod.js
// ****************************************************************************
